#ifndef __CANNON_H__
#define __CANNON_H__

#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include <assert.h>

#include "globals.h"
#include "vect3.h"
#include "model.h"
#include "console.h"

///////////////////////////////////////////////////////////////
class cannon_t
{
public:
  cannon_t() : mValid( false ), mReadyToFire( true ) {}
  ~cannon_t() {}

	void set_pos( const vect3_t &aPos ) 
  { 
    mPos = aPos; 
    // Add 1/2 the block size to the position
    mPos.x += con_block_size * .5f; 
    mPos.z += con_block_size * .5f; 
  }

	const vect3_t &get_pos() const { return mPos; }

	void set_model( model_node_t *apModel ) { mpModel = apModel; }

  void update();
  void draw();

  void set_valid( bool aValid ) { mValid = aValid; }
  bool get_valid() const { return mValid; }

  bool fire();
  void ready_to_fire() { mReadyToFire = true; }

  void set_xz( int aX, int aZ ) { mX = aX; mZ = aZ; }
  void get_xz( int &aX, int &aZ ) { aX = mX; aZ = mZ; }

private:
  bool mValid;

  vect3_t mPos;

	model_node_t *mpModel;

  bool mReadyToFire;

  int mX, mZ;
};

#endif