#include "castle.h"

//////////////////////////////////////////////////
// Reset castle after "GAME OVER"
//////////////////////////////////////////////////
void 
castle_t::reset() 
{ 
  clear(); 
  set_pieces(); 
  disable_ghost();

  reset_conflict_pieces();

  for ( int i = 0; i < mCastle.size(); ++i )
  {
    mCastle[ i ].mPieceType = BLOCK;
    mCastle[ i ].mPieceRot = NO_ROT;
  }
}

//////////////////////////////////////////////////
void 
castle_t::reset_conflict_pieces()
{  
  // Reset conflict pieces
  for ( int i = 0; i < 4; ++i )
    mConflictPieces[ i ].mOK = true;

  mPieceX = mPieceZ = 0x7FFFFFFF;
}

//////////////////////////////////////////////////
// This function is for debug only
//////////////////////////////////////////////////
void 
castle_t::remove_all_pieces()
{
  for ( int i = 0; i < mCastle.size(); ++i )
  {
    if ( mCastle[ i ].mContents == PIECE )
      mCastle[ i ].mContents = EMPTY;  
  }

  mValidPieces.clear();
}

//////////////////////////////////////////////////
void 
castle_t::set_piece_models( model_node_t *apModels[] )
{
  for ( int i = 0; i < 11; ++i )
    mpPieceModels[ i ] = apModels[ i ];
}

//////////////////////////////////////////////////
void 
castle_t::set_cannon_model( model_node_t *apModel ) 
{ 
  mpCannonModel = apModel; 

  vect3_t lCannonMin, lCannonMax;
  mpCannonModel->get_box( lCannonMin, lCannonMax );
  mCannonOffX = con_block_size / 2.0f;
  mCannonOffZ = con_block_size / 2.0f;
}

//////////////////////////////////////////////////
void 
castle_t::set_flag_model( model_node_t *apModels[] )
{
  for ( int i = 0; i < 10; ++i )
    mpFlagModels[ i ] = apModels[ i ];
}

//////////////////////////////////////////////////
// Draws castle. Parameter aBuild is true when
// game is in the BUILD state, false when in the
// state
//////////////////////////////////////////////////
void
castle_t::draw( bool aLighting )
{	
  static int lName, i, lPieceIdx;
  static float lPosX, lPosZ, lOffX, lOffZ;
  static float lRotateAmount;
  lName = 1000;
  lRotateAmount = 0.0f;

  for ( i = 0; i < mCastle.size(); ++i )
  {
    const castle_piece_t &lCastle = mCastle[ i ];

    if ( lCastle.mContents )
    {
	    glPushMatrix(); 

      switch ( lCastle.mContents )
      {
      case PIECE :
        // set the model to the proper piece type
        if ( lCastle.mPieceType == WALL && con_geometry == 1 )
          mpPieceModel = mpPieceModels[ WALL_LOW ];
        else if ( lCastle.mPieceType == WALL && con_geometry == 2 )
          mpPieceModel = mpPieceModels[ WALL_HIGH ];
        else if ( lCastle.mPieceType == TEE && con_geometry == 1 )
          mpPieceModel = mpPieceModels[ TEE_LOW ];
        else if ( lCastle.mPieceType == TEE && con_geometry == 2 )
          mpPieceModel = mpPieceModels[ TEE_HIGH ];
        else if ( lCastle.mPieceType == CORNER && con_geometry == 1 )
          mpPieceModel = mpPieceModels[ CORNER_LOW ];
        else if ( lCastle.mPieceType == CORNER && con_geometry == 2 )
          mpPieceModel = mpPieceModels[ CORNER_HIGH ];
        else if ( lCastle.mPieceType == CROSS && con_geometry == 1 )
          mpPieceModel = mpPieceModels[ CROSS_LOW ];
        else if ( lCastle.mPieceType == CROSS && con_geometry == 2 )
          mpPieceModel = mpPieceModels[ CROSS_HIGH ];
        else if ( lCastle.mPieceType == TOWER && con_geometry == 1 )
          mpPieceModel = mpPieceModels[ TOWER_LOW ];
        else if ( lCastle.mPieceType == TOWER && con_geometry == 2 )
          mpPieceModel = mpPieceModels[ TOWER_HIGH ];
        else
          mpPieceModel = mpPieceModels[ BLOCK_LOW ];

        if ( lCastle.mPieceRot == QUARTER )
          lRotateAmount = 90.0f;
        else if ( lCastle.mPieceRot == HALF )
          lRotateAmount = 180.0f;
        else if ( lCastle.mPieceRot == THREE_QUARTERS )
          lRotateAmount = 270.0f;
        else
          lRotateAmount = 0.0f;

        if ( ( mDrawCannon && mPieceX == lCastle.mX && mPieceZ == lCastle.mZ ) ||
             ( !mConflictPieces[ 0 ].mOK && 
               mConflictPieces[ 0 ].mX == lCastle.mX && mConflictPieces[ 0 ].mZ == lCastle.mZ ) ||
             ( !mConflictPieces[ 1 ].mOK && 
               mConflictPieces[ 1 ].mX == lCastle.mX && mConflictPieces[ 1 ].mZ == lCastle.mZ ) ||
             ( !mConflictPieces[ 2 ].mOK && 
               mConflictPieces[ 2 ].mX == lCastle.mX && mConflictPieces[ 2 ].mZ == lCastle.mZ ) ||
             ( !mConflictPieces[ 3 ].mOK && 
               mConflictPieces[ 3 ].mX == lCastle.mX && mConflictPieces[ 3 ].mZ == lCastle.mZ ) )
        {
          if ( aLighting )
          {
            glDisable( GL_LIGHTING );
            glColor4f( 1, 0, 0, 0 );
            glTranslatef( (con_block_size*0.5f), 0.0, (con_block_size*0.5f) );
            glTranslatef( lCastle.mPos.x, lCastle.mPos.y, lCastle.mPos.z );
            glRotatef( lRotateAmount, 0.0, 1.0, 0.0 );
            glTranslatef( (-con_block_size*0.5f), 0.0, (-con_block_size*0.5f) );
            mpPieceModel->draw();
            glColor4f( 1, 1, 1, 1 );
            glEnable( GL_LIGHTING );
          }
          else
          {
            glColor4f( 1, 0, 0, 0 );
            glTranslatef( (con_block_size*0.5f), 0.0, (con_block_size*0.5f) );
            glTranslatef( lCastle.mPos.x, lCastle.mPos.y, lCastle.mPos.z );
            glRotatef( lRotateAmount, 0.0, 1.0, 0.0 );
            glTranslatef( (-con_block_size*0.5f), 0.0, (-con_block_size*0.5f) );
	          mpPieceModel->draw();
            glColor4f( 1, 1, 1, 1 );
          }
        }
        else
        {
          glTranslatef( (con_block_size*0.5f), 0.0, (con_block_size*0.5f) );
          glTranslatef( lCastle.mPos.x, lCastle.mPos.y, lCastle.mPos.z );
          glRotatef( lRotateAmount, 0.0, 1.0, 0.0 );
          glTranslatef( (-con_block_size*0.5f), 0.0, (-con_block_size*0.5f) );
	        mpPieceModel->draw();
        }
        break;
      
      case RUBBLE0 :
        glTranslatef( lCastle.mPos.x, lCastle.mPos.y, lCastle.mPos.z );
	      mpRubbleModel[ 0 ]->draw();
        break;

      case RUBBLE1 :
	      glTranslatef( lCastle.mPos.x, lCastle.mPos.y, lCastle.mPos.z );
	      mpRubbleModel[ 1 ]->draw();
        break;

      case CANNON :
	      glTranslatef( lCastle.mPos.x + mCannonOffX, 
                      lCastle.mPos.y, 
                      lCastle.mPos.z + mCannonOffZ );
	      mpCannonModel->draw();
        break;
      }

	    glPopMatrix();
    }
  }

  draw_flag();

  // Do we draw the ghost piece?
  if ( mGhostEnabled )
  {
    glEnable( GL_BLEND );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    glColor4f( 1, 1, 1, 0.5f );

    lPosX = mMin.x + ( mPieceX * con_block_size );
    lPosZ = mMin.z + ( mPieceZ * con_block_size );

    for ( i = 0; i < 4; ++i )
    {
      lOffX = lPosX + ( gPieceOffset[ mCurrPiece.mType ][ mCurrPiece.mAngle ].mX[ i ] * con_block_size );
      lOffZ = lPosZ + ( gPieceOffset[ mCurrPiece.mType ][ mCurrPiece.mAngle ].mZ[ i ] * con_block_size );

      glPushMatrix(); 
	    glTranslatef( lOffX, mMin.y, lOffZ );
	    mpPieceModel->draw();      
	    glPopMatrix();
    }

    glDisable( GL_BLEND );

    //  There is z fighting between the lines and polygons
    //  Maybe we need to scale the object slightly before drawing the lines
    //  We need a push and pop method to modify the render states
    //  Use glPushAttrib() and glPopAttrib()
    glPushAttrib( GL_ALL_ATTRIB_BITS );
    glDisable( GL_TEXTURE_2D );
    glDepthFunc( GL_LEQUAL );
    glColor3f( 1, 1, 1 );

    for ( i = 0; i < 4; ++i )
    {
      lOffX = lPosX + ( gPieceOffset[ mCurrPiece.mType ][ mCurrPiece.mAngle ].mX[ i ] * con_block_size );
      lOffZ = lPosZ + ( gPieceOffset[ mCurrPiece.mType ][ mCurrPiece.mAngle ].mZ[ i ] * con_block_size );

      glPushMatrix(); 
	    glTranslatef( lOffX, mMin.y, lOffZ );
	    mpPieceModel->draw_lines();      
	    glPopMatrix();
    }
    
    glPopAttrib();
  }

  static float lCannonX, lCannonZ;

  // Do we draw the cannon?
  if ( mDrawCannon )
  {
    glEnable( GL_BLEND );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    glColor4f( 1, 1, 1, 0.5f );

    lPosX = mMin.x + ( mPieceX * con_block_size );
    lPosZ = mMin.z + ( mPieceZ * con_block_size );

    lCannonX = lPosX + mCannonOffX;
    lCannonZ = lPosZ + mCannonOffZ;

    glPushMatrix(); 
	  glTranslatef( lCannonX, mMin.y, lCannonZ );
	  mpCannonModel->draw();
	  glPopMatrix();

    glDisable( GL_BLEND );
  }
}
/*
void
castle_t::draw( bool aLighting )
{	
  static int lName, i;
  static float lPosX, lPosZ, lOffX, lOffZ;
  lName = 1000;

  mpPieceModel->set_pointers();
  int lCastleSize = mCastle.size();
  for ( i = 0; i < lCastleSize; ++i )
  {
    const castle_piece_t &lCastle = mCastle[ i ];

    if ( lCastle.mContents == PIECE )
    {
	    glPushMatrix();
      if ( ( mDrawCannon && mPieceX == lCastle.mX && mPieceZ == lCastle.mZ ) ||
           ( !mConflictPieces[ 0 ].mOK && 
             mConflictPieces[ 0 ].mX == lCastle.mX && mConflictPieces[ 0 ].mZ == lCastle.mZ ) ||
           ( !mConflictPieces[ 1 ].mOK && 
             mConflictPieces[ 1 ].mX == lCastle.mX && mConflictPieces[ 1 ].mZ == lCastle.mZ ) ||
           ( !mConflictPieces[ 2 ].mOK && 
             mConflictPieces[ 2 ].mX == lCastle.mX && mConflictPieces[ 2 ].mZ == lCastle.mZ ) ||
           ( !mConflictPieces[ 3 ].mOK && 
             mConflictPieces[ 3 ].mX == lCastle.mX && mConflictPieces[ 3 ].mZ == lCastle.mZ ) )
      {
        if ( aLighting )
        {
          glDisable( GL_LIGHTING );
          glColor4f( 1, 0, 0, 0 );
          glTranslatef( lCastle.mPos.x, lCastle.mPos.y, lCastle.mPos.z );
	        mpPieceModel->draw_pointers();
          glColor4f( 1, 1, 1, 1 );
          glEnable( GL_LIGHTING );
        }
        else
        {
          glColor4f( 1, 0, 0, 0 );
          glTranslatef( lCastle.mPos.x, lCastle.mPos.y, lCastle.mPos.z );
	        mpPieceModel->draw_pointers();
          glColor4f( 1, 1, 1, 1 );
        }
      }
      else
      {
        glTranslatef( lCastle.mPos.x, lCastle.mPos.y, lCastle.mPos.z );
	      mpPieceModel->draw_pointers();
      }
    
      glPopMatrix();
    }
  }
  mpPieceModel->unset_pointers();

  mpRubbleModel[ 0 ]->set_pointers();
  for ( i = 0; i < lCastleSize; ++i )
  {
    const castle_piece_t &lCastle = mCastle[ i ];

    if ( lCastle.mContents == RUBBLE0 )
    {
	    glPushMatrix(); 
      glTranslatef( lCastle.mPos.x, lCastle.mPos.y, lCastle.mPos.z );
      mpRubbleModel[ 0 ]->draw_pointers();
	    glPopMatrix();
    }
  }
  mpRubbleModel[ 0 ]->unset_pointers();

  mpRubbleModel[ 1 ]->set_pointers();
  for ( i = 0; i < lCastleSize; ++i )
  {
    const castle_piece_t &lCastle = mCastle[ i ];

    if ( lCastle.mContents == RUBBLE1 )
    {
	    glPushMatrix(); 
      glTranslatef( lCastle.mPos.x, lCastle.mPos.y, lCastle.mPos.z );
      mpRubbleModel[ 1 ]->draw_pointers();
	    glPopMatrix();
    }
  }
  mpRubbleModel[ 1 ]->unset_pointers();

  mpCannonModel->set_pointers();
  for ( i = 0; i < mpCannons->size(); ++i )
  {
    if ( ( *mpCannons )[ i ].get_valid() )
    {
      const vect3_t &lCannonPos = ( *mpCannons )[ i ].get_pos();
	    glPushMatrix(); 
      glTranslatef( lCannonPos.x, lCannonPos.y, lCannonPos.z );
      mpCannonModel->draw_pointers();
	    glPopMatrix();
    }
  }
  mpCannonModel->unset_pointers();

  draw_flag();

  // Do we draw the ghost piece?
  if ( mGhostEnabled )
  {
    glEnable( GL_BLEND );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    glColor4f( 1, 1, 1, 0.5f );

    lPosX = mMin.x + ( mPieceX * con_block_size );
    lPosZ = mMin.z + ( mPieceZ * con_block_size );

    for ( i = 0; i < 4; ++i )
    {
      lOffX = lPosX + ( gPieceOffset[ mCurrPiece.mType ][ mCurrPiece.mAngle ].mX[ i ] * con_block_size );
      lOffZ = lPosZ + ( gPieceOffset[ mCurrPiece.mType ][ mCurrPiece.mAngle ].mZ[ i ] * con_block_size );

      glPushMatrix(); 
	    glTranslatef( lOffX, mMin.y, lOffZ );
	    mpPieceModel->draw();      
	    glPopMatrix();
    }

    glDisable( GL_BLEND );

    //  There is z fighting between the lines and polygons
    //  Maybe we need to scale the object slightly before drawing the lines
    //  We need a push and pop method to modify the render states
    //  Use glPushAttrib() and glPopAttrib()
    glPushAttrib( GL_ALL_ATTRIB_BITS );
    glDisable( GL_TEXTURE_2D );
    glDepthFunc( GL_LEQUAL );
    glColor3f( 1, 1, 1 );

    for ( i = 0; i < 4; ++i )
    {
      lOffX = lPosX + ( gPieceOffset[ mCurrPiece.mType ][ mCurrPiece.mAngle ].mX[ i ] * con_block_size );
      lOffZ = lPosZ + ( gPieceOffset[ mCurrPiece.mType ][ mCurrPiece.mAngle ].mZ[ i ] * con_block_size );

      glPushMatrix(); 
	    glTranslatef( lOffX, mMin.y, lOffZ );
	    mpPieceModel->draw_lines();      
	    glPopMatrix();
    }
    
    glPopAttrib();
  }

  static float lCannonX, lCannonZ;

  // Do we draw the cannon?
  if ( mDrawCannon )
  {
    glEnable( GL_BLEND );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    glColor4f( 1, 1, 1, 0.5f );

    lPosX = mMin.x + ( mPieceX * con_block_size );
    lPosZ = mMin.z + ( mPieceZ * con_block_size );

    lCannonX = lPosX + mCannonOffX;
    lCannonZ = lPosZ + mCannonOffZ;

    glPushMatrix(); 
	  glTranslatef( lCannonX, mMin.y, lCannonZ );
	  mpCannonModel->draw();
	  glPopMatrix();

    glDisable( GL_BLEND );
  }
}
*/
/////////////////////////////////////////////////
void
castle_t::clear_rubble()
{
  for ( int i = 0; i < mCastle.size(); ++i )
  {
    castle_piece_t *lpCastle = &( mCastle[ i ] );

    if ( lpCastle->mContents == RUBBLE0 ||  lpCastle->mContents == RUBBLE1 )
    {
      lpCastle->mContents = EMPTY;
    }
  }
}

/////////////////////////////////////////////////
void
castle_t::draw_flag()
{
  static float lCount = 0;
  static int lIndex = 0;

  lCount += gFactTick;
  if ( lCount > 10.0f )
  {
    lIndex = ( lIndex + 1 ) & 0x7;
    lCount = 0;
  }

  // Draw flag
  static int lPosX, lPosZ;
  
  lPosX = mMin.x + ( mFlagBlockX * con_block_size );
  lPosZ = mMin.z + ( mFlagBlockZ * con_block_size ) + con_block_size;
  
  glPushMatrix(); 
	glTranslatef( lPosX, mMin.y, lPosZ );
	mpFlagModels[ lIndex ]->draw();
	mpFlagModels[ 8 ]->draw();              // FlagPole
	mpFlagModels[ 9 ]->draw();              // FlagBase
	glPopMatrix();
}

/////////////////////////////////////////////////
void
castle_t::draw_debug()
{  
  glColor3f( 0, 0.65f, 0 );
  glBegin( GL_LINES );

  float lMaxX = mMin.x + mX * con_block_size;
  float lMaxZ = mMin.z + mZ * con_block_size;

  float lCurrX;
  float lCurrZ;

  lCurrZ = mMin.z;
  for ( int z = 0; z <= mZ; ++z )
  {
    lCurrX = mMin.x;
    for ( int x = 0; x <= mX; ++x )
    {
      if ( inpoly( mPolyPoints, lCurrX, lCurrZ ) )
      {
	      glVertex3f( lCurrX, mMin.y + 1, lCurrZ );
        glVertex3f( (lCurrX + con_block_size), mMin.y + 1, lCurrZ );
      }
      lCurrX += con_block_size;
    }
    lCurrZ += con_block_size;
  }

  lCurrX = mMin.x;
  for ( int x = 0; x <= mX; ++x )
  {
    lCurrZ = mMin.z;
    for ( int z = 0; z <= mZ; ++z )
    {
      if ( inpoly( mPolyPoints, lCurrX, lCurrZ ) )
      {
	      glVertex3f( lCurrX, mMin.y + 1, lCurrZ );
        glVertex3f( lCurrX, mMin.y + 1, (lCurrZ + con_block_size) );
      }
      lCurrZ += con_block_size;
    }
    lCurrX += con_block_size;
  }

  glEnd();
}

/////////////////////////////////////////////////
// Checks if e particular 'tetris' piece fits
// in the current location
/////////////////////////////////////////////////
bool 
castle_t::check( const vect3_t &aHitPos, const tetris_piece_t &aPiece, 
                 bool aClicked )
{
  static bool lChecked;

  if ( inpoly( mPolyPoints, aHitPos.x, aHitPos.z ) )
  {
    int lX, lZ;

    lX = ( int ) ( aHitPos.x - mMin.x ) / con_block_size;
    lZ = ( int ) ( aHitPos.z - mMin.z ) / con_block_size;

    mPieceX = lX;
    mPieceZ = lZ;

    //gCursor.disable();
    enable_ghost();

    lChecked = check_piece( lX, lZ, aPiece );
    if ( lChecked && aClicked )
    {
      insert_piece( lX, lZ, aPiece );
      return true;
    }
    else if ( aClicked )
    {
      mpSounds->play_sample( mNegateSound );
    }
  }
  else
  {
    gCursor.enable();
    disable_ghost();
  }

  return false;
}


/////////////////////////////////////////////////
// Checks if the cannon fits in the current location
/////////////////////////////////////////////////
bool 
castle_t::check( const vect3_t &aHitPos, bool aClicked )
{
  static bool lChecked, lFoundCannon;
  static vect3_t lCannonPos;
  static int i;

  lFoundCannon = false;

  // check the build area bounds as defined in the level file
  if ( inpoly( mPolyPoints, aHitPos.x, aHitPos.z ) )
  {
    static int lX, lZ;

    lX = ( int ) ( aHitPos.x - mMin.x ) / con_block_size;
    lZ = ( int ) ( aHitPos.z - mMin.z ) / con_block_size;

    mPieceX = lX;
    mPieceZ = lZ;

    gCursor.disable();
    enable_draw_cannon();

    lChecked = piece_ok( lX, lZ );
    if ( lChecked && aClicked && check_enclosure( lX, lZ ) )
    {
      insert_cannon( lX, lZ );

      lCannonPos.x = mMin.x + ( lX * con_block_size );
      lCannonPos.y = mMin.y;
      lCannonPos.z = mMin.z + ( lZ * con_block_size );

      // Insert new cannon at specified position
      cannon_t lCannon;
      lCannon.set_pos( lCannonPos );
      lCannon.set_valid( true );
      lCannon.set_xz( lX, lZ );

      for ( i = 0; i < mpCannons->size(); ++i )
      {
        if ( ( *mpCannons )[ i ].get_valid() == false )
        {
          ( *mpCannons )[ i ] = lCannon;
          lFoundCannon = true;
        }
      }

      if ( !lFoundCannon )
        mpCannons->push_back( lCannon );

      return true;
    }
    else if ( aClicked )
    {
      mpSounds->play_sample( mNegateSound );
    }
  }
  else
  {
    gCursor.enable();
    disable_draw_cannon();
  }

  return false;
}

/////////////////////////////////////////////////
bool 
castle_t::piece_ok( int aX, int aZ )
{
  return ( aX >= 0 && aX < mX && aZ >= 0 && aZ < mZ && 
           mCastle[ mZ * aX + aZ ].mContents == EMPTY );
}

/////////////////////////////////////////////////
void
castle_t::set_piece( int aX, int aZ )
{
  if ( !piece_ok( aX, aZ ) )
    return;

  int lIndex = mZ * aX + aZ;
  mCastle[ lIndex ].mContents = PIECE;
  mCastle[ lIndex ].mDestroyed = false;
  mCastle[ lIndex ].mShotAt = false;
  mCastle[ lIndex ].mIndex = lIndex;
  mCastle[ lIndex ].mX = aX;
  mCastle[ lIndex ].mZ = aZ;

  mValidPieces.push_back( lIndex );
}

/////////////////////////////////////////////////
void
castle_t::insert_cannon( int aX, int aZ )
{
  int lIndex = mZ * aX + aZ;
  mCastle[ lIndex ].mContents = CANNON;
  mCastle[ lIndex ].mDestroyed = false;
  mCastle[ lIndex ].mShotAt = false;
  mCastle[ lIndex ].mIndex = lIndex;
}

/////////////////////////////////////////////////
bool 
castle_t::check_piece( int aX, int aZ, const tetris_piece_t &aPiece )
{
  mConflictPieces[ 0 ].mOK = piece_ok( aX + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mX[ 0 ],
                                       aZ + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mZ[ 0 ] );
  mConflictPieces[ 1 ].mOK = piece_ok( aX + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mX[ 1 ],
                                       aZ + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mZ[ 1 ] );
  mConflictPieces[ 2 ].mOK = piece_ok( aX + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mX[ 2 ],
                                       aZ + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mZ[ 2 ] );
  mConflictPieces[ 3 ].mOK = piece_ok( aX + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mX[ 3 ],
                                       aZ + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mZ[ 3 ] );
  bool lRes = true;
  if ( !mConflictPieces[ 0 ].mOK )
  {
    mConflictPieces[ 0 ].mX = aX + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mX[ 0 ];
    mConflictPieces[ 0 ].mZ = aZ + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mZ[ 0 ];
    lRes = false;
  }
  if ( !mConflictPieces[ 1 ].mOK )
  {
    mConflictPieces[ 1 ].mX = aX + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mX[ 1 ];
    mConflictPieces[ 1 ].mZ = aZ + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mZ[ 1 ];
    lRes = false;
  }
  if ( !mConflictPieces[ 2 ].mOK )
  {
    mConflictPieces[ 2 ].mX = aX + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mX[ 2 ];
    mConflictPieces[ 2 ].mZ = aZ + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mZ[ 2 ];
    lRes = false;
  }
  if ( !mConflictPieces[ 3 ].mOK )
  {
    mConflictPieces[ 3 ].mX = aX + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mX[ 3 ];
    mConflictPieces[ 3 ].mZ = aZ + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mZ[ 3 ];
    lRes = false;
  }

  return lRes;
}

/////////////////////////////////////////////////
void 
castle_t::insert_piece( int aX, int aZ, const tetris_piece_t &aPiece )
{
  // only up to 4 blocks are allowed for a tetris piece
  set_piece( aX + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mX[ 0 ],
             aZ + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mZ[ 0 ] );
  set_piece( aX + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mX[ 1 ],
             aZ + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mZ[ 1 ] );
  set_piece( aX + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mX[ 2 ],
             aZ + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mZ[ 2 ] );
  set_piece( aX + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mX[ 3 ],
             aZ + gPieceOffset[ aPiece.mType ][ aPiece.mAngle ].mZ[ 3 ] );
}

/////////////////////////////////////////////////
// Returns a pointer to the castle piece the ship
// is going to shoot at
/////////////////////////////////////////////////
castle_piece_t *
castle_t::shoot()
{
  int val, i, index;

  if ( mValidPieces.size() == 0 )
    return NULL;

  // Randomly pick a valid piece
  val = rand() % mValidPieces.size();

  list<int>::iterator it = mValidPieces.begin();
  for ( i = 0; i < val; ++i )
    ++it;

  index = *it;

  castle_piece_t *lPiece = &( mCastle[ index ] );

  lPiece->mShotAt = true;

  // Remove piece from the list
  mValidPieces.erase( it );

  return lPiece;
}

/////////////////////////////////////////////////
void
castle_t::set_pieces()
{
  vect3_t lCurr = mMin;
  castle_piece_t lPiece;
  int i = 0;

  for ( int x = 0; x < mX; ++x )
  {
    for ( int z = 0; z < mZ; ++z )
    {
      lPiece.mContents = EMPTY;

      lPiece.mPos = lCurr;

      lCurr.z += con_block_size;

      lPiece.mIndex = i++;

      mCastle.push_back( lPiece );
    }
    
    lCurr.z = mMin.z;
    lCurr.x += con_block_size;
  }

  // Figure out where the flag and castle pieces go
  int lCenterX, lCenterZ;
  lCenterX = mFlagBlockX;
  lCenterZ = mFlagBlockZ;

  // Draw the initial castle pieces
  set_piece( lCenterX - 3, lCenterZ - 2 );
  set_piece( lCenterX - 2, lCenterZ - 2 );
  set_piece( lCenterX - 1, lCenterZ - 2 );
  set_piece( lCenterX    , lCenterZ - 2 );
  set_piece( lCenterX + 1, lCenterZ - 2 );
  set_piece( lCenterX + 2, lCenterZ - 2 );
  
  set_piece( lCenterX - 3, lCenterZ - 1 );
  set_piece( lCenterX + 2, lCenterZ - 1 );
  
  set_piece( lCenterX - 3, lCenterZ     );
  set_piece( lCenterX + 2, lCenterZ     );
  
  set_piece( lCenterX - 3, lCenterZ + 2 );
  set_piece( lCenterX - 3, lCenterZ + 1 );
  set_piece( lCenterX + 2, lCenterZ + 1 );
  set_piece( lCenterX + 2, lCenterZ + 2 );

  set_piece( lCenterX - 3, lCenterZ + 3 );
  set_piece( lCenterX - 2, lCenterZ + 3 );
  set_piece( lCenterX - 1, lCenterZ + 3 );
  set_piece( lCenterX    , lCenterZ + 3 );
  set_piece( lCenterX + 1, lCenterZ + 3 );
  set_piece( lCenterX + 2, lCenterZ + 3 );

  // Insert flag - it occupies 4 squares
  if ( piece_ok( lCenterX, lCenterZ ) )
  {
    int lIndex = mZ * lCenterX + lCenterZ;
    mCastle[ lIndex ].mContents = FLAG;
    mCastle[ lIndex ].mDestroyed = false;
    mCastle[ lIndex ].mShotAt = false;
    mCastle[ lIndex ].mIndex = lIndex;
  }
  if ( piece_ok( lCenterX, (lCenterZ + 1) ) )
  {
    int lIndex = mZ * lCenterX + lCenterZ + 1;
    mCastle[ lIndex ].mContents = FLAG;
    mCastle[ lIndex ].mDestroyed = false;
    mCastle[ lIndex ].mShotAt = false;
    mCastle[ lIndex ].mIndex = lIndex;
  }
  if ( piece_ok( (lCenterX - 1), (lCenterZ + 1) ) )
  {
    int lIndex = mZ * (lCenterX - 1) + lCenterZ + 1;
    mCastle[ lIndex ].mContents = FLAG;
    mCastle[ lIndex ].mDestroyed = false;
    mCastle[ lIndex ].mShotAt = false;
    mCastle[ lIndex ].mIndex = lIndex;
  }
  if ( piece_ok( (lCenterX - 1), lCenterZ ) )
  {
    int lIndex = mZ * (lCenterX - 1) + lCenterZ;
    mCastle[ lIndex ].mContents = FLAG;
    mCastle[ lIndex ].mDestroyed = false;
    mCastle[ lIndex ].mShotAt = false;
    mCastle[ lIndex ].mIndex = lIndex;
  }
}

/////////////////////////////////////////////////
void
castle_t::set_rect( const vect3_t &aMin, const vect3_t &aMax )
{
  mMin = aMin;
  mMax = aMax;

  // Figure out how many blocks in each direction
  mX = ( mMax.x - mMin.x ) / con_block_size;
  mZ = ( mMax.z - mMin.z ) / con_block_size;

  // The FlagBlock is one square of the Flag area
  mFlagBlockX = ( mX - ( mMax.x - mFlagPos.x ) / con_block_size );
  mFlagBlockZ = ( mZ - ( mMax.z - mFlagPos.z ) / con_block_size );

  set_pieces();

  if ( mCheckEnclosureArray )
    delete [] mCheckEnclosureArray;

  mCheckEnclosureArray = new bool [ mX * mZ ];
}

/////////////////////////////////////////////////
// This function checks to see if the cannons are
// enclosed at the end of the build round. The ones
// the aren't are removed
/////////////////////////////////////////////////
void 
castle_t::check_cannon_enclosure()
{
  int lX, lZ;

  for ( int i = 0; i < mpCannons->size(); ++i )
  {
    ( *mpCannons )[ i ].get_xz( lX, lZ );

    if ( !check_enclosure( lX, lZ ) )
    {
      ( *mpCannons )[ i ].set_valid( false );
      
      int lIndex = mZ * lX + lZ;

      mCastle[ lIndex ].mContents = EMPTY;
    }
  }
}

/////////////////////////////////////////////////
// This function returns the number of enclosed
// spaces in the castle
/////////////////////////////////////////////////
int 
castle_t::count_enclosed_pieces()
{
  int lCount = 0, lIndex;

  for ( int x = 0; x < mX; ++x )
  {
    for ( int z = 0; z < mZ; ++z )
    {
      lIndex = mZ * x + z;

      if ( mCastle[ lIndex ].mContents != PIECE && check_enclosure( x, z ) )
        ++lCount;
    }
  }

  return lCount;
}

/////////////////////////////////////////////////
// Checks if the flag is fully enclosed by a wall
// returns true if successful
/////////////////////////////////////////////////
bool 
castle_t::check_enclosure()
{
  return check_enclosure( mFlagBlockX, mFlagBlockZ );
}

/////////////////////////////////////////////////
// Checks if a cannon is fully enclosed by a wall
// returns true if successful
/////////////////////////////////////////////////
bool 
castle_t::check_enclosure( int aX, int aZ )
{
  bool lEnclosed = true;

  for ( int i = 0; i < mX * mZ; ++i )
    mCheckEnclosureArray[ i ] = false;

  mCheckEnclosureArray[ mZ * aX + aZ ] = true;

  check_enclosure( aX - 1, aZ - 1, lEnclosed );
  check_enclosure( aX    , aZ - 1, lEnclosed );
  check_enclosure( aX + 1, aZ - 1, lEnclosed );

  check_enclosure( aX - 1, aZ    , lEnclosed );
  check_enclosure( aX + 1, aZ    , lEnclosed );

  check_enclosure( aX - 1, aZ + 1, lEnclosed );
  check_enclosure( aX    , aZ + 1, lEnclosed );
  check_enclosure( aX + 1, aZ + 1, lEnclosed );

  return lEnclosed;
}

/////////////////////////////////////////////////
void
castle_t::check_enclosure( int aX, int aZ, bool &aEnclosed )
{
  // Stop the search if we've already determined that the castle
  // is not enclosed
  if ( !aEnclosed )
    return;

  // If we've reached the map boundaries -> the castle is NOT enclosed
  if ( aX < 0 || aX >= mX || aZ < 0 || aZ >= mZ ) 
  {
    aEnclosed = false;
    return;
  }

  static int lIndex;
  lIndex = mZ * aX + aZ;

  // If we've already checked this piece -> go back
  if ( mCheckEnclosureArray[ lIndex ] )
    return;

  mCheckEnclosureArray[ lIndex ] = true;

  // If we've reached a wall -> go back
  if ( mCastle[ lIndex ].mContents == PIECE )
    return;

  // If we're here -> check all 8 neighbors
  check_enclosure( aX - 1, aZ - 1, aEnclosed );
  check_enclosure( aX    , aZ - 1, aEnclosed );
  check_enclosure( aX + 1, aZ - 1, aEnclosed );

  check_enclosure( aX - 1, aZ    , aEnclosed );
  check_enclosure( aX + 1, aZ    , aEnclosed );

  check_enclosure( aX - 1, aZ + 1, aEnclosed );
  check_enclosure( aX    , aZ + 1, aEnclosed );
  check_enclosure( aX + 1, aZ + 1, aEnclosed );
}

/////////////////////////////////////////////////
// Change the castle pieces from boxes to walls
// and towers
/////////////////////////////////////////////////
void
castle_t::convert_pieces()
{
  int lIndex;
  int lIdx0, lIdx1, lIdx2, lIdx3;
  int lIdx5, lIdx6, lIdx7, lIdx8;
  bool lPiece0, lPiece1, lPiece2, lPiece3;
  bool lPiece5, lPiece6, lPiece7, lPiece8;
  bool lTower;
  bool lTower0, lTower1, lTower2, lTower3;
  bool lTower5, lTower6, lTower7, lTower8;

  // Loop through all of the grid squares
  for ( int x = 0; x < mX; ++x )
  {
    for ( int z = 0; z < mZ; ++z )
    {
      lIndex = mZ * x + z;    // Equivalent to index 4 below

      // These indices are of a 3x3 with 0 being the top left
      // 0 1 2
      // 3 4 5
      // 6 7 8
      lIdx0 = mZ * (x - 1) + (z - 1);
      lIdx1 = mZ *  x      + (z - 1);
      lIdx2 = mZ * (x + 1) + (z - 1);

      lIdx3 = mZ * (x - 1) + z;
      lIdx5 = mZ * (x + 1) + z;

      lIdx6 = mZ * (x - 1) + (z + 1);
      lIdx7 = mZ *  x      + (z + 1);
      lIdx8 = mZ * (x + 1) + (z + 1);

      // If a piece is present figure out what to place there
      // Look at the piece's neighbors, including diagonals
      if ( mCastle[ lIndex ].mContents == PIECE )
      {
        lPiece0 = lPiece1 = lPiece2 = lPiece3 = false;
        lPiece5 = lPiece6 = lPiece7 = lPiece8 = false;
        lTower0 = lTower1 = lTower2 = lTower3 = false;
        lTower5 = lTower6 = lTower7 = lTower8 = false;

        // Are pieces present at each index?
        if ( mCastle[ lIdx0 ].mContents == PIECE )
        {
          lPiece0 = true;
          if ( mCastle[ lIdx0 ].mPieceType == TOWER )
            lTower0 = true;
        }
        if ( mCastle[ lIdx1 ].mContents == PIECE )
        {
          lPiece1 = true;
          if ( mCastle[ lIdx1 ].mPieceType == TOWER )
            lTower1 = true;
        }
        if ( mCastle[ lIdx2 ].mContents == PIECE )
        {
          lPiece2 = true;
          if ( mCastle[ lIdx2 ].mPieceType == TOWER )
            lTower2 = true;
        }
        if ( mCastle[ lIdx3 ].mContents == PIECE )
        {
          lPiece3 = true;
          if ( mCastle[ lIdx3 ].mPieceType == TOWER )
            lTower3 = true;
        }
        if ( mCastle[ lIdx5 ].mContents == PIECE )
        {
          lPiece5 = true;
          if ( mCastle[ lIdx5 ].mPieceType == TOWER )
            lTower5 = true;
        }
        if ( mCastle[ lIdx6 ].mContents == PIECE )
        {
          lPiece6 = true;
          if ( mCastle[ lIdx6 ].mPieceType == TOWER )
            lTower6 = true;
        }
        if ( mCastle[ lIdx7 ].mContents == PIECE )
        {
          lPiece7 = true;
          if ( mCastle[ lIdx7 ].mPieceType == TOWER )
            lTower7 = true;
        }
        if ( mCastle[ lIdx8 ].mContents == PIECE )
        {
          lPiece8 = true;
          if ( mCastle[ lIdx8 ].mPieceType == TOWER )
            lTower8 = true;
        }

        // Are any of the pieces in the 3x3 towers?
        if ( lTower0 || lTower1 || lTower2 || lTower3 ||
             lTower5 || lTower6 || lTower7 || lTower8 )
        {
          lTower = true;
        }
        else
          lTower = false;

        // A switch-case statement might be faster here, but performance doesn't matter
        // because this function is called during a scene switch

        // If there is a piece above and/or below and none to the left or right
        if ( (lPiece1 || lPiece7) && (!lPiece3 && !lPiece5) )
        {
          // Place a wall piece and rotate it 90 degrees so it is vertical
          mCastle[ lIndex ].mPieceType = WALL;
          mCastle[ lIndex ].mPieceRot = QUARTER;
        }

        // If there is a piece to the right and/or to the left and none above or below
        else if ( (lPiece3 || lPiece5) && (!lPiece1 && !lPiece7) )
        {
          // Place a wall piece
          mCastle[ lIndex ].mPieceType = WALL;
          mCastle[ lIndex ].mPieceRot = NO_ROT;
        }

        // If there are pieces to the left, right, above, and none below
        else if ( lPiece3 && lPiece5 && lPiece1 && !lPiece7 )
        {
          // Place a tee
          mCastle[ lIndex ].mPieceType = TEE;
          mCastle[ lIndex ].mPieceRot = NO_ROT;
        }

        // If there are pieces to the left, right, below, and none above
        else if ( lPiece3 && lPiece5 && lPiece7 && !lPiece1 )
        {
          // Place a tee and rotate it 180 degrees
          mCastle[ lIndex ].mPieceType = TEE;
          mCastle[ lIndex ].mPieceRot = HALF;
        }

        // Is it a tower?  This is the only time corner pieces matter
        // 4 possible tower configs where the current piece is the center of an elbow
        // If there are only pieces to the right and below and they aren't towers
        else if ( !lPiece1 && !lPiece3 && lPiece5 && lPiece7 && !lTower )
        {
          // Place a tower
          mCastle[ lIndex ].mPieceType = TOWER;
          mCastle[ lIndex ].mPieceRot = NO_ROT;
        }

        // If there are only pieces to the left and below and they aren't towers
        else if ( !lPiece1 && lPiece3 && !lPiece5 && lPiece7 && !lTower )
        {
          // Place a tower and rotate it 270 degrees
          mCastle[ lIndex ].mPieceType = TOWER;
          mCastle[ lIndex ].mPieceRot = THREE_QUARTERS;
        }

        // If there are only pieces above and to the right and they aren't towers
        else if ( lPiece1 && !lPiece3 && lPiece5 && !lPiece7 && !lTower )
        {
          // Place a tower and rotate it 90 degrees
          mCastle[ lIndex ].mPieceType = TOWER;
          mCastle[ lIndex ].mPieceRot = QUARTER;
        }

        // If there are only pieces above and to the left and they aren't towers
        else if ( lPiece1 && lPiece3 && !lPiece5 && !lPiece7 && !lTower )
        {
          // Place a tower and rotate it 180 degrees
          mCastle[ lIndex ].mPieceType = TOWER;
          mCastle[ lIndex ].mPieceRot = HALF;
        }

        // If there are pieces to the right and below, but not above or to the left
        else if ( !lPiece1 && !lPiece3 && lPiece5 && lPiece7 )
        {
          // Place a corner
          mCastle[ lIndex ].mPieceType = CORNER;
          mCastle[ lIndex ].mPieceRot = NO_ROT;
        }

        // If there are pieces to the left and below, but not above or to the right
        else if ( !lPiece1 && lPiece3 && !lPiece5 && lPiece7 )
        {
          // Place a corner and rotate it 270 degrees
          mCastle[ lIndex ].mPieceType = CORNER;
          mCastle[ lIndex ].mPieceRot = THREE_QUARTERS;
        }

        // If there are pieces above and to the right, but not below and to the left
        else if ( lPiece1 && !lPiece3 && lPiece5 && !lPiece7 )
        {
          // Place a corner and rotate it 90 degrees
          mCastle[ lIndex ].mPieceType = CORNER;
          mCastle[ lIndex ].mPieceRot = QUARTER;
        }

        // If there are pieces above and to the left, but not below and to the right
        else if ( lPiece1 && lPiece3 && !lPiece5 && !lPiece7 )
        {
          // Place a corner and rotate it 180 degrees
          mCastle[ lIndex ].mPieceType = CORNER;
          mCastle[ lIndex ].mPieceRot = HALF;
        }

        // Else place a cross
        else
        {
          mCastle[ lIndex ].mPieceType = CROSS;
          mCastle[ lIndex ].mPieceRot = NO_ROT;
        }
      }
    }
  }
}

/////////////////////////////////////////////////
// Change the castle pieces from walls and towers
// back to boxes
/////////////////////////////////////////////////
void
castle_t::unconvert_pieces()
{
  int lIndex;

  // Loop through all of the grid squares
  for ( int x = 0; x < mX; ++x )
  {
    for ( int z = 0; z < mZ; ++z )
    {
      lIndex = mZ * x + z;    // Equivalent to index 4 below

      mCastle[ lIndex ].mPieceType = BLOCK;
      mCastle[ lIndex ].mPieceRot = NO_ROT;
    }
  }
}

