#ifndef __CASTLE_PIECE_H__
#define __CASTLE_PIECE_H__

#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include <vector>
#include <list>

#include <assert.h>

#include "globals.h"
#include "vect3.h"
#include "model.h"
#include "console.h"

enum { PIECE_I = 0, PIECE_L, PIECE_T, PIECE_S, PIECE_2I, PIECE_SING, 
  PIECE_CRNR, PIECE_S_INV };

enum { GHOST_NAME = 0xffff };

enum { EMPTY = 0, PIECE, CANNON, FLAG, RUBBLE0, RUBBLE1 };

// Model indices
enum { BLOCK_LOW = 0, WALL_LOW, WALL_HIGH, TEE_LOW, TEE_HIGH, 
  CORNER_LOW, CORNER_HIGH, CROSS_LOW, CROSS_HIGH, TOWER_LOW, TOWER_HIGH };

// PieceType indices
enum { BLOCK = 0, WALL, TEE, CORNER, CROSS, TOWER };

enum { NO_ROT = 0, QUARTER, HALF, THREE_QUARTERS };

struct castle_piece_t
{
  int mContents;
  int mPieceType;   // Specifies a wall, tower, etc.
  int mPieceRot;    // Specifies the piece's rotation
  bool mShotAt;     // Is this space being shot at?
  bool mDestroyed;  // It had a piece but got destroyed
  vect3_t mPos;     // World position
  int mX, mZ;       // Block position
  int mIndex;       // Index of this piece in the array
};

struct tetris_piece_t
{
  tetris_piece_t() : mAngle( 0 ) {}

  int mType;
  int mAngle;    // Angle of the piece, increments of 90 degrees
};

struct piece_offset_t
{
  int mX[ 4 ];
  int mZ[ 4 ];
};

// This array holds 4 piece offsets per piece type and per rotation angle
extern piece_offset_t gPieceOffset[ 8 ][ 4 ];


void init_tetris_pieces();

#endif