#include "console.h"

console_t *gCon;

//////////////////////////////////////////////////////////////////////
console_t::console_t( int pNumLines ) : 
	enable( false ), numLines( pNumLines ), 
	cursorCount( 0.0f ), tcinc( 0.0 )
{ 
    lines = new Line [ numLines ];
    clean();

	
  strcpy( mInstantMsg, "" );
  mInstantMsgValid = false;
  mInstantMsgCount = 0.0f;

  mFromFile = false;
}

//////////////////////////////////////////////////////////////////////
void
console_t::init( texture_manager_t *apTextures, font_t *apFont )
{
	assert( apFont );
	mpFont = apFont;

	mpTextures = apTextures;
	conTexID = mpTextures->insert( "water_tile.jpg" );
}

//////////////////////////////////////////////////////////////////////
void
console_t::reset()
{
  cursorPos = 5;
  maxPos = 5;
  strcpy( lines[ numLines - 1 ].data, "CON> " );
}

//////////////////////////////////////////////////////////////////////
void
console_t::clean()
{
  for ( int i = 0; i < numLines; i++ )
  {
    strcpy( lines[ i ].data, "" );
  }
  reset();
}

//////////////////////////////////////////////////////////////////////
void 
console_t::toggle() 
{ 
	if ( enable == false )
	{
		enable = true;
		drawY = 700.0; 
		yinc = 10.0;
	}
	else
	{
		yinc = -10.0;
		enable = true;
	}
}

//////////////////////////////////////////////////////////////////////
void 
console_t::reg_float( const string &name, void *data, bool constant,
		              HANDLER_TYPE handlerType, 
			          void ( *handler )( float ) )
{
  console_entry_t entry;
  entry.type = FT_FLOAT;
  entry.name = name;
  entry.data = data;
  entry.hdType = handlerType;
  entry.handlerf = handler;
  entry.constant = constant;
  entryList.push_back( entry );
}

//////////////////////////////////////////////////////////////////////
void 
console_t::reg_int( const string &name, void *data, bool constant,
		             HANDLER_TYPE handlerType, 
		             void ( *handler )( int ) )
{
  console_entry_t entry;
  entry.type = FT_INT;
  entry.name = name;
  entry.data = data;
  entry.hdType = handlerType;
  entry.handleri = handler;
  entry.constant = constant;
  entryList.push_back( entry );
}

//////////////////////////////////////////////////////////////////////
void 
console_t::reg_bool( const string &name, void *data, bool constant,
		             HANDLER_TYPE handlerType, 
		             void ( *handler )( bool ) )
{
  console_entry_t entry;
  entry.type = FT_BOOL;
  entry.name = name;
  entry.data = data;
  entry.hdType = handlerType;
  entry.handlerb = handler;
  entry.constant = constant;
  entryList.push_back( entry );
}

//////////////////////////////////////////////////////////////////////
void 
console_t::reg_string( const string &name, void *data, bool constant,
		               HANDLER_TYPE handlerType, 
			           void ( *handler )( const char * ) )
{
  console_entry_t entry;
  entry.type = FT_STRING;
  entry.name = name;
  entry.data = data;
  entry.hdType = handlerType;
  entry.handlers = handler;
  entry.constant = constant;
  entryList.push_back( entry );
}

//////////////////////////////////////////////////////////////////////
void 
console_t::reg_cmd( const string &name, void ( *handler )( void ) )
{
  console_entry_t entry;
  entry.type = FT_COMMAND;
  entry.name = name;
  entry.handlercmd = handler;
  entryList.push_back( entry );
}

//////////////////////////////////////////////////////////////////////
void 
console_t::exec_delay_handlers()
{
	list<console_entry_t*>::iterator it;
	for ( it = mGLEntries.begin(); it != mGLEntries.end(); it++ )
	{
		console_entry_t *entry = *it;

		switch ( entry->type ) 
		{
		case FT_FLOAT :    

			entry->handlerf( *( ( float * ) entry->data ) );
			break;

		case FT_INT :
			entry->handleri( *( ( int * ) entry->data ) );
			break;

		case FT_BOOL :
			entry->handlerb( *( ( bool * ) entry->data ) );
			break;

		case FT_STRING :
			entry->handlers( ( char * ) entry->data );
			break;	
		}
	}

	mGLEntries.clear();
}

//////////////////////////////////////////////////////////////////////
bool 
console_t::set( const string &name, const char *valstr )
{
  list<console_entry_t>::iterator it;
  for ( it = entryList.begin(); it != entryList.end(); it++ )
  {
    console_entry_t *entry = &( *it );
    if ( entry->name == name )
    {
      if ( entry->constant )
	  {
		add_msg( "Variable %s is a constant", name.c_str() );
		return false;
	  }

      switch ( entry->type ) 
      {
      case FT_FLOAT :
		{
			float val; 
		    sscanf( valstr, "%f", &val );
		    *( ( float * ) entry->data ) = val;

		    // If this entry has a handler -> call it
		    if ( entry->handlerf )
			{
				if ( entry->hdType == HD_GL && mFromFile )
					mGLEntries.push_back( entry );
				else
					entry->handlerf( val );
			}
		    break;
		}
      case FT_INT :
		{
			int val; 
			sscanf( valstr, "%d", &val );
			*( ( int * ) entry->data ) = val;

			// If this entry has a handler -> call it
			if ( entry->handleri )
			{
				if ( entry->hdType == HD_GL && mFromFile )
					mGLEntries.push_back( entry );
				else
					entry->handleri( val );
			}
			break;
		}
      case FT_BOOL :
		{
			bool val; 
			sscanf( valstr, "%d", &val );
			*( ( bool * ) entry->data ) = val;

		    // If this entry has a handler -> call it
			if ( entry->handlerb )
			{
				if ( entry->hdType == HD_GL && mFromFile )
					mGLEntries.push_back( entry );
				else
					entry->handlerb( val );
			}
			break;
		}
      case FT_STRING :
		{
			strcpy( ( char * ) entry->data, valstr );
     
			// If this entry has a handler -> call it
			if ( entry->handlers )
			{
				if ( entry->hdType == HD_GL && mFromFile )
					mGLEntries.push_back( entry );
				else
					entry->handlers( valstr );
			}
			break;
		}	
      }
      return true;
    }
  }
  return false;
}

//////////////////////////////////////////////////////////////////////
// Toggles boolean console vars
//////////////////////////////////////////////////////////////////////
bool 
console_t::toggle( const string &name )
{
  list<console_entry_t>::iterator it;
  for ( it = entryList.begin(); it != entryList.end(); it++ )
  {
    console_entry_t *entry = &( *it );
    if ( entry->name == name )
    {
      switch ( entry->type ) 
      {
      case FT_BOOL :
        *( ( bool * ) entry->data ) = !( *( ( bool * ) entry->data ) );

		    // If this entry has a handler -> call it
        if ( entry->handlerb )
        {
          entry->handlerb( *( ( bool * ) entry->data ) );
        }
        break;
      }
      return true;
    }
  }
  return false;
}

//////////////////////////////////////////////////////////////////////
bool 
console_t::echo( const string &name )
{
	char msg[ LINE_WIDTH ];
	float val;
	list<console_entry_t>::iterator it;
	for ( it = entryList.begin(); it != entryList.end(); it++ )
	{	
		console_entry_t *entry = &( *it );
		if ( entry->name == name )
		{
			shift();
			switch ( entry->type )
			{
			case FT_BOOL :
				val = *( ( bool * ) entry->data );
				sprintf( msg, "%s = %s", name.c_str(), 
					( val ) ? "true" : "false" );
				break;
			case FT_INT :
				val = *( ( int * ) entry->data );
				sprintf( msg, "%s = %d", name.c_str(), ( int ) val );
				break;
			case FT_FLOAT :
				val = *( ( float * ) entry->data );
				sprintf( msg, "%s = %1.3f", name.c_str(), val );
				break;
			case FT_STRING :
				sprintf( msg, "%s = %s", name.c_str(), ( char * ) entry->data );
				break;
			}
      
			add_msg( msg );

			return true;
		}
	}
	return false;
}

//////////////////////////////////////////////////////////////////////
bool 
console_t::process_cmd( const string &command )
{
	list<console_entry_t>::iterator it;
	for ( it = entryList.begin(); it != entryList.end(); it++ )
	{	
		console_entry_t *entry = &( *it );
		if ( entry->name == command )
		{
			shift();
			if ( entry->type == FT_COMMAND )
			{
				entry->handlercmd();
				return true;
			}
			else
				return false;
		}
	}
	return false;
}

//////////////////////////////////////////////////////////////////////
void
console_t::shift()
{
  for ( int i = 0; i < numLines - 1; i++ )
    strcpy( lines[ i ].data, lines[ i + 1 ].data );
}

//////////////////////////////////////////////////////////////////////
// This function handles keystrokes
//////////////////////////////////////////////////////////////////////
void 
console_t::add_char( unsigned short key )
{
	if ( !enable )
		return;

	if ( key == '`' )
	{
		enable = false;
		return;
	}

	if ( isprint( key ) )
	{
		if ( cursorPos == LINE_WIDTH - 1 )
			return;
		
		lines[ numLines - 1 ].data[ cursorPos++ ] = key;
		
		if ( cursorPos > maxPos )
			maxPos = cursorPos;
		
		lines[ numLines - 1 ].data[ maxPos ] = '\0';
	}
}

//////////////////////////////////////////////////////////////////////
// This function handles keystrokes
//////////////////////////////////////////////////////////////////////
void 
console_t::add( unsigned short key )
{
	if ( !enable )
		return;

	switch ( key )
	{
	case VK_BACK : 
		if ( cursorPos > 5 )
		{
			lines[ numLines - 1 ].data[ --cursorPos ] = ' ';
			for ( int i = cursorPos + 1; i < maxPos; i++ )
				lines[ numLines - 1 ].data[ i ] = lines[ numLines - 1 ].data[ i + 1 ];
			maxPos--;
		}
		break;
	case VK_DELETE : 
		if ( maxPos > cursorPos )
		{
			for ( int i = cursorPos; i < maxPos; i++ )
				lines[ numLines - 1 ].data[ i ] = lines[ numLines - 1 ].data[ i + 1 ];
			maxPos--;
		}
		break;
	case VK_LEFT :
		if ( cursorPos > 5 )
		{
			--cursorPos;
		}
		break;
	case VK_RIGHT :
		if ( cursorPos < maxPos )
		{
			++cursorPos;
		}
		break;
	case VK_RETURN :
		parse_line( lines[ numLines - 1 ].data );
		break;
	}
}

//////////////////////////////////////////////////////////////////////
void
console_t::parse_line( const char *pLine, bool fromFile )
{
  char errorLine[ LINE_WIDTH ];
  char *tok, name[ LINE_WIDTH ], line[ LINE_WIDTH ];
  strcpy( line, pLine );

  if ( fromFile )
  {
    tok = strtok( line, "\t\n " ); // command
  }
  else
  {
    tok = strtok( line, "\t\n " ); // prompt
    tok = strtok( NULL, "\t\n " ); // command
  }
  
  if ( !tok )
  {  
    shift();
    reset();
    return;
  }

  if ( tok[ 0 ] == '#' )
  {
	return;
  }
  else if ( strcmp( tok, "set" ) == 0 )
  {
    tok = strtok( NULL, "\t\n " ); // var name
    if ( !tok )
    {
      shift();
      add_msg( "-- Syntax: set <var> <value>" );
      return;
    }
    strcpy( name, tok );
    tok = strtok( NULL, "\t\n " ); // var value
    if ( !tok )
    {
      shift();
      add_msg( "-- Syntax: set <var> <value>" );
      return;
    }
    if ( !set( name, tok ) )
    {
      shift();
      sprintf( errorLine, "-- Var %s not found!", name );
      add_msg( errorLine );
      return;
    }
  }
  else if ( strcmp( tok, "echo" ) == 0 )
  {
    tok = strtok( NULL, "\t\n " ); // var name
    strcpy( name, tok );
    if ( !tok )
    {
      shift();
      add_msg( "-- Syntax: set <var> <value>" );
      return;
    }
    if ( echo( name ) )
    {
      return;
    }
    else
    {
      shift();
      sprintf( errorLine, "-- Var %s not found!", name );
      add_msg( errorLine );
      return;
    }
  }
  else
  {
    if ( !process_cmd( tok ) )
	  {
      shift();
      sprintf( errorLine, "-- Invalid command '%s'!", tok );
      add_msg( errorLine );
	    return;
	  }
  }
  
  if ( !fromFile )
	  shift();

  reset();
}


//////////////////////////////////////////////////////////////////////
// Used to save options set through the menus to a file
//////////////////////////////////////////////////////////////////////
void
console_t::save_options( const char *fileName )
{
  FILE *file = fopen( fileName, "w" );
  if ( !file )
  {
    add_msg( "-- Cannot open file %s for writing", fileName );
    return;
  }

	list<console_entry_t>::iterator it;
	for ( it = entryList.begin(); it != entryList.end(); it++ )
	{	
		console_entry_t *entry = &( *it );

    if ( entry->name == "fullscreen" )
      fprintf( file, "set %s %d\n", entry->name.c_str(), *( ( bool * ) entry->data ) );

    if ( entry->name == "difficulty" )
      fprintf( file, "set %s %d\n", entry->name.c_str(), *( ( bool * ) entry->data ) );

    if ( entry->name == "shadows" )
      fprintf( file, "set %s %d\n", entry->name.c_str(), *( ( bool * ) entry->data ) );

    if ( entry->name == "sndfx_vol" )
      fprintf( file, "set %s %d\n", entry->name.c_str(), *( ( int * ) entry->data ) );

    if ( entry->name == "res" )
      fprintf( file, "set %s %s\n", entry->name.c_str(), ( entry->data ) );

    if ( entry->name == "filtering" )
      fprintf( file, "set %s %d\n", entry->name.c_str(), *( ( int * ) entry->data ) );

    if ( entry->name == "geometry" )
      fprintf( file, "set %s %d\n", entry->name.c_str(), *( ( int * ) entry->data ) );
	}

  fclose( file );
}


//////////////////////////////////////////////////////////////////////
void
console_t::save_file( const char *fileName )
{
  FILE *file = fopen( fileName, "w" );
  if ( !file )
  {
    add_msg( "-- Cannot open file %s for writing", fileName );
    return;
  }

	list<console_entry_t>::iterator it;
	for ( it = entryList.begin(); it != entryList.end(); it++ )
	{	
		console_entry_t *entry = &( *it );
		switch ( entry->type )
		{
		case FT_BOOL :
      fprintf( file, "set %s %d\n", entry->name.c_str(), *( ( bool * ) entry->data ) );
			break;
		case FT_INT :
      fprintf( file, "set %s %d\n", entry->name.c_str(), *( ( int * ) entry->data ) );
			break;
		case FT_FLOAT :
      fprintf( file, "set %s %1.3f\n", entry->name.c_str(), *( ( float * ) entry->data ) );
			break;
		case FT_STRING :
      fprintf( file, "set %s %s\n", entry->name.c_str(), ( entry->data ) );
			break;
		}
	}

  fclose( file );
}

//////////////////////////////////////////////////////////////////////
void
console_t::exec_file( const char *fileName )
{
  char errorLine[ LINE_WIDTH ];
  FILE *file = fopen( fileName, "r" );
  if ( !file )
  {
    sprintf( errorLine, "-- Cannot open file %s", fileName );
    add_msg( errorLine );
    return;
  }
  
  mGLEntries.clear();

  mFromFile = true;

  sprintf( errorLine, "-- Parsing script '%s'", fileName );
  add_msg( errorLine );  

  char line[ 500 ];
  while ( fgets( line, 500, file ) != NULL ) // TOP
  {
    parse_line( line, true );
  }

  mFromFile = false;

  fclose( file );
}


//////////////////////////////////////////////////////////////////////
void 
console_t::draw()
{	
	if ( mInstantMsgValid )
	{
		mInstantMsgCount -= gFactTick;
		if ( mInstantMsgCount < 0 )
			mInstantMsgValid = false;

		mpFont->print( 90, 60, 1, 0, 0, mInstantMsg );
	}

	if ( !enable )
		return;

	drawY -= yinc * gFactTick;
	if ( drawY < 480 )
		drawY = 480;

	if ( drawY > 700 )
		enable = false;

	// Animate the texture coords of the console background
	tcinc += ( 0.0005f * gFactTick );
	glMatrixMode( GL_TEXTURE );
	glTranslatef( tcinc, tcinc, 0 );
	glMatrixMode( GL_MODELVIEW );

	glDisable( GL_BLEND );
	glBindTexture( GL_TEXTURE_2D, conTexID );
	glBegin( GL_QUADS );
  
	glTexCoord2f( 1, 0 );
	glVertex2f( 640, drawY );
  
	glTexCoord2f( 1, 1 );
	glVertex2f( 640, drawY - ( numLines * 20 ) );
  
	glTexCoord2f( 0, 1 );
	glVertex2f( 0, drawY - ( numLines * 20 ) );

	glTexCoord2f( 0, 0 );
	glVertex2f( 0, drawY );

	glEnd();
	glEnable( GL_BLEND );

	glMatrixMode( GL_TEXTURE );
	glLoadIdentity();
	glMatrixMode( GL_MODELVIEW );

	mpFont->setDefaultDim();
	float cony = drawY - 20;
	for ( int i = 0; i < numLines; i++ )
	{		
		mpFont->print( 0, cony, lines[ i ].data );
		cony -= 20;
	}
  
	// Print Version number
	mpFont->setDim( 8, 8, 5 );
	mpFont->print( 600, cony + 20, 1, 1, 0, "VER 1.0" );

	mpFont->setDefaultDim();

	cursorCount += gFactTick;

	if ( cursorCount < 50.0f )
		mpFont->print( mpFont->getSpace() * cursorPos, cony + 20.0, "_" );
	else if ( cursorCount >= 100.0f )
		cursorCount = 0.0f;

  glColor4f( 1, 1, 1, 1 );
}

//////////////////////////////////////////////////////////////////////
void 
console_t::add_msg( const char *msg, ... )
{  
	va_list		ap;
	if ( msg == NULL )
		return;

	va_start( ap, msg );
	vsprintf( lines[ numLines - 1 ].data, msg, ap );
	va_end( ap );

	gLog->add_msg( lines[ numLines - 1 ].data );

	shift();
	reset();
}

//////////////////////////////////////////////////////////////////////
void 
console_t::add_instant_msg( const char *msg, ... )
{  
	va_list		ap;
	if ( msg == NULL )
		return;

	va_start( ap, msg );
	vsprintf( mInstantMsg, msg, ap );
	va_end( ap );

	mInstantMsgValid = true;
	mInstantMsgCount = 200.0f;

	gLog->add_msg( mInstantMsg );
}
