
#include "fmath.h"

float frand()
{
	return ( ( float ) ( rand() & 32767 ) ) * ( 1.0f / 32767.0f );
}

float crand()
{
	return ( ( float ) ( rand() & 32767 ) ) * ( 2.0f / 32767.0f ) - 1.0f;
}	

int irand( int aMin, int aMax )
{		
	return ( rand() % ( aMax - aMin + 1 ) ) + aMin;
}

static unsigned int fast_sqrt_table[0x10000];  // declare table of square roots 

float_t bias = {((23 + 127) << 23) + (1 << 22)};

int FloatToInt( float f ) 
{
	static float_t n;

	n.f = f;

	n.f += bias.f;
	n.i -= bias.i;

	return n.i;
}

void  
build_sqrt_table()
{
  unsigned int i;
  float_t s;
  
  for (i = 0; i <= 0x7FFF; i++)
  {
    
    // Build a float with the bit pattern i as mantissa
    //  and an exponent of 0, stored as 127
    
    s.i = (i << 8) | (0x7F << 23);
    s.f = (float)sqrt(s.f);
    
    // Take the square root then strip the first 7 bits of
    //  the mantissa into the table
    
    fast_sqrt_table[i + 0x8000] = (s.i & 0x7FFFFF);
    
    // Repeat the process, this time with an exponent of 1, 
    //  stored as 128
    
    s.i = (i << 8) | (0x80 << 23);
    s.f = (float)sqrt(s.f);
    
    fast_sqrt_table[i] = (s.i & 0x7FFFFF);
  }
}

float fsqrt( float n )
{
  static float_t val;
  val.f = n;
  //if (FP_BITS(n) == 0)
  //  return 0.0;                 // check for square root of 0
  
  val.i = fast_sqrt_table[(val.i >> 8) & 0xFFFF] | ((((val.i - 0x3F800000) >> 1) + 0x3F800000) & 0x7F800000);
  
  //gLog << "val = " << n << " SQRT = " << val.f << ' ' << sqrt( n ) << endl;
  return val.f;
}

/**************************************************************
    Linear lookup for sine and cosine

    This section implements a lookup table version of sin
    and cos, both using the same table.
 **************************************************************/

static float    sintable[SINTABLESIZE];

float two = 2.0f;

/*
    Initialize the lookup table.
*/

void 
triginit()
{
    int     i;
    double  theta;

    /* Integer range 0 to SINTABLESIZE is converted to floating
       point 0 to 2 pi range (radians for a full circle). */
    for (i = 0; i < SINTABLESIZE; ++i)
    {
        theta = (double)i/TWOPISCALE;
        sintable[i] = (float)sin(theta);
    }
}

/*
    fsin - sin calculation with lookup table.
*/

float    
fsin( float theta )
{
    static unsigned int    i;
	static float_t  fi;

    fi.f = theta * TWOPISCALE + bias.f; /* Scale to table index range and add conversion bias. */
    i = fi.i & (SINTABLESIZE - 1);         /* Mask off lower bits, assuming SINTABLESIZE is power of 2. */

    /* Return sine from table. */
    return sintable[i];
}

/*
    fcos - cos calculation with lookup table.
*/

float    
fcos( float theta )
{
    static unsigned int i;
    static float_t fi;

    fi.f = theta * TWOPISCALE + bias.f;             /* Scale to table index range and add conversion bias. */

    /* Note how adding SINTABLESIZE/4 lets us compute cos instead of sin,
       since cos(theta) = sin(theta + Pi/4). */
    i = (fi.i + (SINTABLESIZE/4)) & (SINTABLESIZE - 1);   /* Mask off lower bits, assuming SINTABLESIZE is power of 2. */

    return sintable[i];
}
