#ifndef __FMATH__
#define __FMATH__

#include <stdlib.h>
#include <math.h>
#include <wtypes.h>
#include <time.h>
#include <stdio.h>

float frand();

float crand();

int irand( int aMin, int aMax );	

typedef union
{
    int     i;          // as integer
    float   f;          // as float
    struct              // as bit fields
    {
        unsigned int    mantissa:23;
        unsigned int    exponent:8;
        unsigned int    sign:1;
    }
    bits;
}
float_t;

__forceinline void FloatToInt(int *int_pointer, float f) 
{
  __asm  fld  f
  __asm  mov  edx,int_pointer
  __asm  FRNDINT
  __asm  fistp dword ptr [edx];

}

inline
float fastabs( float val )
{
	static float_t ftemp;
	ftemp.f = val;

	ftemp.i &= 0x7fffffff;
	return ftemp.f;
}

#define FP_BITS(fp) (*(DWORD *)&(fp))
#define FP_ABS_BITS(fp) (FP_BITS(fp)&0x7FFFFFFF)
#define FP_SIGN_BIT(fp) (FP_BITS(fp)&0x80000000)
#define FP_ONE_BITS 0x3F800000 

// r = 1/p
#define FP_INV(r,p)                                                          \
{                                                                            \
    int _i = 2 * FP_ONE_BITS - *(int *)&(p);                                 \
    r = *(float *)&_i;                                                       \
    r = r * (2.0f - (p) * r);                                                \
}

/////////////////////////////////////////////////
// The following comes from Vincent Van Eeckhout
// Thanks for sending us the code!
// It's the same thing in assembly but without this C-needed line:
//    r = *(float *)&_i;

extern float   two;

#define FP_INV2(r,p)                     \
{                                        \
    __asm { mov     eax,0x7F000000    }; \
    __asm { sub     eax,dword ptr [p] }; \
    __asm { mov     dword ptr [r],eax }; \
    __asm { fld     dword ptr [p]     }; \
    __asm { fmul    dword ptr [r]     }; \
    __asm { fsubr   [two]             }; \
    __asm { fmul    dword ptr [r]     }; \
    __asm { fstp    dword ptr [r]     }; \
}

void build_sqrt_table();
float fsqrt(float n);

/**************************************************************
    Linear lookup for sine and cosine

    This section implements a lookup table version of sin
    and cos, both using the same table.
 **************************************************************/

#define SINTABLESIZE    256
#define TWOPISCALE      ((float)SINTABLESIZE / (2.0f * 3.14159265f))

void triginit();

float fsin( float theta );
float fcos( float theta );

#pragma warning (push)
#pragma warning (disable : 4035)	// disable no return value warning

__forceinline  DWORD GetPentiumCounter()
{
	__asm
	{
		xor   eax,eax	          // VC won't realize that eax is modified w/out this
								  //   instruction to modify the val.
								  //   Problem shows up in release mode builds
		_emit 0x0F		          // Pentium high-freq counter to edx;eax
		_emit 0x31		          // only care about low 32 bits in eax

		xor   edx,edx	          // so VC gets that edx is modified
	}
}

#pragma warning (pop)

#endif