#include "font.h"

//////////////////////////////////////////////////////////////
void
font_t::build( texture_manager_t *apTextures )
{
	// Build font texture coordinate LUTS 
	mFontID = apTextures->insert( "Font.bmp" );

	for ( int i = 0; i < 256; i++ )
	{
	    mXArr[ i ] = float( i % 16 ) / 16.0f;
	    mYArr[ i ] = 1.0 - float( i / 16 ) / 16.0f;    
	}
}

//////////////////////////////////////////////////////////////
void
font_t::pre() const
{
  glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
  glFrontFace( GL_CW );

  /*
  glActiveTextureARB( GL_TEXTURE1_ARB );
  glDisable( GL_TEXTURE_2D );
  glClientActiveTextureARB( GL_TEXTURE1_ARB );
  glDisableClientState( GL_TEXTURE_COORD_ARRAY );
*/
  glActiveTextureARB( GL_TEXTURE0_ARB );
  glEnable( GL_TEXTURE_2D );
  glClientActiveTextureARB( GL_TEXTURE0_ARB );
  glDisableClientState( GL_TEXTURE_COORD_ARRAY );

  glBindTexture( GL_TEXTURE_2D, mFontID );
  glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

  glDisable( GL_DEPTH_TEST );
  glEnable( GL_BLEND );
  glBlendFunc( GL_ONE, GL_ONE );
  
  glMatrixMode( GL_PROJECTION );
  glPushMatrix();
  glLoadIdentity();
  glOrtho( 0, 640, 0, 480, -100, 100 );
  glMatrixMode( GL_MODELVIEW );
  glPushMatrix();
  glLoadIdentity(); 
}

//////////////////////////////////////////////////////////////
void 
font_t::post() const
{
  //glActiveTextureARB( GL_TEXTURE1_ARB );
  //glEnable( GL_TEXTURE_2D );
  //glClientActiveTextureARB( GL_TEXTURE1_ARB );
  //glEnableClientState( GL_TEXTURE_COORD_ARRAY );

  //glActiveTextureARB( GL_TEXTURE0_ARB );
  //glEnable( GL_TEXTURE_2D );
  //glClientActiveTextureARB( GL_TEXTURE0_ARB );
  //glEnableClientState( GL_TEXTURE_COORD_ARRAY );

  //glEnableClientState( GL_VERTEX_ARRAY );

  glActiveTextureARB( GL_TEXTURE0_ARB );
  glDisable( GL_TEXTURE_2D );
  //glClientActiveTextureARB( GL_TEXTURE0_ARB );
  //glEnableClientState( GL_TEXTURE_COORD_ARRAY );

  glFrontFace( GL_CCW );
  glDisable( GL_BLEND );
  glEnable( GL_DEPTH_TEST );

  glMatrixMode( GL_PROJECTION );
  glPopMatrix();
  glMatrixMode( GL_MODELVIEW );
  glPopMatrix();

  glPolygonMode( GL_FRONT_AND_BACK, gWireFrame );
}

//////////////////////////////////////////////////////////////
void 
font_t::print( float aX, float aY, const char *aStr ) const
{
  const int lLen = strlen( aStr );
  const float lYPlusHeight = aY + mHeight;
  
  glColor3f( 1.0, 1.0, 1.0 );
  glBindTexture( GL_TEXTURE_2D, mFontID );

  glBegin( GL_QUADS );
  for ( int i = 0; i < lLen; i++ )
  {
    int val = aStr[ i ] - 32;

    // Get texture coordinates from prebuilt LUTS
    float cx = mXArr[ val ];
    float cy = mYArr[ val ];
    
    glTexCoord2f( cx, cy ); 
    glVertex2f( aX, lYPlusHeight );
    
    glTexCoord2f( cx + 0.0625, cy ); 
    glVertex2f( aX + mWidth, lYPlusHeight );

    glTexCoord2f( cx + 0.0625, cy - 0.0625 ); 
    glVertex2f( aX + mWidth, aY );

    glTexCoord2f( cx, cy - 0.0625 ); 
    glVertex2f( aX, aY );
    
    aX += mSpace;
  }
  glEnd();
}

//////////////////////////////////////////////////////////////
void 
font_t::print( float aX, float aY, float aR, float aG, float aB,
			   const char *aStr ) const
{
  const int lLen = strlen( aStr );
  const float lYPlusHeight = aY + mHeight;
  
  glColor3f( aR, aG, aB );
  glBindTexture( GL_TEXTURE_2D, mFontID );

  glBegin( GL_QUADS );
  for ( int i = 0; i < lLen; i++ )
  {
    int val = aStr[ i ] - 32;

    // Get texture coordinates from prebuilt LUTS
    float cx = mXArr[ val ];
    float cy = mYArr[ val ];
    
    glTexCoord2f( cx, cy ); 
    glVertex2f( aX, lYPlusHeight );
    
    glTexCoord2f( cx + 0.0625, cy ); 
    glVertex2f( aX + mWidth, lYPlusHeight );

    glTexCoord2f( cx + 0.0625, cy - 0.0625 ); 
    glVertex2f( aX + mWidth, aY );

    glTexCoord2f( cx, cy - 0.0625 ); 
    glVertex2f( aX, aY );
    
    aX += mSpace;
  }
  glEnd();
}
