#include "game.h"

//////////////////////////////////////////////////////////
game_t::game_t() 
{
	mpTextures = NULL;
	mpSounds = NULL;

  mStage = PLACE_CANNON;

  mNumShipProjs = 3;      // 1 for each ship
  mNumPlayerProjs = 1;
  mBuildRound = 0;
  mBattleRound = 0;

  mScore = 0;
  char lName[] = "";
  strcpy( mName, lName );

	mQuadric = gluNewQuadric();
	gluQuadricDrawStyle( mQuadric, GLU_LINE );

  mShipModel = new shadow_model_node_t;
  mShip2Model = new shadow_model_node_t;

  // Pre-allocate PCOUNT water particle systems
  int i;
  mWaterParticles = new particle_sys_t [ PCOUNT ];
	vect3_t vel;
	vel.x = 0.04f; vel.y = 0.45f; vel.z = 0.04f;
	vect3_t velDelta;
	velDelta.x = 0.03f; velDelta.y = 0.03f; velDelta.z = 0.03f;
  for ( i = 0; i < PCOUNT; ++i )
  {
    mWaterParticles[ i ].init( 40, 400, 20, gParticleTexture, 2, 0.03f, 2.2f,
		                           0.006f, 0.004f, 1, false, vel, velDelta, false,
							                 GL_ONE, GL_ONE );
  }

  // Pre-allocate PCOUNT wake particle systems
  mWakeParticles = new particle_sys_t [ PCOUNT ];
	vel.x = 0.01f; vel.y = 0.05f; vel.z = 0.01f;
	velDelta.x = 0.15f; velDelta.y = 0.02f; velDelta.z = 0.15f;
  for ( i = 0; i < PCOUNT; ++i )
  {
    mWakeParticles[ i ].init( 150,        // StartNum
                              400,        // Energy
                              20,         // EnergyDelta
                              gParticleTexture,   // Texid
                              4,          // SizeStart
                              0.05f,      // SizeInc
                              10.0f,      // SizeMax
                              0.002f,     // Gravity
                              0.003f,     // Fade
                              0.6f,       // Color
                              true,       // Loop
                              vel,        // Velocity
                              velDelta,   // Velocity Delta
                              true,       // Update
							                GL_ONE,     // Src function
                              GL_ONE );   // Dst function
  }

  // Pre-allocate PCOUNT smoke particle systems
  mSmokeParticles = new particle_sys_t [ PCOUNT ];
	vel.x = 0.01f; vel.y = 0.2f; vel.z = 0.01f;
	velDelta.x = 0.001f; velDelta.y = 0.01f; velDelta.z = 0.001f;
  for ( i = 0; i < PCOUNT; ++i )
  {
    mSmokeParticles[ i ].init( 30, 800, 40, gParticleTexture, 16, 0.2f, 50.0f,
		                           0, 0.003f, 1, true, vel, velDelta, true,
							                 GL_ZERO, GL_ONE_MINUS_SRC_COLOR );
  }

  // Pre-allocate PCOUNT cannon smoke particle systems
  mCannonParticles = new particle_sys_t [ PCOUNT ];
	vel.x = 0.01f; vel.y = 0.06f; vel.z = 0.01f;
	velDelta.x = 0.1f; velDelta.y = 0.03f; velDelta.z = 0.1f;
  for ( i = 0; i < PCOUNT; ++i )
  {
    mCannonParticles[ i ].init( 20, 300, 20, gParticleTexture, 6, 0.26f, 14,
		                            0, 0.008f, 1, false, vel, velDelta, false,
							                  GL_ONE, GL_ONE );
  }

  // Pre-allocate PCOUNT explosion particle systems
  mExplosionParticles = new particle_sys_t [ PCOUNT ];
	vel.x = 0.04f; vel.y = 0.45f; vel.z = 0.04f;
	velDelta.x = 0.03f; velDelta.y = 0.03f; velDelta.z = 0.03f;
  for ( i = 0; i < PCOUNT; ++i )
  {
    mExplosionParticles[ i ].init( 20, 400, 20, gParticleTexture, 
                                   16, 0.2f, 22.0f,
		                               0.005f, 0.007f, 1, false, vel, velDelta, false,
							                     GL_ZERO, GL_ONE_MINUS_SRC_COLOR );
  }

  mCastle.set_cannons( &mCannons );

	clear();

  init_tetris_pieces();
}

//////////////////////////////////////////////////////////
game_t::~game_t()
{
  gDebug->add_msg( "~game_t() : start" );

  clear();

  if ( mProjectiles )
    delete [] mProjectiles;

  if ( mProjectiles2 )
    delete [] mProjectiles2;

  if ( mWaterParticles )
    delete [] mWaterParticles;

  if ( mWakeParticles )
    delete [] mWakeParticles;

  if ( mSmokeParticles )
    delete [] mSmokeParticles;

  if ( mCannonParticles )
    delete [] mCannonParticles;  

  if ( mExplosionParticles )
    delete [] mExplosionParticles;

  gDebug->add_msg( "~game_t() : end" );
}

//////////////////////////////////////////////////////////
void
game_t::clear()
{
  int i;

	mValid = false;
  mPause = false;

	mShips.clear();
	mSegments.clear();
	mShipPoly.clear();
	mModelManager.clear();
  mCastle.reset();
  mCannons.clear();

  mStage = PLACE_CANNON;

  for ( i = 0; i < mNumShipProjs; ++i )
  {
    if ( mProjectiles )
      mProjectiles[ i ].set_valid( false );
  }
  for ( i = 0; i < mNumPlayerProjs; ++i )
  {
    if ( mProjectiles2 )
      mProjectiles2[ i ].set_valid( false );
  }

  // Set the number of seconds for the pre-battle phase
  mSeconds = con_warmup_time;
  mSecondCount = 0;

  // Disable particle systems
  disable_particles();
}

//////////////////////////////////////////////////////////
// Restart current game: Don't reload anything, just reset
// objects to their original states
//////////////////////////////////////////////////////////
void
game_t::restart()
{
  int i;

  con_difficulty = con_difficulty_next;

  
  mStage = PLACE_CANNON;

  for ( i = 0; i < mNumShipProjs; ++i )
  {
    if ( mProjectiles )
      mProjectiles[ i ].set_valid( false );
  }
  for ( i = 0; i < mNumPlayerProjs; ++i )
  {
    if ( mProjectiles2 ) 
      mProjectiles2[ i ].set_valid( false );
  }

  mCannons.clear();

  // Set the number of seconds for CANNON_PLACE phase
  mSeconds = con_cannon_time;
  mSecondCount = 0;
  mBuildRound = 0;
  mBattleRound = 0;

  // Disable particle systems
  disable_particles();

  // Reset ships
  if ( con_geometry == 2 )
  {
    mShipModel = &mShipModelHigh;
    mShip2Model = &mShip2ModelHigh;
  }
  else
  {
    mShipModel = &mShipModelLow;
    mShip2Model = &mShip2ModelLow;
  }

  mNumValidShips = con_num_start_ships;
  if ( mNumValidShips > mShips.size() )
    mNumValidShips = mShips.size();

	for ( i = 0; i < mNumValidShips; ++i )
	{
    mShips[ i ].set_vel( get_ship_velocity() );
		mShips[ i ].reset();

    // This is necessary because the ship type can be changed during a previous game
    if ( mOrigShipType[ i ] == 2 )
    {
      mShips[ i ].set_shiptype( 2 );
      mShips[ i ].set_model( mShip2Model );
    }
    else
    {
      mShips[ i ].set_shiptype( 1 );
      mShips[ i ].set_model( mShipModel );
    }
  }

  mCastle.reset();

  mScore = 0;
  char lName[] = "";
  strcpy( mName, lName );

  // Set view pos for cannon-place round
  mViewerPos.x = 500; 
  mViewerPos.y = 1000;
  mViewerPos.z = 500;
  mLookAtPos = mCastle.getFlagPos();

  mViewDir = mLookAtPos - mViewerPos;
  mViewDir.normalize();

  mZoom = 0;
  mViewAngle = 0;
  
  // Clear the high score list states
  gpScoreList->clear_name_entered();

  // Make sure we are pointing to the correct list
  gpScoreList->restart();

  // Reset the player name text box
  gpNameTextBox->reset();
}

//////////////////////////////////////////////////////////
void
game_t::set_sound_manager( sound_manager_t *apSounds )
{
	assert( apSounds );
	mpSounds = apSounds;

  mCastle.set_sound_manager( apSounds );
}

//////////////////////////////////////////////////////////
void
game_t::set_texture_manager( texture_manager_t *apTextures )
{
	assert( apTextures );
	mpTextures = apTextures;

  mShaderManager.set_texture_manager( mpTextures );
}
  
//////////////////////////////////////////////////////////
void 
game_t::zoom_in() 
{ 
  if ( mZoom < 600.0f )
  {
    mViewerPos += mViewDir * 20.0f; 
    mZoom += 20.0f;
  }
}
  
//////////////////////////////////////////////////////////
void 
game_t::zoom_out() 
{ 
  if ( mZoom > -400.0f )
  {
    mViewerPos -= mViewDir * 20.0f; 
    mZoom -= 20.0f;
  }
}

/////////////////////////////////////////////////////////
void 
game_t::rotate_left()
{
  mViewAngle += 2;
  if ( mViewAngle > 360 )
    mViewAngle -= 360;
}

/////////////////////////////////////////////////////////
void 
game_t::rotate_right()
{
  mViewAngle -= 2;
  if ( mViewAngle < 0 )
    mViewAngle += 360;
}

//////////////////////////////////////////////////////////
#define case_enum(a) case(a): if ( con_debug ) gCon->add_instant_msg( #a );

void
game_t::update()
{
  if ( mPause )
    return;

  switch ( mStage )
  {
  case_enum( PLACE_CANNON_BONUS )
    update_placecannon();
  break;

  case_enum( PLACE_CANNON )
    update_placecannon();
    break;

  case_enum( PRE_BATTLE )
    update_prebattle();
    break;

  case_enum( BATTLE )
    update_battle();
    break;

  case_enum( POST_BATTLE )
    update_postbattle();
    break;

  case_enum( PRE_BUILD )
    update_prebuild();
    break;

  case_enum( BUILD )
    update_build();
    break;

  case_enum( WAIT_TO_RESTART )
    mRestartDelay -= gFactTick;
    break;
  }

	static const float lUp[] = { 0, 1, 0 };
	mpSounds->set_listener( mViewerPos.mVect, lUp );
}

//////////////////////////////////////////////////////////
// Update function for when game is in PLACE_CANNON mode
//////////////////////////////////////////////////////////
void
game_t::update_placecannon()
{
  handle_mouse_placecannon( false );

  mSecondCount += gFactTick;
  if ( mSecondCount >= 100.0f )
  {
    mSecondCount = 0;
    --mSeconds;

    // Are we ready for battle?
    if ( mSeconds < 0 )
    {
      // Set the number of seconds for the pre-battle round
      mSeconds = con_warmup_time;

      // Set view pos for battle
      mViewerPos.x = 700;
      mViewerPos.y = 1200;
      mViewerPos.z = 1300;

      mLookAtPos.x = 600;
      mLookAtPos.y = 0;
      mLookAtPos.z = 600;

      mViewDir = mLookAtPos - mViewerPos;
      mViewDir.normalize();
      mZoom = 0;
      mViewAngle = 0;

      //Renable the cursor
      gCursor.enable();

      mCastle.disable_ghost();
      mCastle.disable_draw_cannon();

      // Convert the castle before the pre_battle round
      if ( con_geometry != 0 )
        mCastle.convert_pieces();

      // We placed the cannon > now let's fight
      mStage = PRE_BATTLE;

    }
  }
}

//////////////////////////////////////////////////////////
// Update function for when game is in PRE_BATTLE mode
//////////////////////////////////////////////////////////
void
game_t::update_prebattle()
{
  mSecondCount += gFactTick;
  if ( mSecondCount >= 100.0f )
  {
    mSecondCount = 0;
    --mSeconds;

    // Are we ready for battle?
    if ( mSeconds < 0 )
    {
      mStage = BATTLE;

      mProjectiles = new projectile_t [ mNumShipProjs ];

      for ( int i = 0; i < mNumShipProjs; ++i )
      {
        mProjectiles[ i ].set_model( &mCannonBallModel );
        mProjectiles[ i ].set_valid( false );
      }
      
      mNumPlayerProjs = mCastle.get_num_cannons();
      mProjectiles2 = new projectile2_t [ mNumPlayerProjs ];

      for ( i = 0; i < mNumPlayerProjs; ++i )
      {
        mProjectiles2[ i ].set_model( &mCannonBallModel );
        mProjectiles2[ i ].set_valid( false );
        mCannons[ i ].ready_to_fire();
      }

      // Set the number of seconds for the battle round
      mSeconds = con_battle_time;
    }
  }
}

//////////////////////////////////////////////////////////
// Update function for when game is in PRE_BUILD mode
//////////////////////////////////////////////////////////
void
game_t::update_prebuild()
{
  mSecondCount += gFactTick;
  if ( mSecondCount >= 100.0f )
  {
    mSecondCount = 0;
    --mSeconds;

    // Are we ready for battle?
    if ( mSeconds < 0 )
    {
      mStage = BUILD;

      mCastle.enable_ghost();

      // Randomize the next pieces
      mNextPiece.mType = rand() & 7;
      mCurrPiece.mType = rand() & 7;

      // Set the number of seconds for the build round
      if ( con_difficulty == HARD )
        mSeconds = con_build_time - mBuildRound;
      else
        mSeconds = con_build_time;

      mPreviousSecond = mSeconds;

      if ( mBuildRound <= 5 )
        mBuildRound++;
    }
  }
}

//////////////////////////////////////////////////////////
// Update function for when game is in BUILD mode
//////////////////////////////////////////////////////////
void
game_t::update_build()
{
  handle_mouse_build( false );

  mSecondCount += gFactTick;
  if ( mSecondCount >= 100.0f )
  {
    mSecondCount = 0;
    --mSeconds;

    // Are we ready for battle?
    if ( mSeconds < 0 )
    {
      mCastle.disable_ghost();
      mCastle.reset_conflict_pieces();

      // See if the flag is fully enclosed
      if ( mCastle.check_enclosure() )
      {
        // Let's place a new cannon first
        mStage = PLACE_CANNON_BONUS;
        mSeconds = con_cannon_time;

        // Do territory bonus
        mTerritoryBonus = mCastle.count_enclosed_pieces();
        
        // The goal behind modifing the territory bonus is to balance
        // the points between the battle and build rounds and to
        // encourage riskier play by enticing player's to build large castles

        // Subtract 4 for the flagbase
        mTerritoryBonus -= 4;
        // Subtract the number of cannons
        mTerritoryBonus -= mCannons.size();
        // Multiply by 2
        mTerritoryBonus *= 2;
        mBonusTime = 0;

        mScore += mTerritoryBonus;

        reset_ships();
        mCastle.check_cannon_enclosure();
      }
      else
      {
        // Game over here
        mStage = GAME_OVER;
      }
    }
  }
}

//////////////////////////////////////////////////////////
// Update function for when game is in BATTLE mode
//////////////////////////////////////////////////////////
void
game_t::update_battle()
{
	static int i;
  static vect3_t lHitPos;

  update_ships( true );

  // Update cannons
  for ( i = 0; i < mCannons.size(); ++i )
    mCannons[ i ].update();

  // Update ship projectiles
  for ( i = 0; i < mNumShipProjs; ++i )
  {
    if ( mProjectiles[ i ].update( lHitPos ) )
    {
      mpSounds->play_sample( mExplosionSound );

      add_explosion_particles( lHitPos );
    }
  }

  update_player_projectiles();

  mSecondCount += gFactTick;
  if ( mSecondCount >= 100.0f )
  {
    mSecondCount = 0;
    --mSeconds;

    // Is the battle stage done?
    if ( mSeconds < 0 )
    {
      mStage = POST_BATTLE;
    }
  }
}

//////////////////////////////////////////////////////////
// Update function for when game is in POST_BATTLE mode.
// In this state we let all the cannonballs hit their
// targets but neither the player nor the ships will
// be able to continue firing
//////////////////////////////////////////////////////////
void
game_t::update_postbattle()
{
	static int i;
  static vect3_t lHitPos;
  static bool lDone;
  lDone = true;

  // Keep moving the ships but don't fire
  update_ships( false );

  // Update ship projectiles
  for ( i = 0; i < mNumShipProjs; ++i )
  {
    if ( mProjectiles[ i ].get_valid() )
      lDone = false;

    if ( mProjectiles[ i ].update( lHitPos ) )
    {
      mpSounds->play_sample( mExplosionSound );

      add_explosion_particles( lHitPos );
    }
  }

  // Update player projectiles
  lDone &= update_player_projectiles();

  if ( lDone )
  {
    mSecondCount += gFactTick;
    if ( mSecondCount >= 300.0f )
    {
      // If the castle has no blocks left -> game over
      if ( mCastle.get_num_valid() == 0 )
      {
        mStage = GAME_OVER;
      }
      else
      {
        mSecondCount = 0;

        // Set the number of seconds for the PRE_BUILD phase
        mSeconds = con_warmup_time;

        // Set view pos for cannon-place round
        mViewerPos.x = 500; 
        mViewerPos.y = 1000;
        mViewerPos.z = 500;
        mLookAtPos = mCastle.getFlagPos();

        mViewDir = mLookAtPos - mViewerPos;
        mViewDir.normalize();

        mZoom = 0;
        mViewAngle = 0;

        // Disable all particle systems
        disable_particles();

        // if necessary, delete mProjectiles and mProjectiles2
        if ( mProjectiles ) 
        {
          delete [] mProjectiles;
          mProjectiles = 0;
        }
        if ( mProjectiles2 ) 
        {
          delete [] mProjectiles2;
          mProjectiles2 = 0;
        }

        mBattleRound++;

        mCastle.clear_rubble();

        // Unconvert the pieces before PRE_BUILD
        mCastle.unconvert_pieces();

        mStage = PRE_BUILD;
      }
    }
  }
}


///////////////////////////////////////////////////////////
// Checks if the player's score is good enough for the list
///////////////////////////////////////////////////////////
void
game_t::update_highscores()
{
  static score_data_t lNewScore;

  if ( gpScoreList->check_score( mScore ) )
  {
    // Print a congratulations message and ask for the player's name
    gpScoreList->ask_name( mName, mpFont );
  }
  else
  {
    mStage = WAIT_TO_RESTART;   // no high score was set
    mRestartDelay = 200.0f;
  }

  if ( gpScoreList->get_name_entered() )
  {
    // The player's name has been entered so update the list
    gpScoreList->clear_name_entered();
    strcpy( lNewScore.mName, mName );
    lNewScore.mScore = mScore;
    gpScoreList->update_list( lNewScore );
    mStage = WAIT_TO_RESTART;
    mRestartDelay = 0.0f;
  }
}
  
//////////////////////////////////////////////////////////
void
game_t::update_ships( bool aFire )
{
	static int i;
  static vect3_t lShipPos, lShipPrevPos, lDir, lShipDir;
  static projectile_t *lpProjectile;
  static castle_piece_t *lpPiece;
  static float lDist, lVel;

	for ( i = 0; i < mNumValidShips; ++i )
	{
		mShips[ i ].clear_ship_collide();
  }

	for ( i = 0; i < mNumValidShips; ++i )
	{
    // Get the previous ship position
    lShipPrevPos = mShips[ i ].get_pos();

    if ( mShips[ i ].update( mSegments, mShipPoly, mShips, mBattleRound ) && aFire )
    {
      lShipPos = mShips[ i ].get_pos();

  		// Draw wake particles
      add_wake_particles( mShips[ i ], lShipPrevPos );

      // Must check for a free projectile before shooting
      lpProjectile = find_free_projectile();

      if ( !lpProjectile )
        continue;

      lpPiece = mCastle.shoot(); 

      if ( lpPiece == NULL )
        continue;

      // Ship is shooting at castle
      lpProjectile->set_pos( lShipPos );

      lDir = lpPiece->mPos - lShipPos;
      lpProjectile->set_dir( lDir );

      add_cannon_particles( lShipPos, lpPiece->mPos );
      
      lpProjectile->set_valid( true );
      lpProjectile->set_target_piece( lpPiece );

      lpProjectile->set_vel( 300 );

      lpProjectile->calc_physics();

      mpSounds->play_sample( mCannonSound );
    }
    else if ( aFire )
    {
  		// Draw wake particles
      add_wake_particles( mShips[ i ], lShipPrevPos );
    }
	}
}

//////////////////////////////////////////////////////////
// This function is called after the BUILD round to reset
// the existing ships plus add new ones when necesary
//////////////////////////////////////////////////////////
void
game_t::reset_ships()
{
  // See if we have any ships to add
  if ( mShips.size() > mNumValidShips )
  {
    mNumValidShips++;
  }
  else
  {
    for ( int i = 0; i < mShips.size(); ++i )
    {
      if ( mShips[ i ].get_shiptype() == 1 )
      {
        mShips[ i ].set_shiptype( 2 );
        mShips[ i ].set_model( mShip2Model );
        break;
      }
    }
  }

  for ( int i = 0; i < mNumValidShips; ++i )
    mShips[ i ].reset();
}

//////////////////////////////////////////////////////////
// Check collision between the player's cannon balls and
// the ships, returns true if there are NO cannonbals
// flying
//////////////////////////////////////////////////////////
bool
game_t::update_player_projectiles()
{
  static vect3_t lHitPos, lShipPos;
  static float lDist;
  static bool lHit, lHitShip;
  static bool lDone;

  lDone = true;

  for ( int lProj = 0; lProj < mNumPlayerProjs; ++lProj )
  {
    lHitShip = false;

    if ( mProjectiles2[ lProj ].get_valid() )
      lDone = false;

    // See if the player hit one of the ships
    lHit = mProjectiles2[ lProj ].update( lHitPos );
    if ( lHit )
    {
      // clear mFireCount for the cannon
      mCannons[ lProj ].ready_to_fire();

      for ( int lShip = 0; lShip < mNumValidShips; ++lShip )
      {
        if ( mShips[ lShip ].is_alive() && 
             lHitPos.distanceSquared( mShips[ lShip ].get_pos() ) < 3600.0f )
        {
          if ( mShips[ lShip ].get_health() == con_ship_health )
          {
            add_smoke_particles( mShips[ lShip ] );
          }

          add_explosion_particles( lHitPos );

          mShips[ lShip ].hit();
          lHitShip = true;
          mpSounds->play_sample( mExplosionSound );

          // If we killed the ship -> add to our score
          if ( mShips[ lShip ].get_health() == 0 )
          {
            mScore += 10 * mShips[ lShip ].get_shiptype();
          }

          break;
        }
      }

      if ( !lHitShip )
      {
        // If it didn't hit any ships -> add a water particle system
        add_water_particles( lHitPos );
      }
    }
  }

  return lDone;
}

//////////////////////////////////////////////////////////
projectile_t *
game_t::find_free_projectile()
{
  for ( int i = 0; i < mNumShipProjs; ++i )
  {
    if ( !mProjectiles[ i ].get_valid() )
      return &mProjectiles[ i ];
  }

  return NULL;
}

//////////////////////////////////////////////////////////
projectile2_t *
game_t::find_free_projectile2()
{
  for ( int i = 0; i < mNumPlayerProjs; ++i )
  {
    if ( !mProjectiles2[ i ].get_valid() )
      return &mProjectiles2[ i ];
  }

  return NULL;
}

//////////////////////////////////////////////////////////
void
game_t::draw()
{
  switch ( mStage )
  {
  case PRE_BATTLE :
//    if ( con_geometry != 0 )
//      mCastle.convert_pieces();
  case BATTLE :
  case POST_BATTLE :
    draw_battle();
    break;

  case PRE_BUILD :
//    mCastle.unconvert_pieces();
  case PLACE_CANNON :
  case PLACE_CANNON_BONUS :
  case BUILD :
  case GAME_OVER :
  case WAIT_TO_RESTART :
    draw_build();
    break;
  }
}

//////////////////////////////////////////////////////////
void
game_t::draw_battle()
{
	int i;
  
  glLoadIdentity();
  gluLookAt( mViewerPos.x, mViewerPos.y, mViewerPos.z, 
	           mLookAtPos.x, mLookAtPos.y, mLookAtPos.z, 0, 1, 0 );  

  glRotatef( mViewAngle, 0, 1, 0 );

  glEnable( GL_DEPTH_TEST );
  glDisable( GL_BLEND );
  glCullFace( GL_BACK );

	glEnableClientState( GL_VERTEX_ARRAY );
	glEnableClientState( GL_NORMAL_ARRAY );

	glActiveTextureARB( GL_TEXTURE0_ARB );
	glClientActiveTextureARB( GL_TEXTURE0_ARB );
	glEnable( GL_TEXTURE_2D );
	glEnableClientState( GL_TEXTURE_COORD_ARRAY );

	//glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

  // Light Stuff
  if ( mGlobalLight.get_enable() && con_light )
  {
    glPushMatrix();
    glColorMaterial( GL_FRONT, GL_AMBIENT_AND_DIFFUSE );
    glLightfv( GL_LIGHT0, GL_DIFFUSE, mGlobalLight.get_color() );
    glLightfv( GL_LIGHT0, GL_POSITION, mGlobalLight.get_pos() );
    glEnable( GL_LIGHT0 );
    glEnable( GL_LIGHTING );
    GLfloat lmodel_ambient[] = { 1.5f, 1.5f, 1.5f, 1.5f };
    glLightModelfv( GL_LIGHT_MODEL_AMBIENT, lmodel_ambient );
    glEnable ( GL_LIGHT_MODEL_AMBIENT );
    glPopMatrix();
  }

  int lNumModels = mWorldModels.size();
	for ( i = 0; i < lNumModels; ++i )
		mWorldModels[ i ].draw_pos();

	for ( i = 0; i < mNumValidShips; ++i )
		mShips[ i ].draw();

	mCastle.draw( mGlobalLight.get_enable() );

  if ( mProjectiles ) 
  {
    // Disable the depth test so the projectiles don't disappear
    // behind objects they aren't supposed to hit
    glDisable( GL_DEPTH_TEST );

    for ( i = 0; i < mNumShipProjs; ++i )
      mProjectiles[ i ].draw();
    
    glEnable( GL_DEPTH_TEST );
  }

  if ( mProjectiles2 ) 
  {
    for ( i = 0; i < mNumPlayerProjs; ++i )
      mProjectiles2[ i ].draw();
  }

  glDisable( GL_LIGHTING );

	glDisableClientState( GL_VERTEX_ARRAY );
	glDisableClientState( GL_NORMAL_ARRAY );  
	//

	glActiveTextureARB( GL_TEXTURE0_ARB );
	glClientActiveTextureARB( GL_TEXTURE0_ARB );
	glDisable( GL_TEXTURE_2D );
	glDisableClientState( GL_TEXTURE_COORD_ARRAY );

  // Draw particle systems
  draw_particles();

  if ( con_shadows )
  {
    vect3_t lLightPos;
    lLightPos = mGlobalLight.get_spot_pos();
    
	  glDisableClientState( GL_NORMAL_ARRAY );

	  glActiveTextureARB( GL_TEXTURE0_ARB );
	  glClientActiveTextureARB( GL_TEXTURE0_ARB );
	  glDisable( GL_TEXTURE_2D );
	  glDisableClientState( GL_TEXTURE_COORD_ARRAY );

    glDepthMask( GL_FALSE );
    glEnable( GL_STENCIL_TEST );
    glColorMask( 0, 0, 0, 0 );

    if ( gSeparateStencilATI )
    { 
      glDisable( GL_CULL_FACE );

      glStencilFuncSeparateATI( GL_EQUAL, GL_EQUAL, 0, 0 );

      glStencilOpSeparateATI( GL_FRONT, GL_KEEP, GL_KEEP, GL_INCR_WRAP_EXT );
      glStencilOpSeparateATI( GL_BACK, GL_KEEP, GL_KEEP, GL_DECR_WRAP_EXT );

	    for ( i = 0; i < mNumValidShips; ++i )
		    mShips[ i ].draw_shadow( lLightPos );
    }
    else if ( gStencilTwoSideEXT )
    {
      glDisable( GL_CULL_FACE ); 

      glEnable( GL_STENCIL_TEST_TWO_SIDE_EXT );  
      
      glActiveStencilFaceEXT( GL_FRONT );
      glStencilFunc( GL_ALWAYS, 1, 0xffffffff );
      glStencilOp( GL_KEEP, GL_KEEP, GL_INCR_WRAP_EXT );        
      glStencilMask( -1 );

      glActiveStencilFaceEXT( GL_BACK );
      glStencilFunc( GL_ALWAYS, 1, 0xffffffff );
      glStencilOp( GL_KEEP, GL_KEEP, GL_DECR_WRAP_EXT );        
      glStencilMask( -1 );

	    for ( i = 0; i < mNumValidShips; ++i )
		    mShips[ i ].draw_shadow( lLightPos );
       
      glDisable( GL_STENCIL_TEST_TWO_SIDE_EXT );  
    }
    else
    {
      // Unoptimized stencil path
      glStencilFunc( GL_ALWAYS, 1, 0xffffffff );

	    for ( i = 0; i < mNumValidShips; ++i )
		    mShips[ i ].draw_shadow_two_pass( lLightPos );
    }

    glEnable( GL_CULL_FACE );
    glFrontFace( GL_CCW );
    glColorMask( 1, 1, 1, 1 );

	  glDisableClientState( GL_VERTEX_ARRAY );

    draw_screen_quad();

    glDepthMask( GL_TRUE );
    glDisable( GL_STENCIL_TEST );

    // Reenable arrays
	  glEnableClientState( GL_VERTEX_ARRAY );
	  glEnableClientState( GL_NORMAL_ARRAY );

	  glActiveTextureARB( GL_TEXTURE0_ARB );
	  glClientActiveTextureARB( GL_TEXTURE0_ARB );
	  glEnable( GL_TEXTURE_2D );
	  glEnableClientState( GL_TEXTURE_COORD_ARRAY );
  }

  if ( con_draw_debug )
  {
    static vect3_t lPos;
    static sphere_t lSphere;

    draw_coord_ref();

	  // Draw outline of the area the ships will navigate in
	  draw_debug_ai_area();

  	mCastle.draw_debug();	

    lSphere = mShipModel->get_sphere();

	  for ( i = 0; i < mNumValidShips; ++i )
    {
      if ( !mShips[ i ].is_alive() )
        continue;

      lPos = mShips[ i ].get_pos();

		  glPushMatrix();
		  glTranslatef( lPos.x, lPos.y, lPos.z );
		  gluSphere( mQuadric, lSphere.mRad.f, 4, 4 );
		  glPopMatrix();

      mShips[ i ].draw_debug();
    }
  }
}

//////////////////////////////////////////////////////////
void
game_t::draw_build()
{
	int i;
  
  glLoadIdentity();
  gluLookAt( mViewerPos.x, mViewerPos.y, mViewerPos.z, 
	           mLookAtPos.x, mLookAtPos.y, mLookAtPos.z, 0, 1, 0 );  
 
  glRotatef( mViewAngle, 0, 1, 0 );

  glEnable( GL_DEPTH_TEST );
  glDisable( GL_BLEND );
  glCullFace( GL_BACK );

	glEnableClientState( GL_VERTEX_ARRAY );
	glEnableClientState( GL_NORMAL_ARRAY );

	glActiveTextureARB( GL_TEXTURE0_ARB );
	glClientActiveTextureARB( GL_TEXTURE0_ARB );
	glEnable( GL_TEXTURE_2D );
	glEnableClientState( GL_TEXTURE_COORD_ARRAY );

	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

  // Light Stuff
  if ( mGlobalLight.get_enable() && con_light )
  {
    glPushMatrix();
    glColorMaterial( GL_FRONT, GL_AMBIENT_AND_DIFFUSE );
    glLightfv( GL_LIGHT0, GL_DIFFUSE, mGlobalLight.get_color() );
    glLightfv( GL_LIGHT0, GL_POSITION, mGlobalLight.get_pos() );
    glEnable( GL_LIGHT0 );
    glEnable( GL_LIGHTING );
    GLfloat lmodel_ambient[] = { 1.5f, 1.5f, 1.5f, 1.5f };
    glLightModelfv( GL_LIGHT_MODEL_AMBIENT, lmodel_ambient );
    glEnable ( GL_LIGHT_MODEL_AMBIENT );
    glPopMatrix();
  }

  int lNumModels = mWorldModels.size();
	for ( i = 0; i < lNumModels; ++i )
		mWorldModels[ i ].draw_pos();

	mCastle.draw( mGlobalLight.get_enable() );

	glDisableClientState( GL_VERTEX_ARRAY );
	glDisableClientState( GL_NORMAL_ARRAY );  
	//

	glActiveTextureARB( GL_TEXTURE0_ARB );
	glClientActiveTextureARB( GL_TEXTURE0_ARB );
	glDisable( GL_TEXTURE_2D );
	glDisableClientState( GL_TEXTURE_COORD_ARRAY );

  glDisable( GL_LIGHTING );

  mCastle.draw_debug();	

  if ( con_draw_debug )
  {
    draw_coord_ref();
  }
}   

//////////////////////////////////////////////////////////
void
game_t::draw_text()
{
  static char lStr[ 50 ];

	mpFont->setDefaultDim();
  sprintf( lStr, "Score: %d", mScore );
	mpFont->print( 510, 460, lStr );

  if ( mPause )
  {
    // Blink "PAUSED" in the middle of the screen
    static float lPauseCount = 0;
    lPauseCount += gFactTick;
    if ( lPauseCount >= 100.0f )
    {   
	    mpFont->setDim( 22.5, 30, 16 );
	    mpFont->print( 270, 260, 1, 0, 0, "PAUSED" );      
    }
    if ( lPauseCount >= 200.0f )
    {   
      lPauseCount = 0;
    }

    glColor4f( 1, 1, 1, 1 );
    return;
  }

	mpFont->setDefaultDim();

  switch ( mStage )
  {
  case PLACE_CANNON :
    draw_text_placecannon();
    break;

  case PLACE_CANNON_BONUS :
    draw_text_placecannonbonus();
    break;

  case PRE_BATTLE :
    draw_text_prebattle();
    break;

  case BATTLE :
    draw_text_battle();
    break;

  case PRE_BUILD :
    draw_text_prebuild();
    break;

  case BUILD :
    draw_text_build();
    break;

  case GAME_OVER :
    update_highscores();
    draw_text_gameover();
    break;

  case WAIT_TO_RESTART :
    mpFont->print( 160, 80, 1, 0, 0, "Click a mouse button to continue" );
    gpScoreList->display( mpFont );   // Draw the high score list
    draw_text_gameover();
    break;
  }
	mpFont->setDefaultDim();
}

//////////////////////////////////////////////////////////
void
game_t::draw_text_placecannon()
{  
  static char lStr[ 50 ];

  sprintf( lStr, "%d", mSeconds );    
	mpFont->setDim( 30, 40, 20 );
	mpFont->print( 560, 420, 1, 0, 0, lStr );

	mpFont->setDim( 22.5, 30, 16 );
	mpFont->print( 185, 370, "PLACE YOUR CANNON!" );
  mpFont->setDefaultDim();
  mpFont->print( 195, 350, "( Inside the Castle Walls )" );
}


//////////////////////////////////////////////////////////
void
game_t::draw_text_placecannonbonus()
{  
  static char lStr[ 50 ];

  sprintf( lStr, "%d", mSeconds );    
	mpFont->setDim( 30, 40, 20 );
	mpFont->print( 560, 420, 1, 0, 0, lStr );

	mpFont->setDim( 22.5, 30, 16 );
	mpFont->print( 185, 370, "PLACE YOUR CANNON!" );
  mpFont->setDefaultDim();
  mpFont->print( 195, 350, "( Inside the Castle Walls )" );

  mBonusTime += gFactTick;
  if ( mBonusTime < 300.0f )
  {
	  mpFont->setDim( 22.5, 30, 16 );
    sprintf( lStr, "TERRITORY BONUS: %d", mTerritoryBonus ); 
	  mpFont->print( 170, 170, lStr );
  }
}

//////////////////////////////////////////////////////////
void
game_t::
draw_text_gameover()
{
	mpFont->setDim( 22.5, 30, 16 );
	mpFont->print( 175, 420, "YOU WERE DEFEATED!" );
}

//////////////////////////////////////////////////////////
void
game_t::draw_text_battle()
{
  static char lStr[ 50 ];

  sprintf( lStr, "%d", mSeconds );    
	mpFont->setDim( 30, 40, 20 );
	mpFont->print( 560, 420, 1, 0, 0, lStr );
}

//////////////////////////////////////////////////////////
void
game_t::draw_text_prebattle()
{  
	mpFont->setDim( 22.5, 30, 16 );
	mpFont->print( 180, 260, "GET READY TO FIGHT!" );
}

//////////////////////////////////////////////////////////
void
game_t::draw_text_prebuild()
{  
	mpFont->setDim( 22.5, 30, 16 );
	mpFont->print( 180, 260, "GET READY TO BUILD!" );
  mpFont->setDefaultDim();
	mpFont->print( 170, 240, "( Right Click to Rotate Pieces )" );
}

//////////////////////////////////////////////////////////
void
game_t::draw_text_build()
{
  static char lStr[ 50 ];
  static int lCountSoundVol;

  sprintf( lStr, "%d", mSeconds );
	mpFont->setDim( 30, 40, 20 );
	mpFont->print( 560, 420, 1, 0, 0, lStr );

  // If the count changed and is < 5 play a countdown sound
  if ( mSeconds != mPreviousSecond && mSeconds < 5 )
  {
    mPreviousSecond = mSeconds;
    switch ( mSeconds )
    {
    case 4 :
      lCountSoundVol = con_sndfx_vol * 0.1f;
      break;
    case 3 :
      lCountSoundVol = con_sndfx_vol * 0.2f;
      break;
    case 2 :
      lCountSoundVol = con_sndfx_vol * 0.3f;
      break;
    case 1 :
      lCountSoundVol = con_sndfx_vol * 0.5f;
      break;
    case 0 :
      lCountSoundVol = con_sndfx_vol;
      break;
    }
    if ( lCountSoundVol < 0 )
      lCountSoundVol = 0;

    mpSounds->set_sample_volume( mNegateSound, lCountSoundVol );
    mpSounds->play_sample( mNegateSound );

    // Reset the sound volume
    mpSounds->set_sample_volume( mNegateSound, con_sndfx_vol );
  }

  draw_next_piece();
}

//////////////////////////////////////////////////////////
void
game_t::draw_next_piece()
{      
	glPushMatrix(); 
	glTranslatef( 540, 400, 0 ); 
  glColor3f( 1, 1, 1 );
  glBindTexture( GL_TEXTURE_2D, mPieceTextures[ mNextPiece.mType ] );

  glBegin( GL_QUADS );
  glTexCoord2f( 1, 0 ); 
  glVertex2f( 20, 20 );

  glTexCoord2f( 1, 1 ); 
  glVertex2f( 20, -20 );

  glTexCoord2f( 0, 1 ); 
  glVertex2f( -20, -20 );
    
  glTexCoord2f( 0, 0 ); 
  glVertex2f( -20, 20 );
  glEnd();  
	glPopMatrix(); 
}

//////////////////////////////////////////////////////////
void 
game_t::draw_debug_ai_area()
{
	glColor3f( 1, 1, 0 );
	glBegin( GL_LINE_LOOP );

	for ( int i = 0; i < mShipPoly.size(); ++i )
	{
		glVertex3fv( mShipPoly[ i ].mVect );
	}

	glEnd();
}

//////////////////////////////////////////////////////////
void 
game_t::draw_coord_ref()
{
  // Draw coordinate reference
  glBegin( GL_LINES );
  glColor3f( 0, 0, 1 );
  glVertex3f( 0, 0.1f, 0 );
  glVertex3f( 100, 0.1f, 0 );
  
  glColor3f( 1, 1, 1 );
  glVertex3f( 0, 0.1f, 0 );
  glVertex3f( 0, 100, 0 );
  
  glColor3f( 1, 0, 0 );
  glVertex3f( 0, 0.1f, 0 );
  glVertex3f( 0, 0.1f, 100 );
  glEnd();
}
	
//////////////////////////////////////////////////////////
void 
game_t::load_level( const char *apFileName )
{
  static bool lLoaded = false;

	if ( mValid )
		clear();

  if ( !lLoaded )
  {
    lLoaded = true;

    // Load game data constant to all levels
    load_constant_data();
  }

	string lFileName = "data/levels/";
	lFileName += apFileName;

	FILE *lFP = fopen( lFileName.c_str(), "r" );

	if ( !lFP )
	{		
		gCon->add_msg( "Could not load map %s", apFileName );
		return;
	}

	gCon->add_msg( "Parsing map %s", apFileName );


	int lLineCount = 0;
	char lLine[ 500 ], *tok; 	
	while ( fgets( lLine, 500, lFP ) != NULL )
	{    
		lLineCount++;

    tok = strtok( lLine, " \n\t\r" );
    if ( !tok || tok[ 0 ] == '-' || tok[ 0 ] == '#' )
      continue;
    
    if ( strcmp( tok, "include" ) == 0 )
    {
      tok = strtok( NULL, " \n\t\r" );
      mShaderManager.readFromFile( tok );
    }
		else if ( strcmp( tok, "seg" ) == 0 )
		{  
			segment_t lSeg;
			vect3_t lP;
			lP.y = 10;

			tok = strtok( NULL, " \n\t\r" );
			sscanf( tok, "%f", &lP.x );
			tok = strtok( NULL, " \n\t\r" );
			sscanf( tok, "%f", &lP.z );
			lSeg.set_point( 0, lP );

			tok = strtok( NULL, " \n\t\r" );
			sscanf( tok, "%f", &lP.x );
			tok = strtok( NULL, " \n\t\r" );
			sscanf( tok, "%f", &lP.z );
			lSeg.set_point( 1, lP );

			mSegments.push_back( lSeg );
		}
		else if ( strcmp( tok, "ship2" ) == 0 )
		{  
			parse_ship( lLineCount, 2 );
		}
		else if ( strcmp( tok, "ship" ) == 0 )
		{  
			parse_ship( lLineCount, 1 );
		}
		else if ( strcmp( tok, "model" ) == 0 )
		{  
			parse_model( lLineCount );
		}
		else if ( strcmp( tok, "castle" ) == 0 )
		{  
			parse_castle( lLineCount );
		}
		else if ( strcmp( tok, "light" ) == 0 )
		{  
			parse_light( lLineCount );
		}
		else
		{
			gCon->add_msg( "ERROR: Illegal token '%s' in line %d",
				           tok, lLineCount );
		}
	}

  restart();

  if ( mNumValidShips > mShips.size() )
    mNumValidShips = mShips.size();

  int i;
  mShipPoly.push_back( mSegments[ 0 ].get_point( 0 ) );
  mShipPoly.push_back( mSegments[ 0 ].get_point( 1 ) );
  for ( i = 1; i < mSegments.size() - 1; i++ )
  {
    mShipPoly.push_back( mSegments[ i ].get_point( 1 ) );
  }
}

//////////////////////////////////////////////////////////
// Load game data constant to all levels
//////////////////////////////////////////////////////////
void 
game_t::load_constant_data()
{
  // Load and scale block model by con var block_size
  model_t *lModel;
  shadow_model_t *lShadowModel;

  // Load and scale rubble model by con var block_size
  model_node_t *lPieceModels[ 11 ];

  lModel = mModelManager.insert( "castle_block.b3d", con_block_size );
  mCastleModel[ BLOCK_LOW ].set_model( lModel );
  mCastleModel[ BLOCK_LOW ].set_texture( mpTextures->insert( "blocks.bmp" ) );
  lPieceModels[ BLOCK_LOW ] = &mCastleModel[ BLOCK_LOW ];

  lModel = mModelManager.insert( "wall_low.b3d", con_block_size );
  mCastleModel[ WALL_LOW ].set_model( lModel );
  mCastleModel[ WALL_LOW ].set_texture( mpTextures->insert( "blocks.bmp" ) );
  lPieceModels[ WALL_LOW ] = &mCastleModel[ WALL_LOW ];

  lModel = mModelManager.insert( "wall_high.b3d", con_block_size );
  mCastleModel[ WALL_HIGH ].set_model( lModel );
  mCastleModel[ WALL_HIGH ].set_texture( mpTextures->insert( "blocks.bmp" ) );
  lPieceModels[ WALL_HIGH ] = &mCastleModel[ WALL_HIGH ];

  lModel = mModelManager.insert( "tee_low.b3d", con_block_size );
  mCastleModel[ TEE_LOW ].set_model( lModel );
  mCastleModel[ TEE_LOW ].set_texture( mpTextures->insert( "blocks.bmp" ) );
  lPieceModels[ TEE_LOW ] = &mCastleModel[ TEE_LOW ];

  lModel = mModelManager.insert( "tee_high.b3d", con_block_size );
  mCastleModel[ TEE_HIGH ].set_model( lModel );
  mCastleModel[ TEE_HIGH ].set_texture( mpTextures->insert( "blocks.bmp" ) );
  lPieceModels[ TEE_HIGH ] = &mCastleModel[ TEE_HIGH ];

  lModel = mModelManager.insert( "corner_low.b3d", con_block_size );
  mCastleModel[ CORNER_LOW ].set_model( lModel );
  mCastleModel[ CORNER_LOW ].set_texture( mpTextures->insert( "blocks.bmp" ) );
  lPieceModels[ CORNER_LOW ] = &mCastleModel[ CORNER_LOW ];

  lModel = mModelManager.insert( "corner_high.b3d", con_block_size );
  mCastleModel[ CORNER_HIGH ].set_model( lModel );
  mCastleModel[ CORNER_HIGH ].set_texture( mpTextures->insert( "blocks.bmp" ) );
  lPieceModels[ CORNER_HIGH ] = &mCastleModel[ CORNER_HIGH ];

  lModel = mModelManager.insert( "cross_low.b3d", con_block_size );
  mCastleModel[ CROSS_LOW ].set_model( lModel );
  mCastleModel[ CROSS_LOW ].set_texture( mpTextures->insert( "blocks.bmp" ) );
  lPieceModels[ CROSS_LOW ] = &mCastleModel[ CROSS_LOW ];

  lModel = mModelManager.insert( "cross_high.b3d", con_block_size );
  mCastleModel[ CROSS_HIGH ].set_model( lModel );
  mCastleModel[ CROSS_HIGH ].set_texture( mpTextures->insert( "blocks.bmp" ) );
  lPieceModels[ CROSS_HIGH ] = &mCastleModel[ CROSS_HIGH ];

  lModel = mModelManager.insert( "tower_low.b3d", con_block_size );
  mCastleModel[ TOWER_LOW ].set_model( lModel );
  mCastleModel[ TOWER_LOW ].set_texture( mpTextures->insert( "blocks.bmp" ) );
  lPieceModels[ TOWER_LOW ] = &mCastleModel[ TOWER_LOW ];

  lModel = mModelManager.insert( "tower_high.b3d", con_block_size );
  mCastleModel[ TOWER_HIGH ].set_model( lModel );
  mCastleModel[ TOWER_HIGH ].set_texture( mpTextures->insert( "blocks.bmp" ) );
  lPieceModels[ TOWER_HIGH ] = &mCastleModel[ TOWER_HIGH ];

  mCastle.set_piece_models( lPieceModels );
  mCastle.set_piece_model( lPieceModels[ BLOCK_LOW ] );   // Initialize the model

  // Load and scale rubble model by con var block_size
  model_node_t *lRubbleModels[ 2 ];

  lModel = mModelManager.insert( "rubble1.b3d", con_block_size );
  mRubbleModel[ 0 ].set_model( lModel );
  mRubbleModel[ 0 ].set_texture( mpTextures->insert( "rubble.jpg" ) );
  lRubbleModels[ 0 ] = &mRubbleModel[ 0 ];

  lModel = mModelManager.insert( "rubble2.b3d", con_block_size );
  mRubbleModel[ 1 ].set_model( lModel );
  mRubbleModel[ 1 ].set_texture( mpTextures->insert( "rubble.jpg" ) );
  lRubbleModels[ 1 ] = &mRubbleModel[ 1 ];
  
  mCastle.set_rubble_model( lRubbleModels );

  // Load cannonball model
  lModel = mModelManager.insert( "cannonball.b3d" );
  mCannonBallModel.set_model( lModel );
  mCannonBallModel.set_texture( mpTextures->insert( "cannon.jpg" ) );

  // Load cannon model
  lModel = mModelManager.insert( "cannon.b3d", con_block_size );
  mCannonModel.set_model( lModel );
  mCannonModel.set_texture( mpTextures->insert( "cannon.jpg" ) );

  // Load ship model
  lShadowModel = mModelManager.insert_shadow( "ship_low.b3d" );
  mShipModelLow.set_model( lShadowModel );
  mShipModelLow.set_texture( mpTextures->insert( "ship.jpg" ) );

  // Load ship model
  lShadowModel = mModelManager.insert_shadow( "ship_high.b3d" );
  mShipModelHigh.set_model( lShadowModel );
  mShipModelHigh.set_texture( mpTextures->insert( "ship.jpg" ) );

  // Load ship2 model
  lShadowModel = mModelManager.insert_shadow( "ship2_low.b3d" );
  mShip2ModelLow.set_model( lShadowModel );
  mShip2ModelLow.set_texture( mpTextures->insert( "ship.jpg" ) );

  // Load ship2 model
  lShadowModel = mModelManager.insert_shadow( "ship2_high.b3d" );
  mShip2ModelHigh.set_model( lShadowModel );
  mShip2ModelHigh.set_texture( mpTextures->insert( "ship.jpg" ) );

  // Assign the proper ship models
  if ( con_geometry == 2 )
  {
    mShipModel = &mShipModelHigh;
    mShip2Model = &mShip2ModelHigh;
  }
  else
  {
    mShipModel = &mShipModelLow;
    mShip2Model = &mShip2ModelLow;
  }

  mCannonSound = mpSounds->load_sample( "cannon.wav" );
  mExplosionSound = mpSounds->load_sample( "explosion.wav" );
  mNegateSound = mpSounds->load_sample( "click_enter.wav" );
  mCastle.set_negate_sound( mNegateSound );

  // Load piece textures
  mPieceTextures[ 0 ] = mpTextures->insert( "piece_i.jpg" );
  mPieceTextures[ 1 ] = mpTextures->insert( "piece_l.jpg" );
  mPieceTextures[ 2 ] = mpTextures->insert( "piece_t.jpg" );
  mPieceTextures[ 3 ] = mpTextures->insert( "piece_s.jpg" );
  mPieceTextures[ 4 ] = mpTextures->insert( "piece_2i.jpg" );
  mPieceTextures[ 5 ] = mpTextures->insert( "piece_sing.jpg" );
  mPieceTextures[ 6 ] = mpTextures->insert( "piece_crnr.jpg" );
  mPieceTextures[ 7 ] = mpTextures->insert( "piece_s_inv.jpg" );

  // Load flag
  int lTex;
  lTex = mpTextures->insert( "flag.jpg" );

  model_node_t *lFlagModels[ 10 ];

  lModel = mModelManager.insert( "flag_0.b3d" );
  mFlagModel[ 0 ].set_model( lModel );
  mFlagModel[ 0 ].set_texture( lTex );
  lFlagModels[ 0 ] = &mFlagModel[ 0 ];

  lModel = mModelManager.insert( "flag_1.b3d" );
  mFlagModel[ 1 ].set_model( lModel );
  mFlagModel[ 1 ].set_texture( lTex );
  lFlagModels[ 1 ] = &mFlagModel[ 1 ];

  lModel = mModelManager.insert( "flag_2.b3d" );
  mFlagModel[ 2 ].set_model( lModel );
  mFlagModel[ 2 ].set_texture( lTex );
  lFlagModels[ 2 ] = &mFlagModel[ 2 ];

  lModel = mModelManager.insert( "flag_3.b3d" );
  mFlagModel[ 3 ].set_model( lModel );
  mFlagModel[ 3 ].set_texture( lTex );
  lFlagModels[ 3 ] = &mFlagModel[ 3 ];

  lModel = mModelManager.insert( "flag_4.b3d" );
  mFlagModel[ 4 ].set_model( lModel );
  mFlagModel[ 4 ].set_texture( lTex );
  lFlagModels[ 4 ] = &mFlagModel[ 4 ];

  lModel = mModelManager.insert( "flag_5.b3d" );
  mFlagModel[ 5 ].set_model( lModel );
  mFlagModel[ 5 ].set_texture( lTex );
  lFlagModels[ 5 ] = &mFlagModel[ 5 ];

  lModel = mModelManager.insert( "flag_6.b3d" );
  mFlagModel[ 6 ].set_model( lModel );
  mFlagModel[ 6 ].set_texture( lTex );
  lFlagModels[ 6 ] = &mFlagModel[ 6 ];

  lModel = mModelManager.insert( "flag_7.b3d" );
  mFlagModel[ 7 ].set_model( lModel );
  mFlagModel[ 7 ].set_texture( lTex );
  lFlagModels[ 7 ] = &mFlagModel[ 7 ];

  lModel = mModelManager.insert( "flagpole.b3d" );
  mFlagPoleModel.set_model( lModel );
  mFlagPoleModel.set_texture( lTex );
  lFlagModels[ 8 ] = &mFlagPoleModel;

  lModel = mModelManager.insert( "flagbase.b3d", con_block_size );
  mFlagBaseModel.set_model( lModel );
  mFlagBaseModel.set_texture( mpTextures->insert( "flagbase.bmp" ) );
  lFlagModels[ 9 ] = &mFlagBaseModel;
  
  mCastle.set_flag_model( lFlagModels );

  // Set particle systems texture
  for ( int i = 0; i < PCOUNT; ++i )
  {
    mWaterParticles[ i ].set_texture( gParticleTexture );
    mWakeParticles[ i ].set_texture( gParticleTexture );
    mSmokeParticles[ i ].set_texture( gParticleTexture );
    mCannonParticles[ i ].set_texture( gParticleTexture );
    mExplosionParticles[ i ].set_texture( gParticleTexture );
  }
}
	
//////////////////////////////////////////////////////////
void 
game_t::parse_castle( int &aLineCount )
{
  vect3_t lMin, lMax, lFlagPos, lPolyPoint;
//  castle_t lCastle;
  char *lTok;
  int lNumBoundaryPoints;
  vector<vect3_t> lPolyPoints;

  // Read the flag position
	lTok = strtok( NULL, " \n\t\r" );
	sscanf( lTok, "%f", &lFlagPos.x );
	lTok = strtok( NULL, " \n\t\r" );
	sscanf( lTok, "%f", &lFlagPos.y );
	lTok = strtok( NULL, " \n\t\r" );
	sscanf( lTok, "%f", &lFlagPos.z );
  
  // Read the number of boundary points
  lTok = strtok( NULL, " \n\t\r" );
  sscanf( lTok, "%d", &lNumBoundaryPoints );

  lMin.x = lFlagPos.x;
  lMin.y = lFlagPos.y;
  lMin.z = lFlagPos.z;
  // Loop through all of the boundary points
  for ( int i = 0; i < lNumBoundaryPoints; i++ )
  {
    lTok = strtok( NULL, " \n\t\r" );
    sscanf( lTok, "%f", &lPolyPoint.x );
    lTok = strtok( NULL, " \n\t\r" );
    sscanf( lTok, "%f", &lPolyPoint.y );
    lTok = strtok( NULL, " \n\t\r" );
    sscanf( lTok, "%f", &lPolyPoint.z );

    lPolyPoints.push_back( lPolyPoint );

    // calculate lMin and lMax from the boundary segments
    if ( lPolyPoints[i].x < lMin.x )
      lMin.x = lPolyPoints[i].x;
    if ( lPolyPoints[i].x > lMax.x )
      lMax.x = lPolyPoints[i].x;
    if ( lPolyPoints[i].y < lMin.y )
      lMin.y = lPolyPoints[i].y;
    if ( lPolyPoints[i].y > lMax.y )
      lMax.y = lPolyPoints[i].y;
    if ( lPolyPoints[i].z < lMin.z )
      lMin.z = lPolyPoints[i].z;
    if ( lPolyPoints[i].z > lMax.z )
      lMax.z = lPolyPoints[i].z;
  }

  mCastle.setFlagPos( lFlagPos );
  mCastle.setNumBoundaryPoints( lNumBoundaryPoints );
  mCastle.setPolyPoints( lPolyPoints );
  mCastle.set_rect( lMin, lMax );
  mCastle.set_cannon_model( &mCannonModel );
}

//////////////////////////////////////////////////////////
void 
game_t::parse_ship( int &aLineCount, int shipType )
{
  static int lID = 0; // Create a unique ID for each ship

	char *lTok;
	vect3_t lPos, lDir;
	lPos.y = 0;
	ship_t lShip;

	// Parse ship's starting position
	lTok = strtok( NULL, " \n\t\r" );
	sscanf( lTok, "%f", &lPos.x );
	lTok = strtok( NULL, " \n\t\r" );
	sscanf( lTok, "%f", &lPos.z );
	
	// Set the ship type
  lShip.set_shiptype( shipType );
  
  if ( shipType == 2 )
    lShip.set_model( mShip2Model );
  else
    lShip.set_model( mShipModel );
	lShip.set_pos( lPos );
	lShip.set_vel( get_ship_velocity() );
	
	// Randomize ship's starting direction
  switch ( rand() % 2 )
  {
  case 0 : lDir.x = -1; lDir.z = 0;  break;
  case 1 : lDir.x = 1;  lDir.z = 0;  break;
  case 2 : lDir.x = 0;  lDir.z = -1; break;
  case 3 : lDir.x = 0;  lDir.z = 1;  break;
  }
	lDir.y = 0;
	lShip.set_dir( lDir );

  lShip.set_id( lID++ );
  lShip.reset();

  // Insert ship in array
	mShips.push_back( lShip );

  // Keep track of the original ship type for restart()
  mOrigShipType.push_back( lShip.get_shiptype() );
  
  mNumShipProjs = mShips.size();
}
	
//////////////////////////////////////////////////////////
void 
game_t::parse_model( int &aLineCount )
{
	char *lModelName, *lTexName, *lTok;
	vect3_t lPos;

	// Parse model for the model
	lModelName = strtok( NULL, " \n\t\r" );
	// Parse texture for the model
	lTexName = strtok( NULL, " \n\t\r" );	

	// Parse model's staring position
	lTok = strtok( NULL, " \n\t\r" );
	sscanf( lTok, "%f", &lPos.x );
	lTok = strtok( NULL, " \n\t\r" );
	sscanf( lTok, "%f", &lPos.y );
	lTok = strtok( NULL, " \n\t\r" );
	sscanf( lTok, "%f", &lPos.z );

	model_t *lpModel = mModelManager.insert( lModelName );

  if ( lpModel == NULL )
  {
    // IF the load fails -> just quit
    return;
  }

	model_node_t lModelNode;
	lModelNode.set_model( lpModel );

  char *lExt = COM_FileExtension( lTexName );

  if ( strcmp( lExt, "" ) == 0 )
  {
    shader_t *lpShader = mShaderManager.getShader( lTexName );
    lModelNode.set_shader( lpShader );
  }
  else
  {
	  int lTex = mpTextures->insert( lTexName );
	  lModelNode.set_texture( lTex );
  }
	lModelNode.set_pos( lPos );

	mWorldModels.push_back( lModelNode );
}

//////////////////////////////////////////////////////////
void 
game_t::parse_light( int &aLineCount )
{
  char *lTok;
  float lVal[ 11 ];

	// Parse position and color
  for ( int i = 0; i < 11; i++ )
  {
	  lTok = strtok( NULL, " \n\t\r" );
	  sscanf( lTok, "%f", &lVal[ i ] );
  }

  mGlobalLight.enable();
  mGlobalLight.set_pos( &lVal[ 0 ] );
  mGlobalLight.set_color( &lVal[ 4 ] );
  mGlobalLight.set_spot_pos( &lVal[ 8 ] );
}

//////////////////////////////////////////////////////////
// Look at current difficulty level and assign ship proper 
// velocity
//////////////////////////////////////////////////////////
float 
game_t::get_ship_velocity()
{
	float lVel;
	switch ( con_difficulty )
	{
	case EASY : 
		lVel = 0.20f;
		break;

	case HARD :
		lVel = 0.30f;
		break;
	}
	return lVel;
}

///////////////////////////////////////////////////////////////////////
void
game_t::handle_left_mouse( bool aClicked )
{
  if ( mPause )
    return;

  switch ( mStage )
  {
  case PLACE_CANNON :
  case PLACE_CANNON_BONUS :
    handle_mouse_placecannon( aClicked );
    break;

  case BATTLE :
    handle_mouse_battle();
    break;

  case BUILD :
    handle_mouse_build( aClicked );
    break;

  case WAIT_TO_RESTART :
    if ( mRestartDelay <= 0.0f )
      restart();
    break;

  default :
    return;
  }
}

///////////////////////////////////////////////////////////////////////
// Rotate the current piece
///////////////////////////////////////////////////////////////////////
void
game_t::handle_right_mouse( bool aClicked )
{
   if ( mPause )
    return;
	
  switch ( mStage )
  {
  case BUILD :
    ++mCurrPiece.mAngle;
    mCurrPiece.mAngle &= 3;
    break;

  case WAIT_TO_RESTART :
    if ( mRestartDelay <= 0.0f )
      restart();
    break;

  default :
    return;
  }
}

///////////////////////////////////////////////////////////////////////
void
game_t::handle_mouse_battle()
{
	static double lPos[ 3 ];    
	static int lViewPort[ 4 ];
	static GLuint lDepth;
	static double lModelMatrix[ 16 ];
	static double lProjMatrix[ 16 ];
	static float lDistance;
	static GLuint lBuffer[ 512 ];
	static GLint lHits;
  static vect3_t lTargetPos, lTargetPosY;

	static POINT lMousePos;
	GetCursorPos( &lMousePos );
  lMousePos.y -= 16;    // 16 is half the cursor height

	glGetIntegerv( GL_VIEWPORT, lViewPort );
/*	glSelectBuffer( 512, lBuffer );

	// Puts OpenGL In Selection Mode. Nothing Will Be Drawn.
	glRenderMode( GL_SELECT );

	glInitNames();
	glPushName( 0 );

	glMatrixMode( GL_PROJECTION );
	glPushMatrix();
	glLoadIdentity();

	gluPickMatrix( lMousePos.x, ( lViewPort[ 3 ] - lMousePos.y ), 1.0f, 1.0f, lViewPort );

	// Apply The Perspective Matrix
	gluPerspective( 45.0f, ( float ) gScreenWidth / ( float ) gScreenHeight, 
			            con_znear, con_zfar );
	glMatrixMode( GL_MODELVIEW );
	draw_select_battle();
	glMatrixMode( GL_PROJECTION );
	glPopMatrix();
	glMatrixMode( GL_MODELVIEW );
	lHits = glRenderMode( GL_RENDER );

	if ( lHits )
	{
		lDepth = lBuffer[ 1 ];

    // Loop through all the hits
		for ( int i = 1; i < lHits; ++i )
		{
			// If This Object Is Closer To Us Than The One We Have Selected
			if ( lBuffer[ ( i << 2 ) + 1 ] < lDepth )
			{
				lDepth = lBuffer[ ( i << 2 ) + 1 ];
			}       
		}

    lDistance = ( float ) lDepth / 4294967295.0f;
				 
    glGetDoublev( GL_MODELVIEW_MATRIX, lModelMatrix );
    glGetDoublev( GL_PROJECTION_MATRIX, lProjMatrix );
				
		gluUnProject( lMousePos.x, ( lViewPort[ 3 ] - lMousePos.y ), lDistance, 
				          lModelMatrix, lProjMatrix, lViewPort, 
 							    &lPos[ 0 ], &lPos[ 1 ], &lPos[ 2 ] );
    
    lTargetPos.x = lPos[ 0 ];
    lTargetPos.y = lPos[ 1 ];
    lTargetPos.z = lPos[ 2 ];
    player_shoot( lTargetPos );   
	}
*/

	// Get the Z coordinate for the pixel location
  glReadPixels( lMousePos.x, ( lViewPort[ 3 ] - lMousePos.y ), 1, 1, 
                GL_DEPTH_COMPONENT, GL_UNSIGNED_INT, &lDepth );

  int zdepth;
  float lDistTermInv;
  glGetIntegerv( GL_DEPTH_BITS, &zdepth );
  switch ( zdepth )
  {
  case 16 :
    lDistTermInv = (float) 1.5259018967e-5;     // 65535
    break;
  default :
    lDistTermInv = (float) 2.32830643708e-10;   // 4294967295
    break;
  }

//  lDistance = ( float ) lDepth / 4294967295.0f;
  lDistance = ( float ) lDepth * lDistTermInv;

  glGetDoublev( GL_MODELVIEW_MATRIX, lModelMatrix );
  glGetDoublev( GL_PROJECTION_MATRIX, lProjMatrix );
				
  gluUnProject( lMousePos.x, ( lViewPort[ 3 ] - lMousePos.y ), lDistance, 
                lModelMatrix, lProjMatrix, lViewPort, 
                &lPos[ 0 ], &lPos[ 1 ], &lPos[ 2 ] );
    
  lTargetPos.x = lPos[ 0 ];
  lTargetPos.y = lPos[ 1 ];
  lTargetPos.z = lPos[ 2 ];
  player_shoot( lTargetPos );

}

///////////////////////////////////////////////////////////////////////
void
game_t::handle_mouse_build( bool aClicked )
{
	static double lPos[ 3 ];    
	static int lViewPort[ 4 ];
	static GLuint lDepth;
	static double lModelMatrix[ 16 ];
	static double lProjMatrix[ 16 ];
	static float lDistance;
	static GLuint lBuffer[ 512 ], lName;
	static GLint lHits;
  static vect3_t lTargetPos;

	static POINT lMousePos;
	GetCursorPos( &lMousePos );

	glGetIntegerv( GL_VIEWPORT, lViewPort );
/*	glSelectBuffer( 512, lBuffer );

	// Puts OpenGL In Selection Mode. Nothing Will Be Drawn.
	glRenderMode( GL_SELECT );

	glInitNames();
	glPushName( 0 );

	glMatrixMode( GL_PROJECTION );
	glPushMatrix();
	glLoadIdentity();

	gluPickMatrix( lMousePos.x, ( lViewPort[ 3 ] - lMousePos.y ), 1.0f, 1.0f, lViewPort );

	// Apply The Perspective Matrix
	gluPerspective( 45.0f, ( float ) gScreenWidth / ( float ) gScreenHeight, 
			            con_znear, con_zfar );
	glMatrixMode( GL_MODELVIEW );
	draw_select_build();
	glMatrixMode( GL_PROJECTION );
	glPopMatrix();
	glMatrixMode( GL_MODELVIEW );
	lHits = glRenderMode( GL_RENDER );

  gCursor.enable();
  mCastle.disable_ghost();
	if ( lHits )
	{
		lDepth = lBuffer[ 1 ];
    lName = lBuffer[ 3 ];

    // Loop through all the hits
		for ( int i = 1; i < lHits; ++i )
		{
			// If This Object Is Closer To Us Than The One We Have Selected
			if ( lBuffer[ ( i << 2 ) + 1 ] < lDepth )
			{
				lDepth = lBuffer[ ( i << 2 ) + 1 ];
				lName = lBuffer[ ( i << 2 ) + 3 ];
			}       
		}

    lDistance = ( float ) lDepth / 4294967295.0f;
				 
    glGetDoublev( GL_MODELVIEW_MATRIX, lModelMatrix );
    glGetDoublev( GL_PROJECTION_MATRIX, lProjMatrix );
				
		gluUnProject( lMousePos.x, ( lViewPort[ 3 ] - lMousePos.y ), lDistance, 
				          lModelMatrix, lProjMatrix, lViewPort, 
 							    &lPos[ 0 ], &lPos[ 1 ], &lPos[ 2 ] );
    
    lTargetPos.x = lPos[ 0 ];
    lTargetPos.y = lPos[ 1 ];
    lTargetPos.z = lPos[ 2 ];

    // See if we can place the current piece in the castle
    if ( mCastle.check( lTargetPos, mCurrPiece, aClicked ) )
    {
      // Pick a new next piece
      mCurrPiece = mNextPiece;
      mNextPiece.mType = rand() & 7;
    }
	}
*/
	// Get the Z coordinate for the pixel location
  glReadPixels( lMousePos.x, ( lViewPort[ 3 ] - lMousePos.y ), 1, 1, 
                GL_DEPTH_COMPONENT, GL_UNSIGNED_INT, &lDepth );

  int zdepth;
  float lDistTermInv;
  glGetIntegerv( GL_DEPTH_BITS, &zdepth );
  switch ( zdepth )
  {
  case 16 :
    lDistTermInv = (float) 1.5259018967e-5;     // 65535
    break;
  default :
    lDistTermInv = (float) 2.32830643708e-10;   // 4294967295
    break;
  }

//  lDistance = ( float ) lDepth / 4294967295.0f;
  lDistance = ( float ) lDepth * lDistTermInv;

  glGetDoublev( GL_MODELVIEW_MATRIX, lModelMatrix );
  glGetDoublev( GL_PROJECTION_MATRIX, lProjMatrix );
				
  gluUnProject( lMousePos.x, ( lViewPort[ 3 ] - lMousePos.y ), lDistance, 
                lModelMatrix, lProjMatrix, lViewPort, 
                &lPos[ 0 ], &lPos[ 1 ], &lPos[ 2 ] );
    
  lTargetPos.x = lPos[ 0 ];
  lTargetPos.y = lPos[ 1 ];
  lTargetPos.z = lPos[ 2 ];

  // See if we can place the current piece in the castle
  if ( mCastle.check( lTargetPos, mCurrPiece, aClicked ) )
  {
    // Pick a new next piece
    mCurrPiece = mNextPiece;
    mNextPiece.mType = rand() & 7;
  }

  mCastle.set_tetris_piece( mCurrPiece );
}

///////////////////////////////////////////////////////////////////////
void
game_t::handle_mouse_placecannon( bool aClicked )
{
	static double lPos[ 3 ];    
	static int lViewPort[ 4 ];
	static GLuint lDepth;
	static double lModelMatrix[ 16 ];
	static double lProjMatrix[ 16 ];
	static float lDistance;
	static GLuint lBuffer[ 512 ], lName;
	static GLint lHits;
  static vect3_t lTargetPos, lCannonPos;

	static POINT lMousePos;
	GetCursorPos( &lMousePos );

	glGetIntegerv( GL_VIEWPORT, lViewPort );
/*	glSelectBuffer( 512, lBuffer );

	// Puts OpenGL In Selection Mode. Nothing Will Be Drawn.
	glRenderMode( GL_SELECT );

	glInitNames();
	glPushName( 0 );

	glMatrixMode( GL_PROJECTION );
	glPushMatrix();
	glLoadIdentity();

	gluPickMatrix( lMousePos.x, ( lViewPort[ 3 ] - lMousePos.y ), 1.0f, 1.0f, lViewPort );

	// Apply The Perspective Matrix
	gluPerspective( 45.0f, ( float ) gScreenWidth / ( float ) gScreenHeight, 
			            con_znear, con_zfar );
	glMatrixMode( GL_MODELVIEW );
	draw_select_build();
	glMatrixMode( GL_PROJECTION );
	glPopMatrix();
	glMatrixMode( GL_MODELVIEW );
	lHits = glRenderMode( GL_RENDER );

  gCursor.enable();
	if ( lHits )
	{
		lDepth = lBuffer[ 1 ];
    lName = lBuffer[ 3 ];

    // Loop through all the hits
		for ( int i = 1; i < lHits; ++i )
		{
			// If This Object Is Closer To Us Than The One We Have Selected
			if ( lBuffer[ ( i << 2 ) + 1 ] < lDepth )
			{
				lDepth = lBuffer[ ( i << 2 ) + 1 ];
				lName = lBuffer[ ( i << 2 ) + 3 ];
			}       
		}

    lDistance = ( float ) lDepth / 4294967295.0f;
				 
    glGetDoublev( GL_MODELVIEW_MATRIX, lModelMatrix );
    glGetDoublev( GL_PROJECTION_MATRIX, lProjMatrix );
				
		gluUnProject( lMousePos.x, ( lViewPort[ 3 ] - lMousePos.y ), lDistance, 
				          lModelMatrix, lProjMatrix, lViewPort, 
 							    &lPos[ 0 ], &lPos[ 1 ], &lPos[ 2 ] );
    
    lTargetPos.x = lPos[ 0 ];
    lTargetPos.y = lPos[ 1 ];
    lTargetPos.z = lPos[ 2 ];

    // See if we can place the cannon in the castle
    if ( mCastle.check( lTargetPos, aClicked ) )
    {
      // We placed the cannon > now let's fight
      mStage = PRE_BATTLE;

      // Set the number of seconds for the pre-battle round
      mSeconds = con_warmup_time;

      // Set view pos for battle
      mViewerPos.x = 700;
      mViewerPos.y = 1200;
      mViewerPos.z = 1300;

      mLookAtPos.x = 600;
      mLookAtPos.y = 0;
      mLookAtPos.z = 600;

      mViewDir = mLookAtPos - mViewerPos;
      mViewDir.normalize();
      mZoom = 0;

      //Renable the cursor
      gCursor.enable();

      mCastle.disable_ghost();
      mCastle.disable_draw_cannon();
    }
	}
*/
  // Get the Z coordinate for the pixel location
  glReadPixels( lMousePos.x, ( lViewPort[ 3 ] - lMousePos.y ), 1, 1, 
                GL_DEPTH_COMPONENT, GL_UNSIGNED_INT, &lDepth );

  int zdepth;
  float lDistTermInv;
  glGetIntegerv( GL_DEPTH_BITS, &zdepth );
  switch ( zdepth )
  {
  case 16 :
    lDistTermInv = (float) 1.5259018967e-5;     // 65535
    break;
  default :
    lDistTermInv = (float) 2.32830643708e-10;   // 4294967295
    break;
  }

  lDistance = ( float ) lDepth * lDistTermInv;

  glGetDoublev( GL_MODELVIEW_MATRIX, lModelMatrix );
  glGetDoublev( GL_PROJECTION_MATRIX, lProjMatrix );
				
  gluUnProject( lMousePos.x, ( lViewPort[ 3 ] - lMousePos.y ), lDistance, 
                lModelMatrix, lProjMatrix, lViewPort, 
                &lPos[ 0 ], &lPos[ 1 ], &lPos[ 2 ] );
    
  lTargetPos.x = lPos[ 0 ];
  lTargetPos.y = lPos[ 1 ];
  lTargetPos.z = lPos[ 2 ];

  // See if we can place the cannon in the castle
  if ( mCastle.check( lTargetPos, aClicked ) )
  {
    // Convert the castle before the pre_battle round
    if ( con_geometry != 0 )
      mCastle.convert_pieces();

    // We placed the cannon > now let's fight
    mStage = PRE_BATTLE;

    // Set the number of seconds for the pre-battle round
    mSeconds = con_warmup_time;

    // Set view pos for battle
    mViewerPos.x = 700;
    mViewerPos.y = 1200;
    mViewerPos.z = 1300;

    mLookAtPos.x = 600;
    mLookAtPos.y = 0;
    mLookAtPos.z = 600;

    mViewDir = mLookAtPos - mViewerPos;
    mViewDir.normalize();
    mZoom = 0;
    mViewAngle = 0;

    //Renable the cursor
    gCursor.enable();

    mCastle.disable_ghost();
    mCastle.disable_draw_cannon();
  }
}

//////////////////////////////////////////////////////////
// This function draws the level models in select mode
// so we can know where the player clicked
//////////////////////////////////////////////////////////
void
game_t::draw_select_battle()
{
	int i;
  int lName = 0;

	glEnableClientState( GL_VERTEX_ARRAY );
	glEnableClientState( GL_NORMAL_ARRAY );

  // Disable texturing during select mode
	glActiveTextureARB( GL_TEXTURE0_ARB );
	glClientActiveTextureARB( GL_TEXTURE0_ARB );
	glDisable( GL_TEXTURE_2D );
	glDisableClientState( GL_TEXTURE_COORD_ARRAY );

	for ( i = 0; i < mWorldModels.size(); ++i )
  {
	  glLoadName( lName++ );
		mWorldModels[ i ].draw_pos();
  }

	glDisableClientState( GL_VERTEX_ARRAY );
	glDisableClientState( GL_NORMAL_ARRAY );  
}

//////////////////////////////////////////////////////////
// This function draws the level models in select mode
// so we can know where the player clicked
//////////////////////////////////////////////////////////
void
game_t::draw_select_build()
{
	int i;

	glEnableClientState( GL_VERTEX_ARRAY );
	glEnableClientState( GL_NORMAL_ARRAY );

	glActiveTextureARB( GL_TEXTURE0_ARB );
	glClientActiveTextureARB( GL_TEXTURE0_ARB );
	glDisable( GL_TEXTURE_2D );
	glDisableClientState( GL_TEXTURE_COORD_ARRAY );

	for ( i = 0; i < mWorldModels.size(); ++i )
  {
		mWorldModels[ i ].draw_pos();
  }

	glDisableClientState( GL_VERTEX_ARRAY );
	glDisableClientState( GL_NORMAL_ARRAY );  
}

//////////////////////////////////////////////////////////
void
game_t::player_shoot( const vect3_t &aTargetPos )
{
	static int i;
  static vect3_t lCannonPos, lDir, lCannonParticlePos;
  static projectile2_t *lpProjectile;
  static float lDist, lVel;
  static bool lFoundCannon;

  lpProjectile = find_free_projectile2();

  if ( !lpProjectile )
    return;      

  // Try to find a cannon that's ready to fire
  lFoundCannon = false;
  for ( i = 0; i < mCannons.size(); ++i )
  {
    if ( mCannons[ i ].fire() )   // returns false if cannon is unable to fire
    {
      lCannonPos = mCannons[ i ].get_pos();
      lFoundCannon = true;
      break;
    }
  }

  if ( !lFoundCannon )
    return;   // no cannon is available so do nothing

  lCannonParticlePos = lCannonPos;
  lCannonParticlePos.x -= con_block_size * 0.25f;
  lCannonParticlePos.z -= con_block_size * 0.25f;

  add_cannon_particles( lCannonParticlePos, aTargetPos );

  lpProjectile->set_pos( lCannonPos );

  lDir = aTargetPos - lCannonPos;
  lpProjectile->set_dir( lDir );

  lpProjectile->set_valid( true );

  lpProjectile->set_vel( 300 );

  lpProjectile->calc_physics();

  mpSounds->play_sample( mCannonSound );
}

//////////////////////////////////////////////////////////
void 
game_t::add_water_particles( const vect3_t &aPos )
{
  vect3_t lPos = aPos;
  //lPos.y += 10;

  // First look for a free particle system
  for ( int i = 0; i < PCOUNT; ++i )
  {
    if ( !mWaterParticles[ i ].is_enabled() )
    {
      mWaterParticles[ i ].set_pos( lPos );
      mWaterParticles[ i ].reset();
      break;
    }
  }
}

//////////////////////////////////////////////////////////
void 
game_t::add_wake_particles( ship_t &aShip, const vect3_t &aDstPos )
{
  vect3_t lPos = aShip.get_pos();
//  lPos -= aShip.get_dir() * 30.0f;
//  lPos.y += 5;

  // First look for a free particle system
  for ( int i = 0; i < mNumValidShips; ++i )
  {
    if ( !mWakeParticles[ i ].is_enabled() )
    {
      mWakeParticles[ i ].set_pos( lPos );
      mWakeParticles[ i ].reset( false );

      aShip.set_wake_particles( &mWakeParticles[ i ] );
      break;
    }
  }
}

//////////////////////////////////////////////////////////
void 
game_t::add_smoke_particles( ship_t &aShip )
{
  vect3_t lPos = aShip.get_pos();
  lPos.y += 10;

  // First look for a free particle system
  for ( int i = 0; i < PCOUNT; ++i )
  {
    if ( !mSmokeParticles[ i ].is_enabled() )
    {
      mSmokeParticles[ i ].set_pos( lPos );
      mSmokeParticles[ i ].reset();

      aShip.set_particles( &mSmokeParticles[ i ] );
      break;
    }
  }
}

//////////////////////////////////////////////////////////
void 
game_t::add_cannon_particles( const vect3_t &aSrcPos, const vect3_t &aDstPos )
{
  vect3_t lPos = aSrcPos;
  lPos.y += 20;

  // This is a hack! My particle engine has no set_dir() type function so
  // I came up with the following. If anybody has a better idea...
  vect3_t lVel;
  lVel.y = 0.08f;
  if ( aDstPos.x >= aSrcPos.x )
    lVel.x = 0.25f;
  else if ( aDstPos.x < aSrcPos.x )
    lVel.x = -0.25f;

  if ( aDstPos.z >= aSrcPos.z )
    lVel.z = 0.25f;
  else if ( aDstPos.z < aSrcPos.z )
    lVel.z = -0.25f;

  // First look for a free particle system
  for ( int i = 0; i < PCOUNT; ++i )
  {
    if ( !mCannonParticles[ i ].is_enabled() )
    {
      mCannonParticles[ i ].set_pos( lPos );
      mCannonParticles[ i ].set_vel( lVel );
      mCannonParticles[ i ].reset( false );
      break;
    }
  }
}

//////////////////////////////////////////////////////////
void 
game_t::add_explosion_particles( const vect3_t &aPos )
{
  // First look for a free particle system
  for ( int i = 0; i < PCOUNT; ++i )
  {
    if ( !mExplosionParticles[ i ].is_enabled() )
    {
      mExplosionParticles[ i ].set_pos( aPos );
      mExplosionParticles[ i ].reset();
      break;
    }
  }
}

//////////////////////////////////////////////////////////
void 
game_t::draw_particles()
{
	glDisableClientState( GL_VERTEX_ARRAY );

	glActiveTextureARB( GL_TEXTURE0_ARB );
	glClientActiveTextureARB( GL_TEXTURE0_ARB );
	glDisableClientState( GL_TEXTURE_COORD_ARRAY );
	glEnable( GL_TEXTURE_2D );
  
  glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

	glEnable( GL_BLEND );
  glDepthMask( GL_FALSE );

  for ( int i = 0; i < PCOUNT; ++i )
  {
    mWaterParticles[ i ].draw( mViewerPos );
    if ( con_wake )
      mWakeParticles[ i ].draw( mViewerPos );
    mSmokeParticles[ i ].draw( mViewerPos );
    mCannonParticles[ i ].draw( mViewerPos );
    mExplosionParticles[ i ].draw( mViewerPos );
  }
  
	glEnableClientState( GL_VERTEX_ARRAY );

	glActiveTextureARB( GL_TEXTURE0_ARB );
	glClientActiveTextureARB( GL_TEXTURE0_ARB );
	glEnableClientState( GL_TEXTURE_COORD_ARRAY );
	glDisable( GL_TEXTURE_2D );

	glDisable( GL_BLEND );
  glDepthMask( GL_TRUE );
}

//////////////////////////////////////////////////////////
void
game_t::disable_particles()
{
  for ( int i = 0; i < PCOUNT; ++i )
  {
    mWaterParticles[ i ].disable();
    mWakeParticles[ i ].disable();
    mSmokeParticles[ i ].disable();
    mCannonParticles[ i ].disable();
    mExplosionParticles[ i ].disable();
  }
}

//////////////////////////////////////////////////////////
void
game_t::draw_screen_quad()
{  
  // Draw a shadowing rectangle covering the entire screen
  glColor4f( 0.0f, 0.0f, 0.0f, 0.4f );
  glEnable( GL_BLEND );
  glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
  glStencilFunc( GL_NOTEQUAL, 0, 0xffffffff );
  glStencilOp( GL_KEEP, GL_KEEP, GL_KEEP );
   
  glMatrixMode( GL_PROJECTION );
  glPushMatrix();
  glLoadIdentity();
  glOrtho( 0, 640, 0, 480, -100, 100 );
  glMatrixMode( GL_MODELVIEW );
  glPushMatrix();
  glLoadIdentity(); 
  
  glBegin( GL_QUADS );
  glVertex2d( 0, 0 );
  glVertex2d( 640, 0 );
  glVertex2d( 640, 480 );
  glVertex2d( 0, 480 );
  glEnd();
  
  glMatrixMode( GL_PROJECTION );
  glPopMatrix();
  glMatrixMode( GL_MODELVIEW );
  glPopMatrix();
  
  glDisable( GL_BLEND );
}
