#include <windows.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <gl\gl.h>
#include <gl\glu.h>

#include "fmath.h"
#include "glext.h"
#include "timer.h"
#include "textures.h"
#include "font.h"
#include "model.h"
#include "globals.h"
#include "qsound.h"
#include "menu.h"
#include "console.h"
#include "cursor.h"
#include "ship.h"
#include "program.h"
#include "game.h"
#include "textbox.h"

#pragma warning(disable:4244)   // No warnings on precision truncation
#pragma warning(disable:4305)   // No warnings on precision truncation
#pragma warning(disable:4786)   // Remove stl warnings

HDC				hDC = NULL;			// Private GDI Device Context
HGLRC			hRC = NULL;			// Permanent Rendering Context
HWND			hWnd = NULL;		// Holds Our Window Handle
HINSTANCE	hInstance;			// Holds The Instance Of The Application

DEVMODE			DMsaved;			// Saves the previous screen settings (NEW)

bool			keys[ 256 ];		// Array Used For The Keyboard Routine
bool			active = TRUE;	// Window Active Flag Set To TRUE By Default

LRESULT	CALLBACK WndProc( HWND, UINT, WPARAM, LPARAM );	// Declaration For WndProc

int GLInit();
void GLDrawScene();
void GLDrawMenu();
void GLResize( int width, int height );
void KillGLWindow();

void SysInit();
void SysShutDown();

//////////////////////////////////////////////////////////////////////
// Console functions
//////////////////////////////////////////////////////////////////////
void con_set_res( const char *aRes );
void con_set_res( const char *aRes, bool aCycle );
void con_set_level( const char *apLevelName );
void con_set_wire( bool aWire );
void con_set_wincursor( bool );
void con_set_difficulty( bool aDifficulty );
void con_set_sndfx_vol( int aVol );
void con_set_shadows( bool aShadows );
void con_set_fullscreen( bool aFullScreen );
void con_set_filtering( int aFiltering );
void con_set_filtering( bool aToggle );
void con_set_geometry( int aDetail );
void con_set_geometry( bool aDetail );

void draw_zmglogo( int aX, int aY );
void draw_logo( int aX, int aY );

void con_restart();
void con_quit();
void con_save();

texture_manager_t gTextures;
int gEnvTexture;

timer_t gTimer;
float gCurrTick, gDiffTick, gLastTick, gAccumTick;

sound_manager_t *gSoundSystem;

font_t gFont;

// Declare a text box to receive the player's name
textbox_t *gpNameTextBox;

// Declare a global high score list so the file is only read once
score_list_t *gpScoreList;

bool gDone = false;
menu_sys_t *gpMenu;
menu_sys_t *gpMenuScores;
menu_sys_t *gpMenuOptions;

//Our vertex program definitions
GLuint gEnvVP;

float gScale = 1;

model_t logo3d;

program_manager_t gProgramManager;

game_t gGame;

//////////////////////////////////////////////////////////////////////
void
SysShutDown()
{
  delete gSoundSystem;
	delete gCon;
	delete gLog;
  delete gpMenu;
  delete gpMenuScores;
  delete gpMenuOptions;
  delete gpNameTextBox;
  // gDebug and gpScoreList cannot be deleted here because they are used in destructors;
}

//////////////////////////////////////////////////////////////////////
void
SysInit()
{
  gLog = new log_t( "invasion.log" );
  gDebug = new log_t( "debug.log" );

	gCon = new console_t;

	gSoundSystem = new sound_manager_t;

  // Setup the main menu page.  Keep the enum constants up to date in globals.h
  gpMenu = new menu_sys_t( &gFont, gSoundSystem );
  gpMenu->add_entry( "START NEW GAME" );
  gpMenu->add_entry( "RESUME GAME" );
  gpMenu->add_entry( "OPTIONS" );
  gpMenu->add_entry( "VIEW HIGH SCORES" );
  gpMenu->add_entry( "EXIT" );

  // Setup the high score page
  gpMenuScores = new menu_sys_t( &gFont, gSoundSystem );
  gpMenuScores->disable();
  gpMenuScores->add_entry( "BACK TO MAIN MENU" );

  // Setup the options menu page
  gpMenuOptions = new menu_sys_t( &gFont, gSoundSystem );
  gpMenuOptions->disable();
  gpMenuOptions->add_entry( "DIFFICULTY:    EASY" );
  gpMenuOptions->add_entry( "SHADOWS:       ON" );
  gpMenuOptions->add_entry( "FULL SCREEN:   OFF" );
  gpMenuOptions->add_entry( "RESOLUTION:    640x480" );
  gpMenuOptions->add_entry( "FILTERING:     FAST" );
  gpMenuOptions->add_entry( "GEOMETRY:      MED DETAIL" );
  gpMenuOptions->add_entry( "BACK TO MAIN MENU" );

  gpNameTextBox = new textbox_t;
  gpScoreList = new score_list_t;

  gCon->reg_bool( "fullscreen", &con_fullscreen, false, HD_NORMAL, con_set_fullscreen );

	gCon->reg_bool( "debug", &con_debug, false, HD_NORMAL );

	gCon->reg_bool( "ai", &con_ai, false, HD_NORMAL );
  gCon->reg_float( "block_size", &con_block_size, false, HD_NORMAL );
  gCon->reg_string( "res", &con_res, false, HD_NORMAL, con_set_res );
  gCon->reg_int( "sndfx_vol", &con_sndfx_vol, false, HD_NORMAL, con_set_sndfx_vol );
  gCon->reg_string( "level", &con_level, false, HD_GL, con_set_level );
  gCon->reg_bool( "difficulty", &con_difficulty, false, HD_NORMAL, con_set_difficulty );
  gCon->reg_bool( "difficulty_next", &con_difficulty_next, false, HD_NORMAL, con_set_difficulty );

	gCon->reg_bool( "draw_debug", &con_draw_debug, false, HD_NORMAL );
	gCon->reg_bool( "shadows", &con_shadows, false, HD_NORMAL, con_set_shadows );
	gCon->reg_bool( "light", &con_light, false, HD_NORMAL );
	gCon->reg_int( "geometry", &con_geometry, false, HD_NORMAL, con_set_geometry );
	gCon->reg_bool( "wire", &con_wire, false, HD_NORMAL, con_set_wire );
	gCon->reg_bool( "wincursor", &con_wincursor, false, HD_NORMAL, con_set_wincursor );

  gCon->reg_bool( "wake", &con_wake, false, HD_NORMAL );
  gCon->reg_int( "filtering", &con_filtering, false, HD_NORMAL, con_set_filtering );

  gCon->reg_float( "znear", &con_znear, false, HD_NORMAL );
  gCon->reg_float( "zfar", &con_zfar, false, HD_NORMAL );

  gCon->reg_int( "ship_health", &con_ship_health, false, HD_NORMAL );
  gCon->reg_int( "num_start_ships", &con_num_start_ships, false, HD_NORMAL );  
  
  gCon->reg_int( "warmup_time", &con_warmup_time, false, HD_NORMAL );
  gCon->reg_int( "battle_time", &con_battle_time, false, HD_NORMAL );
  gCon->reg_int( "build_time", &con_build_time, false, HD_NORMAL );
  gCon->reg_int( "cannon_time", &con_cannon_time, false, HD_NORMAL );
  
  gCon->reg_float( "ship_shoot_count", &con_ship_shoot_count, false, HD_NORMAL );
  gCon->reg_float( "cannon_shoot_time", &con_cannon_shoot_time, false, HD_NORMAL );
  
  gCon->reg_cmd( "restart", con_restart );
  gCon->reg_cmd( "quit", con_quit );
  gCon->reg_cmd( "save", con_save );

  gCon->exec_file( "options.conf" );

  gCon->exec_file( "invasion.conf" );    // override options settings if necessary

  con_difficulty_next = con_difficulty;   // make sure these start off the same

}

///////////////////////////////////////////////////////////////////////
void 
GLResize( int width, int height )
{
	if ( height == 0 )
	{
		height = 1;
	}

	glViewport( 0, 0, width, height );

	gCursor.set_dim( width, height );

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();

	// Calculate The Aspect Ratio Of The Window
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	gluPerspective( 45.0f, ( float ) width / ( float ) height, 
                  con_znear, con_zfar );

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
}

///////////////////////////////////////////////////////////////////////
int 
GLInit()
{  
  // Seed the random number generator
  srand( ( unsigned ) time( NULL ) );

  gCon->add_msg( "Initializing rendering system" );

  gCon->init( &gTextures, &gFont );

  gFont.build( &gTextures );

  gCrosshairTexture = gTextures.insert( "crosshair.bmp", texture_manager_t::CLAMP );
  gCursor.set_texture( gCrosshairTexture );

  // Intialize our game system
  gGame.set_sound_manager( gSoundSystem );
  gGame.set_texture_manager( &gTextures );
  gGame.set_font( &gFont );

  char msg[ 200 ];
  strcpy( msg, "OpenGL vendor: " );
  strcat( msg, ( char * ) glGetString( GL_VENDOR ) );
  gCon->add_msg( msg );

  strcpy( msg, "OpenGL version: " );
  strcat( msg, ( char * ) glGetString( GL_VERSION ) );
  gCon->add_msg( msg );

  if ( !init_multitexture_arb() || !init_compiled_vertex_array_ext() )
  {
    gCon->add_msg( "Failed to initialize required extensions" );
	  exit( -1 );
  }
  
  //Initialize ARB VP
  if ( !init_vertex_program_arb() )
  {
	  gCon->add_msg( "No Vertex Program support" );
	  gVertexProgram = false;
  }    
  else
  {
	  gCon->add_msg( "ARB Vertex Program enabled" );
    gVertexProgram = true;
    gEnvVP = gProgramManager.insert( "env.vp", GL_VERTEX_PROGRAM_ARB );
  }

  gSeparateStencilATI = init_separate_stencil_ati();
  gStencilTwoSideEXT = init_stencil_two_side_ext();

  glEnable( GL_CULL_FACE );
  glFrontFace( GL_CCW );
  glCullFace( GL_BACK );
  
  glClearColor( 0, 0, 0, 0 );
  glClearDepth( 1 );
  glClearStencil( 0 );
  glDepthFunc( GL_LESS );
  glEnable( GL_DEPTH_TEST );

  glShadeModel( GL_SMOOTH );

  glPointSize( 5 );

  // Load Textures
  gParticleTexture = gTextures.insert( "proj.bmp" );
  gZMGLogoTexture = gTextures.insert( "zmglogo.bmp" );
  gLogoTexture = gTextures.insert( "logo.bmp" );

  gCon->exec_delay_handlers();
  gCon->add_msg( "End rendering initialization" );

  return true;
}

///////////////////////////////////////////////////////////////////////
void 
GLDrawScene()	
{	
  if ( con_wire )
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT );
  else
    glClear( GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT );

  // draw must come before update because glReadPixels is being used in game.cpp
  gGame.draw();
  gGame.update();

  static char lStr[ 100 ];
  gFont.pre();

  gGame.draw_text();

  gCursor.draw();

  gCon->draw();
  gFont.post();
}

///////////////////////////////////////////////////////////////////////
void 
GLDrawMenu()	
{	
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

  gFont.pre();

  if ( gpMenuScores->is_enabled() )
    gpMenuScores->draw_scores();
  else if ( gpMenuOptions->is_enabled() )
    gpMenuOptions->draw_options();
  else
    gpMenu->draw();

  // Draw the zmgames logo
  draw_zmglogo( 505, 30 );

  // Draw the Invasion logo
  if ( gpMenu->is_enabled() )
    draw_logo( 200, 288 );

  gCon->draw();

  gFont.post();
}

////////////////////////////////////////////////////////////////
void 
draw_zmglogo( int aX, int aY )
{
  glColor3f( 1.0, 1.0, 1.0 );
  glBindTexture( GL_TEXTURE_2D, gZMGLogoTexture );

  glBegin( GL_QUADS );

    glTexCoord2f( 0, 0 ); 
    glVertex2f( aX, aY );
    
    glTexCoord2f( 0, 1 ); 
    glVertex2f( aX, aY + 115 );

    glTexCoord2f( 1, 1 ); 
    glVertex2f( aX + 100, aY + 115 );

    glTexCoord2f( 1, 0 ); 
    glVertex2f( aX + 100, aY );  
        
  glEnd();
}

////////////////////////////////////////////////////////////////
void 
draw_logo( int aX, int aY )
{
  glColor3f( 1.0, 1.0, 1.0 );
  glBindTexture( GL_TEXTURE_2D, gLogoTexture );

  glBegin( GL_QUADS );

    glTexCoord2f( 0, 0 ); 
    glVertex2f( aX, aY );
    
    glTexCoord2f( 0, 1 ); 
    glVertex2f( aX, aY + 180 );

    glTexCoord2f( 1, 1 ); 
    glVertex2f( aX + 240, aY + 180 );

    glTexCoord2f( 1, 0 ); 
    glVertex2f( aX + 240, aY );  
    
  glEnd();
}

////////////////////////////////////////////////////////////////
void 
con_restart()
{
  gGame.restart();
}

////////////////////////////////////////////////////////////////
// Saves all console variables to the config file
////////////////////////////////////////////////////////////////
void 
con_save()
{
  gCon->save_file( "invasion.conf" );
}

////////////////////////////////////////////////////////////////
void
con_quit()
{
  // Save the option settings
  con_difficulty = con_difficulty_next;
  gCon->save_options( "options.conf" );
  
  SysShutDown();

	// Shutdown
	KillGLWindow();
}

////////////////////////////////////////////////////////////////
void 
con_set_level( const char *apLevelName )
{
  gGame.load_level( apLevelName );
}

////////////////////////////////////////////////////////////////
void 
con_set_wire( bool aWire )
{
  gWireFrame = ( con_wire ) ? GL_LINE : GL_FILL;
}

////////////////////////////////////////////////////////////////
void 
con_set_shadows( bool aShadows )
{
  string lStr = "SHADOWS:       ";
  lStr += ( aShadows ) ? "ON" : "OFF";

  gpMenuOptions->set_entry( OPTION_SHADOWS, lStr );
}

////////////////////////////////////////////////////////////////
void 
con_set_wincursor( bool aCursor )
{
	ShowCursor( con_wincursor );
}

////////////////////////////////////////////////////////////////
void 
con_set_difficulty( bool aDifficulty )
{
  string lStr = "DIFFICULTY:    ";
  lStr += ( aDifficulty ) ? "HARD" : "EASY";

  gpMenuOptions->set_entry( OPTION_DIFF, lStr );
}

////////////////////////////////////////////////////////////////
void
con_set_sndfx_vol( int aVol )
{
  if ( aVol < 0 )
    con_sndfx_vol = 0;
  else if ( aVol > sound_manager_t::MAX_VOLUME )
    con_sndfx_vol = sound_manager_t::MAX_VOLUME;

  gSoundSystem->set_volume( aVol );
}

////////////////////////////////////////////////////////////////
void 
con_set_fullscreen( bool aFullScreen )
{
  string lStr = "FULL SCREEN:   ";
  lStr += ( aFullScreen ) ? "ON" : "OFF";

  gpMenuOptions->set_entry( OPTION_FS, lStr );
  gCon->add_instant_msg( "Please reload Invasion to activate this change" );
}

////////////////////////////////////////////////////////////////
void 
con_set_res( const char *aRes )
{	
  con_set_res( aRes, false );
}

////////////////////////////////////////////////////////////////
void 
con_set_res( const char *aRes, bool aCycle )
{	
  int lWidth, lHeight;
	int count = sscanf( aRes, "%dx%d", &lWidth, &lHeight );

	if ( count != 2 )
	{
		gLog->add_msg( "ERROR: con_set_res: WidthxHeight" );
		return;
	}

	switch ( lWidth )
	{
	case 320 :
		if ( lHeight != 240 )
		{
			gLog->add_msg( "ERROR: con_set_res: Invalid height %d", lHeight );
			return;
		}
		break;
	case 640 :
		if ( lHeight != 480 )
		{
			gLog->add_msg( "ERROR: con_set_res: Invalid height %d", lHeight );
			return;
		}
		break;
	case 800 :
		if ( lHeight != 600 )
		{
			gLog->add_msg( "ERROR: con_set_res: Invalid height %d", lHeight );
			return;
		}
		break;
	case 1024 :
		if ( lHeight != 768 )
		{
			gLog->add_msg( "ERROR: con_set_res: Invalid height %d", lHeight );
			return;
		}
		break;
	case 1280 :
		if ( lHeight != 1024 && lHeight != 960 )
		{
			gLog->add_msg( "ERROR: con_set_res: Invalid height %d", lHeight );
			return;
		}
		break;
	case 1600 :
		if ( lHeight != 1200 )
		{
			gLog->add_msg( "ERROR: con_set_res: Invalid height %d", lHeight );
			return;
		}
		break;
	default :
		gLog->add_msg( "ERROR: con_set_res: Invalid width %d", lWidth );
		return;
	}

  // Cycle to the next resolution
  if ( aCycle )
  {
    switch ( lWidth )
    {
    case 320 :
      lWidth = 640;
      lHeight = 480;
      break;
    case 640 :
      lWidth = 800;
      lHeight = 600;
      break;
    case 800 :
      lWidth = 1024;
      lHeight = 768;
      break;
    case 1024 :
      lWidth = 1280;
      lHeight = 960;
      break;
    case 1280 :
      if ( lHeight == 960 )
        lHeight = 1024;
      else
      {
        lWidth = 1600;
        lHeight = 1200;
      }
      break;
    case 1600 :
      lWidth = 320;
      lHeight = 240;
      break;
    }
  }

  string lStr = "RESOLUTION:    ";
  char lsWidth[5], lsHeight[5];
  sprintf( lsWidth, "%d", lWidth );
  sprintf( lsHeight, "%d", lHeight );
  lStr += lsWidth;
  lStr += "x";
  lStr += lsHeight;

  gpMenuOptions->set_entry( OPTION_RES, lStr );
  gCon->add_instant_msg( "Please reload Invasion to activate this change" );
  strcpy( con_res, lsWidth );
  strcat( con_res, "x" );
  strcat( con_res, lsHeight );

  if ( !aCycle )
  {
    gScreenWidth = lWidth;
	  gScreenHeight = lHeight;

    gCursor.set_dim( gScreenWidth, gScreenHeight );
  }
}

////////////////////////////////////////////////////////////////
void 
con_set_filtering( int aFiltering )
{
  string lStr = "FILTERING:     ";

  if ( aFiltering == 1 )
    lStr += "QUALITY";
  else
    lStr += "FAST";

  con_filtering = aFiltering;

  gTextures.set_filtering( aFiltering );

  gpMenuOptions->set_entry( OPTION_FILTER, lStr );
}

////////////////////////////////////////////////////////////////
void 
con_set_filtering( bool aToggle )
{
  int lFiltering;

  lFiltering = con_filtering;

  if ( aToggle )
  {
    if ( lFiltering == 1 )
      lFiltering = 0;
    else
      lFiltering = 1;
  }

  string lStr = "FILTERING:     ";

  if ( lFiltering == 1 )
    lStr += "QUALITY";
  else
    lStr += "FAST";

  gTextures.set_filtering( lFiltering );
  con_filtering = lFiltering;

  gpMenuOptions->set_entry( OPTION_FILTER, lStr );
  gCon->add_instant_msg( "Please reload Invasion to activate this change" );
}

////////////////////////////////////////////////////////////////
void 
con_set_geometry( int aDetail )
{
  string lStr = "GEOMETRY:      ";

  if ( aDetail == 0 )
    lStr += "LOW DETAIL";
  if ( aDetail == 1 )
    lStr += "MED DETAIL";
  if ( aDetail == 2 )
    lStr += "HIGH DETAIL";

  con_geometry = aDetail;

  gpMenuOptions->set_entry( OPTION_GEOMETRY, lStr );
}

////////////////////////////////////////////////////////////////
void 
con_set_geometry( bool aDetail )
{
  int lDetail = 2;

  if ( aDetail )
  {
    if ( con_geometry == 0 )
      lDetail = 1;
    if ( con_geometry == 1 )
      lDetail = 2;
    if ( con_geometry == 2 )
      lDetail = 0;
  }

  string lStr = "GEOMETRY:      ";

  if ( lDetail == 0 )
    lStr += "LOW DETAIL";
  if ( lDetail == 1 )
    lStr += "MED DETAIL";
  if ( lDetail == 2 )
    lStr += "HIGH DETAIL";

  con_geometry = lDetail;

  gpMenuOptions->set_entry( OPTION_GEOMETRY, lStr );
}

/************************************************************************************/
void 
KillGLWindow()
{
	ShowCursor( true );
  
  if ( con_fullscreen )
	{
		if ( !ChangeDisplaySettings( NULL, CDS_TEST ) ) 
		{
			ChangeDisplaySettings( NULL, CDS_RESET );
			ChangeDisplaySettings( &DMsaved, CDS_RESET );
		} 
		else 
		{
			ChangeDisplaySettings( NULL, CDS_RESET );
		}
	}

	if ( hRC )
	{
		if ( !wglMakeCurrent( NULL, NULL ) )
		{
			MessageBox( NULL, "Release Of DC And RC Failed.", "SHUTDOWN ERROR", 
				          MB_OK | MB_ICONINFORMATION );
		}

		if ( !wglDeleteContext( hRC ) )
		{
			MessageBox( NULL, "Release Rendering Context Failed.", "SHUTDOWN ERROR",
				          MB_OK | MB_ICONINFORMATION );
		}
		hRC = NULL;
	}

	if ( hDC && !ReleaseDC( hWnd, hDC ) )
	{
		MessageBox( NULL, "Release Device Context Failed.",
			          "SHUTDOWN ERROR", MB_OK | MB_ICONINFORMATION );
		hDC = NULL;
	}

	if ( hWnd && !DestroyWindow( hWnd ) )
	{
		MessageBox( NULL, "Could Not Release hWnd.", "SHUTDOWN ERROR",
			          MB_OK | MB_ICONINFORMATION );
		hWnd = NULL;
	}

	if ( !UnregisterClass( "OpenGL", hInstance ) )
	{
		MessageBox( NULL, "Could Not Unregister Class.", "SHUTDOWN ERROR",
			          MB_OK | MB_ICONINFORMATION );
		hInstance = NULL;
	}
}

/************************************************************************************/

/*	This Code Creates Our OpenGL Window.  Parameters Are:					*
 *	title			- Title To Appear At The Top Of The Window				*
 *	width			- Width Of The GL Window Or Fullscreen Mode				*
 *	height			- Height Of The GL Window Or Fullscreen Mode			*
 *	bits			- Number Of Bits To Use For Color (8/16/24/32)			*
 *	fullscreenflag	- Use Fullscreen Mode (TRUE) Or Windowed Mode (FALSE)	*/ 
BOOL 
CreateGLWindow( char* title, int width, int height, int bits, bool fullscreen )
{
	GLuint		PixelFormat;			// Holds The Results After Searching For A Match
	WNDCLASS	wc;						// Windows Class Structure
	DWORD		dwExStyle;				// Window Extended Style
	DWORD		dwStyle;				// Window Style
	RECT		WindowRect;				// Grabs Rectangle Upper Left / Lower Right Values
	WindowRect.left=(long)0;			// Set Left Value To 0
	WindowRect.right=(long)width;		// Set Right Value To Requested Width
	WindowRect.top=(long)0;				// Set Top Value To 0
	WindowRect.bottom=(long)height;		// Set Bottom Value To Requested Height

	hInstance			= GetModuleHandle(NULL);				// Grab An Instance For Our Window
	wc.style			= CS_HREDRAW | CS_VREDRAW | CS_OWNDC;	// Redraw On Size, And Own DC For Window.
	wc.lpfnWndProc		= (WNDPROC) WndProc;					// WndProc Handles Messages
	wc.cbClsExtra		= 0;									// No Extra Window Data
	wc.cbWndExtra		= 0;									// No Extra Window Data
	wc.hInstance		= hInstance;							// Set The Instance
	wc.hIcon			= LoadIcon(NULL, IDI_WINLOGO);			// Load The Default Icon
	wc.hCursor			= LoadCursor(NULL, IDC_ARROW);			// Load The Arrow Pointer
	wc.hbrBackground	= NULL;									// No Background Required For GL
	wc.lpszMenuName		= NULL;									// We Don't Want A Menu
	wc.lpszClassName	= "OpenGL";								// Set The Class Name
	
	EnumDisplaySettings(NULL, ENUM_CURRENT_SETTINGS, &DMsaved); // save the current display state (NEW)

	if (fullscreen)												// Attempt Fullscreen Mode?
	{
		DEVMODE dmScreenSettings;								// Device Mode
		memset(&dmScreenSettings,0,sizeof(dmScreenSettings));	// Makes Sure Memory's Cleared
		dmScreenSettings.dmSize=sizeof(dmScreenSettings);		// Size Of The Devmode Structure
		dmScreenSettings.dmPelsWidth	= width;				// Selected Screen Width
		dmScreenSettings.dmPelsHeight	= height;				// Selected Screen Height
		dmScreenSettings.dmBitsPerPel	= bits;					// Selected Bits Per Pixel
		dmScreenSettings.dmFields=DM_BITSPERPEL|DM_PELSWIDTH|DM_PELSHEIGHT;

		// Try To Set Selected Mode And Get Results.  NOTE: CDS_FULLSCREEN Gets Rid Of Start Bar.
		if ( ChangeDisplaySettings( &dmScreenSettings, CDS_FULLSCREEN ) != 
			                        DISP_CHANGE_SUCCESSFUL )
		{
			// If The Mode Fails, Offer Two Options.  Quit Or Use Windowed Mode.
			if ( MessageBox( NULL,"The Requested Fullscreen Mode Is Not Supported By\nYour Video Card. Use Windowed Mode Instead?","NeHe GL",
				             MB_YESNO | MB_ICONEXCLAMATION ) == IDYES )
			{
				fullscreen = FALSE;
			}
			else
			{
				// Pop Up A Message Box Letting User Know The Program Is Closing.
				MessageBox( NULL, "Program Will Now Close.", "ERROR", 
					        MB_OK | MB_ICONSTOP );
				return FALSE;
			}
		}
	}

	if ( !RegisterClass( &wc ) )
	{
		MessageBox( NULL, "Failed To Register The Window Class.", "ERROR",
			        MB_OK | MB_ICONEXCLAMATION );
		return FALSE;
	}

	if ( fullscreen )
	{
		dwExStyle = WS_EX_APPWINDOW;
		dwStyle = WS_POPUP;
		ShowCursor( FALSE );
	}
	else
	{
		dwExStyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
		dwStyle = WS_OVERLAPPEDWINDOW;
	}

	AdjustWindowRectEx( &WindowRect, dwStyle, FALSE, dwExStyle );

	// Create The Window
	if ( !( hWnd = CreateWindowEx( dwExStyle,							// Extended Style For The Window
								 "OpenGL",							// Class Name
								 title,								  // Window Title
								 dwStyle |							// Defined Window Style
								 WS_CLIPSIBLINGS |			// Required Window Style
								 WS_CLIPCHILDREN,				// Required Window Style
								 0, 0,								              // Window Position
								 WindowRect.right-WindowRect.left,	// Calculate Window Width
								 WindowRect.bottom-WindowRect.top,	// Calculate Window Height
								 NULL,								  // No Parent Window
								 NULL,								  // No Menu
								 hInstance,							// Instance
								 NULL ) ) )							// Dont Pass Anything To WM_CREATE
	{		
		KillGLWindow();
		MessageBox( NULL, "Window Creation Error.", "ERROR",
			          MB_OK | MB_ICONEXCLAMATION );
		return FALSE;
	}

	static	PIXELFORMATDESCRIPTOR pfd=			// pfd Tells Windows How We Want Things To Be
	{
		sizeof(PIXELFORMATDESCRIPTOR),				// Size Of This Pixel Format Descriptor
		1,											// Version Number
		PFD_DRAW_TO_WINDOW |		// Format Must Support Window
		PFD_SUPPORT_OPENGL |		// Format Must Support OpenGL
		PFD_DOUBLEBUFFER,				// Must Support Double Buffering
		PFD_TYPE_RGBA,					// Request An RGBA Format
		bits,										// Select Our Color Depth
		0, 0, 0, 0, 0, 0,				// Color Bits Ignored
		0,											// No Alpha Buffer
		0,											// Shift Bit Ignored
		0,											// No Accumulation Buffer
		0, 0, 0, 0,							// Accumulation Bits Ignored
		24,											// 24Bit Z-Buffer (Depth Buffer)  
		8,											// 8Bit Stencil Buffer
		0,											// No Auxiliary Buffer
		PFD_MAIN_PLANE,					// Main Drawing Layer
		0,											// Reserved
		0, 0, 0									// Layer Masks Ignored
	};
	
	if ( !( hDC = GetDC( hWnd ) ) )
	{
		KillGLWindow();
		MessageBox( NULL,"Can't Create A GL Device Context.", "ERROR",
			          MB_OK | MB_ICONEXCLAMATION );
		return FALSE;
	}

	if ( !( PixelFormat = ChoosePixelFormat( hDC, &pfd ) ) )
	{
		KillGLWindow();
		MessageBox( NULL, "Can't Find A Suitable PixelFormat.", "ERROR",
			          MB_OK | MB_ICONEXCLAMATION );
		return FALSE;
	}

	if( !SetPixelFormat( hDC, PixelFormat, &pfd ) )
	{
		KillGLWindow();
		MessageBox( NULL, "Can't Set The PixelFormat.", "ERROR",
			          MB_OK | MB_ICONEXCLAMATION );
		return FALSE;
	}

	if ( !( hRC = wglCreateContext( hDC ) ) )
	{
		KillGLWindow();
		MessageBox( NULL, "Can't Create A GL Rendering Context.", "ERROR",
			          MB_OK | MB_ICONEXCLAMATION );
		return FALSE;
	}

	if( !wglMakeCurrent( hDC,hRC ) )
	{
		KillGLWindow();
		MessageBox( NULL, "Can't Activate The GL Rendering Context.", "ERROR",
			          MB_OK | MB_ICONEXCLAMATION );
		return FALSE;
	}

	ShowWindow( hWnd, SW_SHOW );
	SetForegroundWindow( hWnd );
	SetFocus( hWnd );
	GLResize( width, height );
	ShowCursor( con_wincursor );

	if ( !GLInit() )
	{
		KillGLWindow();
		MessageBox( NULL, "Initialization Failed.", "ERROR",
			          MB_OK | MB_ICONEXCLAMATION );
		return FALSE;
	}

	return TRUE;
}

/************************************************************************************/
// (no changes)

LRESULT CALLBACK WndProc(	HWND	hWnd,			// Handle For This Window
							UINT	uMsg,			// Message For This Window
							WPARAM	wParam,			// Additional Message Information
							LPARAM	lParam)			// Additional Message Information
{
	switch (uMsg)									// Check For Windows Messages
	{
		case WM_ACTIVATE:							// Watch For Window Activate Message
		{
			if (!HIWORD(wParam))					// Check Minimization State
			{
				active=TRUE;						// Program Is Active
			}
			else
			{
				active=FALSE;						// Program Is No Longer Active
			}

			return 0;								// Return To The Message Loop
		}

		case WM_SYSCOMMAND:							// Intercept System Commands
		{
			switch (wParam)							// Check System Calls
			{
				case SC_SCREENSAVE:					// Screensaver Trying To Start?
				case SC_MONITORPOWER:				// Monitor Trying To Enter Powersave?
				return 0;							// Prevent From Happening
			}
			break;									// Exit
		}

		case WM_CLOSE:								// Did We Receive A Close Message?
		{
			PostQuitMessage(0);						// Send A Quit Message
			return 0;								// Jump Back
		}

		case WM_CHAR :
		{
			if ( gCon->is_enabled() )
			{
				gCon->add_char( wParam );
				return 0;
			}			

      if ( gpNameTextBox->is_enabled() )
      {
        if ( gpNameTextBox->add_char( wParam ) )
          gpScoreList->set_name_entered();
        return 0;
      }

			switch ( wParam )
			{  
				case 'w' :
				case 'W' :
          gCon->toggle( "wire" );
					break;

				case '`' :
				case '~' :
					gCon->toggle();
          break;

				case 'd' :
          gCon->toggle( "draw_debug" );
					break;

				case 'k' :
          gCon->toggle( "wincursor" );
					break;

				case 'p' :
				case 'P' :
          gGame.toggle_pause();
					break;

				case 'l' :
				case 'L' :
          gCon->toggle( "light" );
					break;

//				case 'c' :
//				case 'C' :
//          gGame.clear_castle();
//					break;

//				case 'r' :
//				case 'R' :
//          gGame.restart();

				case 's' :
				case 'S' :
          gCon->toggle( "shadows" );
					break;
			}
			return 0;
		}

		case WM_KEYDOWN :
		{
			if ( !gCon->is_enabled() )
				keys[ wParam ] = TRUE;

			return 0;
		}

		case WM_KEYUP :
		{
			if ( gCon->is_enabled() )
      {
				gCon->add( wParam );
        return 0;
      }

      if ( gpNameTextBox->is_enabled() )
      {
        if ( gpNameTextBox->add( wParam ) )
          gpScoreList->set_name_entered();
        return 0;
      }

			switch ( wParam )
			{
			case VK_UP :
				if ( gpMenu->is_enabled() )
					gpMenu->key_up();
				if ( gpMenuScores->is_enabled() )
					gpMenuScores->key_up();
				if ( gpMenuOptions->is_enabled() )
					gpMenuOptions->key_up();
				break;

			case VK_DOWN :
				if ( gpMenu->is_enabled() )
					gpMenu->key_down();
				if ( gpMenuScores->is_enabled() )
					gpMenuScores->key_down();
				if ( gpMenuOptions->is_enabled() )
					gpMenuOptions->key_down();
				break;

			case VK_RETURN :
			case VK_SPACE :
				if ( gpMenu->is_enabled() )
				{
					switch ( gpMenu->get_entry() )
					{
					case MENU_PLAY :
						// PLAY GAME
						gpMenu->disable();
            gpMenu->set_game_started( true );
            gGame.restart();
						break;

          case MENU_RESUME :
            // RESUME GAME
            if ( gpMenu->in_progress() )
              gpMenu->disable();
            break;

					case MENU_OPTIONS :
						// OPTIONS
            gpMenuOptions->enable();
            gpMenu->disable();
						break;

					case MENU_SCORES : 
						// VIEW HIGH SCORES
            gpMenuScores->enable();
            gpMenu->disable();
						break;

					case MENU_EXIT : 
						// EXIT
						gDone = true;
						break;
					}
				}
        else if ( gpMenuScores->is_enabled() )
        {
          switch ( gpMenuScores->get_entry() )
          {
          case 0 :
            // BACK TO MAIN MENU
          default :
            gpMenu->enable();
            gpMenuScores->disable();
            break;
          }
        }
        else if ( gpMenuOptions->is_enabled() )
        {
          switch ( gpMenuOptions->get_entry() )
          {
          case OPTION_DIFF :
            // DIFFICULTY
            gCon->toggle( "difficulty_next" );
            if ( !gpMenu->in_progress() )
              con_difficulty = con_difficulty_next;
            break;

          case OPTION_SHADOWS :
            // SHADOWS
            gCon->toggle( "shadows" );
            break;

          case OPTION_FS :
            // FULL SCREEN
            gCon->toggle( "fullscreen" );
            break;

          case OPTION_RES :
            // RESOLUTION
            con_set_res( con_res, true );
            break;

          case OPTION_FILTER :
            // FILTERING
            con_set_filtering( true );    // next
            break;

          case OPTION_GEOMETRY :
            // GEOMETRY
            con_set_geometry( true );     // next
            break;

          default :
            gpMenu->enable();
            gpMenuOptions->disable();
          }
        }
				break;

			case VK_ESCAPE :
        if ( gpMenu->in_progress() )
          gpMenu->set_curr_entry( MENU_RESUME );
				gpMenu->enable();
        if ( gpMenuScores->is_enabled() )
          gpMenuScores->disable();
        if ( gpMenuOptions->is_enabled() )
          gpMenuOptions->disable();
				break;
			}

			keys[ wParam ] = FALSE;
			return 0;
		}

		case WM_SIZE:								// Resize The OpenGL Window
		{
			GLResize( LOWORD( lParam ), HIWORD( lParam ) );
			return 0;								// Jump Back
		}	
		
		case WM_LBUTTONUP :
		{			
      gGame.handle_left_mouse( true );
			return 0;
		}
		
		case WM_RBUTTONUP :
		{			
      gGame.handle_right_mouse( true );
			return 0;
		}
	}

	// Pass All Unhandled Messages To DefWindowProc
	return DefWindowProc(hWnd,uMsg,wParam,lParam);
}

/************************************************************************************/

int WINAPI WinMain(	HINSTANCE	hInstance,
					HINSTANCE	hPrevInstance,
					LPSTR		lpCmdLine,
					int			nCmdShow )
{
	MSG		msg;
	
	SysInit();

	// Create Our OpenGL Window
	if ( !CreateGLWindow( "Invasion: Repel and Rebuild", gScreenWidth, gScreenHeight, 32, 
		                    con_fullscreen ) )
	{
		return 0;									// Quit If Window Was Not Created
	}

  // This check is here because some older cards don't return higher than a 16 bit
  // z buffer by default
  int zdepth;
  glGetIntegerv( GL_DEPTH_BITS, &zdepth );
  if ( zdepth != 24 || zdepth != 32 )
    gLog->add_msg( "incorrect zdepth = %d", zdepth );

	while ( !gDone )
	{
		if ( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ) )
		{
			if ( msg.message == WM_QUIT )
			{
				gDone = true;
			}
			else
			{
				TranslateMessage( &msg );
				DispatchMessage( &msg );
			}
		}
		else
		{
			if ( gpMenu->is_enabled() || gpMenuScores->is_enabled() || gpMenuOptions->is_enabled() )
      {
				GLDrawMenu();
      }
			else
      {
				GLDrawScene();
			
/*			  if ( keys[ VK_RIGHT ] )
			  {
          gGame.rotate_right();
			  }
			  if ( keys[ VK_LEFT ] )
			  {
          gGame.rotate_left();
			  }
*/
  			if ( keys[ VK_UP ] )
  			{
          gGame.zoom_in();
  			}
  			if ( keys[ VK_DOWN ] )
  			{
          gGame.zoom_out();
  			}
      }

			gCurrTick = gTimer.tick();
			gDiffTick	= gCurrTick - gLastTick;
			gFactTick = gDiffTick * 100.0;
			gLastTick = gCurrTick;

			gAccumTick += gDiffTick;
			
			SwapBuffers( hDC );
		}
	}

  con_quit();

	return ( msg.wParam );
}
