#include "menu.h"

////////////////////////////////////////////////////
menu_sys_t::menu_sys_t( font_t *apFont, sound_manager_t *apSound ) : 
	mEnabled( true ), 
  mGameStarted( false ),
	mCurrEntry( 0 ), 
	mpFont( apFont ), 
	mpSound( apSound ) 
{
	assert( apFont );

	assert( apSound );
	mSwitchSound = apSound->load_sample( "switchtarget.wav" );
	mClickSound = apSound->load_sample( "click_enter.wav" );
}

////////////////////////////////////////////////////
void
menu_sys_t::key_up()
{ 
	if ( mCurrEntry > 0 )
		mCurrEntry--;
	else
		mCurrEntry = mEntries.size() - 1;

	mpSound->play_sample( mSwitchSound );
}
	
////////////////////////////////////////////////////
int 
menu_sys_t::get_entry() const 
{ 
	mpSound->play_sample( mClickSound );
	return mCurrEntry; 
}
	
////////////////////////////////////////////////////
const string &
menu_sys_t::get_entry_string() const 
{ 
	mpSound->play_sample( mClickSound );
	return mEntries[ mCurrEntry ]; 
}

////////////////////////////////////////////////////
void 
menu_sys_t::set_entry( int aNum, const string &aEntry )
{
  if ( aNum < 0 || aNum >= mEntries.size() )
    return;

  mEntries[ aNum ] = aEntry;
}

////////////////////////////////////////////////////
void 
menu_sys_t::set_curr_entry( int aEntry )
{
  mCurrEntry = aEntry;
}

////////////////////////////////////////////////////
void
menu_sys_t::key_down()
{ 
	if ( mCurrEntry < ( mEntries.size() - 1 ) )
		mCurrEntry++;
	else
		mCurrEntry = 0;

	mpSound->play_sample( mSwitchSound );
}

////////////////////////////////////////////////////
void
menu_sys_t::draw()
{
	mpFont->setDim( 12, 12, 7 );
	mpFont->print( 30, 50, 0.9, 0.9, 0.9, "Lead Programmer: Marcos Zini" );	
	mpFont->print( 30, 35, 0.9, 0.9, 0.9, "Art and Programming: Todd Martin" );	

	mpFont->setDefaultDim();
	float lY = 260;

	for ( int i = 0; i < mEntries.size(); i++ )
	{
		if ( i == MENU_RESUME && mGameStarted == false )
      mpFont->print( 240, lY, 0.6, 0.6, 0.6, mEntries[ i ].c_str() );
    else
      mpFont->print( 240, lY, mEntries[ i ].c_str() );

		if ( i == mCurrEntry )
			mpFont->print( 220, lY, "*" );

		lY -= 20;
	}
}

////////////////////////////////////////////////////
// Draw the High Scores list
////////////////////////////////////////////////////
void
menu_sys_t::draw_scores()
{
	mpFont->setDefaultDim();
	float lY = 70;

	// Draw the menu
  for ( int i = 0; i < mEntries.size(); i++ )
	{
		mpFont->print( 240, lY, mEntries[ i ].c_str() );

		if ( i == mCurrEntry )
			mpFont->print( 220, lY, "*" );

		lY -= 20;
	}

  // Draw the high score list
  gpScoreList->display( mpFont );
}

////////////////////////////////////////////////////
// Draw the options menu
////////////////////////////////////////////////////
void
menu_sys_t::draw_options()
{
  mpFont->setDim( 22.5, 30, 16 );
	mpFont->print( 270, 360, 1, 0, 0, "OPTIONS" );

	mpFont->setDefaultDim();
	float lY = 260;

	for ( int i = 0; i < mEntries.size(); i++ )
	{
    // Separate "BACK TO THE MAIN MENU" from the other options
    if ( i == ( mEntries.size() - 1 ) )
      mpFont->print( 240, lY -= 20, mEntries[ i ].c_str() );
    else
		  mpFont->print( 240, lY, mEntries[ i ].c_str() );

		if ( i == mCurrEntry )
			mpFont->print( 220, lY, "*" );

		lY -= 20;
	}
}
