#ifndef __MODEL__
#define __MODEL__

#include <stdlib.h>
#include <stdio.h>
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <string>
#include <list>
#include <vector>
#include <memory.h>

#include "globals.h"
#include "glext.h"
#include "ijl.h"
#include "misc.h"
#include "log.h"
#include "vect3.h"
#include "shader.h"

struct index_t
{
  int i0, i1, i2;
};

struct vect2_t
{
  float x, y;
};

////////////////////////////////////////////////////////////////////////
class model_t
{
public :
  model_t();
  virtual ~model_t();

	model_t( const model_t &aModel );
	const model_t &operator=( const model_t &aModel );

  bool load( const char *aFileName, float aScale = 1 );
  
  void draw();
  void draw_lines();
  void draw_normals();

  void set_pointers();
  void unset_pointers();
  void draw_pointers();
  
  const string &get_name() const { return mName; }

  const sphere_t &get_sphere() const { return mSphere; }
  void get_box( vect3_t &aMin, vect3_t &aMax ) const { aMin = mMin; aMax = mMax; }

  vect3_t *get_vertices() { return mpVerts; }
  int get_num_vertices() const { return mNumVerts; }

protected :

  bool load_text( const char *aFileName, float aScale );
  bool load_bin( const char *aFileName, float aScale );

  void check_index( vector<index_t> &aIndeces, const index_t aIndex );
  void gen_face_normals();

  void gen_sphere();
  void gen_box();

  void dump_gometry();

  string mName;

  int mNumInds; 
  int mNumVerts;

  GLenum mIndType;

  vect3_t *mpVerts;
  vect2_t *mpTCoords;
  vect3_t *mpNormals;
  uint32 *mpInds32; // indices

  sphere_t mSphere;
  vect3_t mMin, mMax; // Bounding box
};

////////////////////////////////////////////////////////////////////////
class model_node_t
{
public:
	model_node_t();
	~model_node_t() {}

	void set_model( model_t *apModel ) { mpModel = apModel; }
	void set_shader( shader_t *apShader );
	void set_texture( int aTexture ) { mTexture = aTexture; }
  void set_pos( const vect3_t &lPos ) { mPos = lPos; }
  const vect3_t &get_pos() const { return mPos; }

	void draw();
	void draw_lines();
  void draw_pos();

  void set_pointers() { mpModel->set_pointers(); }
  void draw_pointers();
  void unset_pointers() { mpModel->unset_pointers(); } 

  const sphere_t &get_sphere() const { return mpModel->get_sphere(); }
  void get_box( vect3_t &aMin, vect3_t &aMax ) const 
    { mpModel->get_box( aMin, aMax ); }

private:

  // Shader Stuff
	void draw_shader();
	void update_shader_pass( int aNum );
  void transform_vertices();
  //

	model_t *mpModel;

	int mTexture;

	shader_t *mpShader;

	// Data for each shader pass
	vector<PolyPassData> mShaderData;

  vect3_t mPos;
};

#endif
