#include "model_node.h"

////////////////////////////////////////////////////////////////////////
model_node_t::model_node_t() 
{
	mpShader = NULL;
	mpModel = NULL;
}

////////////////////////////////////////////////////////////////////////
void 
model_node_t::set_shader( shader_t *apShader )
{	
	mpShader = apShader;
	assert( mpShader );

	for ( int i = 0; i < mpShader->passes.size(); i++ )
	{
		PolyPassData data;

		for ( int tcpass = 0; tcpass < mpShader->passes[ i ].tcData.size(); tcpass++ )
		{
			ShaderPassData spd;
			data.tcData.push_back( spd );
			data.tcFlags.push_back( mpShader->passes[ i ].tcFlags[ i ] );
		}

		mShaderData.push_back( data );
	}

}

////////////////////////////////////////////////////////////////////////
void 
model_node_t::draw()
{
	if ( mpShader )
	{
		draw_shader();
	}
	else
	{
		glBindTexture( GL_TEXTURE_2D, mTexture );
		mpModel->draw();
	}
}
 
////////////////////////////////////////////////////////////////////////
void 
model_node_t::draw_lines()
{
//	if ( mpShader )
//	{
//		draw_shader();
//	}
//	else
//	{
//		glBindTexture( GL_TEXTURE_2D, mTexture );
		mpModel->draw_lines();
//	}
}
 
////////////////////////////////////////////////////////////////////////
void 
model_node_t::draw_pos()
{
  glPushMatrix(); 
  glTranslatef( mPos.x, mPos.y, mPos.z ); 
  draw();
  glPopMatrix();
}

////////////////////////////////////////////////////////////////////////
void 
model_node_t::draw_shader()
{	
	static ShaderPass *pass;
	static PolyPassData *data;

	if ( mpShader->mFlags & SF_INVISIBLE )
		return;

  // Enable blending for rendering geometry with shaders
  glEnable( GL_BLEND );

	if ( mpShader->mFlags & SF_DEPTH_OFF )
		glDepthMask( GL_FALSE );

	if ( mpShader->mFlags & SF_NO_CULL )
		glDisable( GL_CULL_FACE );

  if ( mpShader->mFlags & SF_TRANSFORM )
  {
    transform_vertices();
  }

	// Now we draw each pass
	for ( int i = 0; i < mpShader->passes.size(); i++ )
	{
		// If we have multiple passes -> set depth test to equal
		if ( i == 1 )			
			glDepthFunc( GL_EQUAL );

		pass = &mpShader->passes[ i ];
		data = &( mShaderData[ i ] );

		// Update the current shader pass
		update_shader_pass( i );

		// Draw the model
		mpModel->draw();

		if ( pass->mFlags & SF_TCMOD )
		{
			glMatrixMode( GL_TEXTURE );
			glLoadIdentity();
			glMatrixMode( GL_MODELVIEW );
		}

		if ( pass->mFlags & SF_ALPHA )
		{
			glDisable( GL_ALPHA_TEST );
		}   

		if ( pass->mFlags & SF_TCGEN )
		{
			glEnableClientState( GL_NORMAL_ARRAY );
			glDisable( GL_TEXTURE_GEN_S );
			glDisable( GL_TEXTURE_GEN_T );
		}	  
	}

	glDepthMask( GL_TRUE );
	glEnable( GL_CULL_FACE );
	glDepthFunc( GL_LESS );
  glDisable( GL_BLEND );
}

////////////////////////////////////////////////////////////////////////
void 
model_node_t::update_shader_pass( int aNum )
{	
	static ShaderPass *pass;
	static PolyPassData *data;

	pass = &mpShader->passes[ aNum ];
	data = &( mShaderData[ aNum ] );

	data->mCurrTexture = pass->mTextures[ 0 ];

	if ( pass->mFlags & SF_TEXCYCLE )
	{
		data->mCurrTexture = pass->mTextures[ data->cycleData.d1.iVal ];
		if ( data->cycleData.d0.iVal++ == pass->cycleData.d1.iVal )
		{
			data->cycleData.d0.iVal = 0;
			if ( ++data->cycleData.d1.iVal == pass->cycleData.d1.iVal )
				data->cycleData.d1.iVal = 0;
		}    
	}  
	
	glBindTexture( GL_TEXTURE_2D, data->mCurrTexture );

	if ( pass->mFlags & SF_ALPHA )
	{
		glEnable( GL_ALPHA_TEST );
	}	  

	if ( pass->mFlags & SF_TCGEN )
	{
		glDisableClientState( GL_NORMAL_ARRAY );
		glTexGeni( GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP );
		glTexGeni( GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP );
		glEnable( GL_TEXTURE_GEN_S );
		glEnable( GL_TEXTURE_GEN_T );
	}	

	if ( pass->mFlags & SF_RGBGEN )
	{
		if ( pass->mFlags & SF_SQUARE )
		{
			// X++
			data->rgbData.d0.fVal += gFactTick;

			if ( data->rgbData.d0.fVal < pass->rgbData.d2.fVal )
			{
				data->rgbData.d1.fVal = 0; // Y = 0
			}
			else
			{
				data->rgbData.d1.fVal = pass->rgbData.d1.fVal; // Y = YMax
			}

			if ( data->rgbData.d0.fVal > ( 2 * pass->rgbData.d2.fVal ) )
			{
				data->rgbData.d0.fVal = 0;
			}

			// Final RGB = currY + YMin
			pass->mRGB = data->rgbData.d1.fVal + pass->rgbData.d0.fVal;
		}
		else if ( pass->mFlags & SF_TRIANGLE )
		{
			data->rgbData.d0.fVal += gFactTick;

			if ( data->rgbData.d0.fVal < pass->rgbData.d2.fVal )
			{
				data->rgbData.d1.fVal += pass->rgbData.d1.fVal * gFactTick;
			}
			else
			{
				data->rgbData.d1.fVal -= pass->rgbData.d1.fVal * gFactTick;
			}

			if ( data->rgbData.d0.fVal > ( 2 * pass->rgbData.d2.fVal ) )
			{
				data->rgbData.d0.fVal = 0;
				data->rgbData.d1.fVal = 0;
			}

			// Final RGB = currY + YMin
			pass->mRGB = data->rgbData.d1.fVal + pass->rgbData.d0.fVal;
		}
		else if ( pass->mFlags & SF_SAWTOOTH )
		{
			data->rgbData.d0.fVal += gFactTick;

			if ( data->rgbData.d0.fVal < pass->rgbData.d2.fVal )
			{
				data->rgbData.d1.fVal += pass->rgbData.d1.fVal * gFactTick;
			}
			else
			{
				data->rgbData.d0.fVal = 0;
				data->rgbData.d1.fVal = 0;
			}

			// Final RGB = currY + YMin
			pass->mRGB = data->rgbData.d1.fVal + pass->rgbData.d0.fVal;
		}
		else if ( pass->mFlags & SF_SINE )
		{
			data->rgbData.d0.fVal += gFactTick * pass->rgbData.d3.fVal;
			data->rgbData.d1.fVal = fsin( data->rgbData.d0.fVal ) * pass->rgbData.d1.fVal;

			// Final RGB = currY + YMin
			pass->mRGB = data->rgbData.d1.fVal + pass->rgbData.d0.fVal;
		}
	}

	if ( pass->mFlags & SF_TCMOD )
	{
		glMatrixMode( GL_TEXTURE );

		for ( int tcpass = 0; tcpass < pass->tcData.size(); tcpass++ )
		{
			unsigned int lFlag = pass->tcFlags[ tcpass ];
			ShaderPassData *shData = &pass->tcData[ tcpass ];	
			ShaderPassData *paData = &data->tcData[ tcpass ];		

			switch ( lFlag )
			{
			case SF_SCROLL :
				{
					paData->d0.fVal += shData->d0.fVal * gFactTick;
					paData->d1.fVal += shData->d1.fVal * gFactTick;
					glTranslatef( paData->d0.fVal, paData->d1.fVal, 0 );
					break;
				}
			case SF_ROTATE :
				{
					paData->d0.fVal += shData->d0.fVal * gFactTick;
					glTranslatef( 0.5, 0.5, 0 );
					glRotatef( paData->d0.fVal, 0.0, 0.0, 1.0 );
					glTranslatef( -0.5, -0.5, 0 );
					break;
				}
			case SF_SCALE :
				{
					if ( paData->d1.iVal == 0 )
					{
						paData->d0.fVal += shData->d0.fVal * gFactTick;
						if ( paData->d0.fVal >= shData->d2.fVal )
						{
							paData->d0.fVal = shData->d2.fVal;
							paData->d1.iVal = !paData->d1.iVal;
						}
					}
					else
					{
						paData->d0.fVal -= shData->d0.fVal * gFactTick;
						if ( paData->d0.fVal < shData->d1.fVal )
						{
							paData->d0.fVal = shData->d1.fVal;
							paData->d1.iVal = !paData->d1.iVal;
						}
					}
					glTranslatef( 0.5, 0.5, 0 );
					glScalef( paData->d0.fVal, paData->d0.fVal, 0 );
					glTranslatef( -0.5, -0.5, 0 );
					break;
				}		
			}
		}
		glMatrixMode( GL_MODELVIEW );
	}

	glColor4f( pass->mRGB, pass->mRGB, pass->mRGB, pass->mRGB );

	glBlendFunc( pass->mSrc, pass->mDst );
}
 
////////////////////////////////////////////////////////////////////////
void 
model_node_t::transform_vertices()
{
  static float lAnim = 0, lFreq, lAmp;
  static vect3_t *lpVect;

  lFreq = mpShader->mXferFunction.mArgs[ 0 ];
  lAmp = mpShader->mXferFunction.mArgs[ 1 ];

  lAnim -= lFreq * gFactTick;

  for ( int i = 0; i < mpModel->get_num_vertices(); ++i )
  {
    lpVect = &( mpModel->get_vertices()[ i ] );

    //y = sin( ( ( x + z ) * freq ) + anim ) * amp;
    //lpVect->y = sin( ( ( lpVect->x + lpVect->z ) * lFreq ) + lAnim ) * lAmp;
    lpVect->y = sin( ( ( ( lpVect->x + lpVect->z ) *100) + lAnim ) ) * lAmp;
  }
}
