#include "particle.h"

//////////////////////////////////////////////////////////////////////
particle_sys_t::particle_sys_t()
{
	mParticles = new particle_t [ MAX_PARTICLES ];
}

//////////////////////////////////////////////////////////////////////
particle_sys_t::~particle_sys_t()
{
	if ( mParticles )
		delete [] mParticles;

	mParticles = NULL;
}

//////////////////////////////////////////////////////////////////////
particle_sys_t::particle_sys_t( const particle_sys_t &aParticleSystem )
{
	mPos = aParticleSystem.mPos;
	mVel = aParticleSystem.mVel;
	mVelDelta = aParticleSystem.mVelDelta;
	mValid = aParticleSystem.mValid;
	mStartNum = aParticleSystem.mStartNum;
	mCurrNum = aParticleSystem.mCurrNum;
	mTexid = aParticleSystem.mTexid;
	mLoop = aParticleSystem.mLoop;
  mLoopEnabled = aParticleSystem.mLoopEnabled;
	mGravity = aParticleSystem.mGravity;
	mEnergy = aParticleSystem.mEnergy;
	mEnergyDelta = aParticleSystem.mEnergyDelta;
	mStartSize = aParticleSystem.mStartSize;
	mSizeInc = aParticleSystem.mSizeInc;
	mSizeMax = aParticleSystem.mSizeMax;
	mFade = aParticleSystem.mFade;
	mColor = aParticleSystem.mColor;
	mUpdate = aParticleSystem.mUpdate;
	mSrc = aParticleSystem.mSrc;
	mDst = aParticleSystem.mDst;

	if ( aParticleSystem.mParticles )
	{
		mParticles = new particle_t [ MAX_PARTICLES ];
		for ( int i = 0; i < MAX_PARTICLES; i++ )
			mParticles[ i ] = aParticleSystem.mParticles[ i ];
	}
}

//////////////////////////////////////////////////////////////////////
const particle_sys_t &
particle_sys_t::operator=( const particle_sys_t &aParticleSystem )
{
	mPos = aParticleSystem.mPos;
	mVel = aParticleSystem.mVel;
	mVelDelta = aParticleSystem.mVelDelta;
	mValid = aParticleSystem.mValid;
	mStartNum = aParticleSystem.mStartNum;
	mCurrNum = aParticleSystem.mCurrNum;
	mTexid = aParticleSystem.mTexid;
	mLoop = aParticleSystem.mLoop;
  mLoopEnabled = aParticleSystem.mLoopEnabled;
	mGravity = aParticleSystem.mGravity;
	mEnergy = aParticleSystem.mEnergy;
	mEnergyDelta = aParticleSystem.mEnergyDelta;
	mStartSize = aParticleSystem.mStartSize;
	mSizeInc = aParticleSystem.mSizeInc;
	mSizeMax = aParticleSystem.mSizeMax;
	mFade = aParticleSystem.mFade;
	mColor = aParticleSystem.mColor;
	mUpdate = aParticleSystem.mUpdate;
	mSrc = aParticleSystem.mSrc;
	mDst = aParticleSystem.mDst;

	if ( aParticleSystem.mParticles )
	{
		mParticles = new particle_t [ MAX_PARTICLES ];
		for ( int i = 0; i < MAX_PARTICLES; i++ )
			mParticles[ i ] = aParticleSystem.mParticles[ i ];
	}
	return *this;
}

//////////////////////////////////////////////////////////////////////
void
particle_sys_t::reset( bool aRandXZ )
{
  static float lVX, lVZ;

	mValid = true;

	mCurrNum = mStartNum;

  mLoopEnabled = mLoop;

	for ( int i = 0; i < mStartNum; i++ )
	{
		mParticles[ i ].mValid = true;
		mParticles[ i ].mPos = mPos;
		mParticles[ i ].mEnergyStart = mEnergy + crand() * mEnergyDelta;
		mParticles[ i ].mEnergy = mParticles[ i ].mEnergyStart;

		mParticles[ i ].mVelStart.x = mVel.x + crand() * mVelDelta.x;
		mParticles[ i ].mVelStart.y = mVel.y + crand() * mVelDelta.y;
		mParticles[ i ].mVelStart.z = mVel.z + crand() * mVelDelta.z;

    if ( aRandXZ )
    {
      mParticles[ i ].mVelStart.x *= ( rand() & 1 ) ? 1 : -1;
      mParticles[ i ].mVelStart.z *= ( rand() & 1 ) ? 1 : -1;
    }

		mParticles[ i ].mVel = mParticles[ i ].mVelStart;
		mParticles[ i ].mColor = mColor;
					
		mParticles[ i ].mCurrSize = mStartSize + crand() * 0.01f;
	}

	if ( mUpdate )
	{
		for ( int i = 0; i < mStartNum; i++ )
		{
			mParticles[ i ].mEnergy = mEnergy + crand() * mEnergy;
		}
	}
}

//////////////////////////////////////////////////////////////////////
void
particle_sys_t::init( int aStartNum, float aEnergy, float aEnergyDelta,
					  int aTexid, 
					  float aSizeStart, float aSizeInc, float aSizeMax,
					  float aGravity, float aFade, float aColor, bool aLoop,
					  const vect3_t &aVel, const vect3_t &aVelDelta,
					  bool aUpdate, GLenum aSrc, GLenum aDst )
{
  assert( aStartNum <= MAX_PARTICLES );

	mStartNum = mCurrNum = aStartNum;

	mLoop = aLoop;
  mLoopEnabled = aLoop;
	mGravity = aGravity;
	mTexid = aTexid;
	mStartSize = aSizeStart;
	mSizeInc = aSizeInc;
	mSizeMax = aSizeMax;
	mFade = aFade;
	mVel = aVel;
	mVelDelta = aVelDelta;
	mEnergy = aEnergy;
	mEnergyDelta = aEnergyDelta;
	mUpdate = aUpdate;
	mColor = aColor;
	mSrc = aSrc;
	mDst = aDst;
}

//////////////////////////////////////////////////////////////////////
void 
particle_sys_t::update( particle_t &aPart )
{
	aPart.mColor -= mFade * gFactTick;

	aPart.mVel.y -= mGravityTick;

	//static vect3_t lTemp;
	//mult( aPart.mVel, gFactTick, lTemp );
	//add( aPart.mPos, lTemp, aPart.mPos );
	//aPart.mPos += aPart.mVel * gFactTick;

  muladd( aPart.mVel, gFactTick, aPart.mPos, aPart.mPos );

	aPart.mCurrSize += mSizeInc * gFactTick;
	if ( aPart.mCurrSize > mSizeMax )
		aPart.mCurrSize = mSizeMax;

	aPart.mEnergy -= gFactTick;
	if ( aPart.mEnergy <= 0.0f )
	{		
		if ( mLoop && mLoopEnabled )
		{
			aPart.mPos = mPos;
			aPart.mEnergy = aPart.mEnergyStart;
			aPart.mVel = aPart.mVelStart;
			aPart.mColor = mColor;
					
			aPart.mCurrSize = mStartSize;
		}
		else
		{
			aPart.mValid = false;
			if ( --mCurrNum == 0 )
			{
				mValid = false;			
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////
bool
particle_sys_t::draw( const vect3_t &aViewerPos )
{
	if ( !mValid )
		return false;

	glBindTexture( GL_TEXTURE_2D, mTexid );
	glBlendFunc( mSrc, mDst );

	mGravityTick = mGravity * gFactTick;

	for ( int i = 0; i < mStartNum; i++ )
	{
		particle_t &part = mParticles[ i ];

		if ( part.mValid )
		{    
			// Update particle
			update( part );

			// Now draw each particle
			glColor4f( part.mColor, part.mColor, part.mColor, 0.5f );

			glPushMatrix();

			mult_look_at( part.mPos, aViewerPos );

			glBegin( GL_QUADS );
			glTexCoord2f( 0.0, 0.0 );
			glVertex3f( 0.0, 0.0, 0.0 );

			glTexCoord2f( 0.0, 1.0 );
			glVertex3f( 0.0, part.mCurrSize, 0.0 );

			glTexCoord2f( 1.0, 1.0 );
			glVertex3f( part.mCurrSize, part.mCurrSize, 0.0 );

			glTexCoord2f( 1.0, 0.0 );
			glVertex3f( part.mCurrSize, 0.0, 0.0 );
			glEnd();      
			glPopMatrix();
		}
	}
	return true;
}

//////////////////////////////////////////////////////////////////////
void 
particle_sys_t::drawParticleQuad( float aSize )
{
	static float lVerts[ 12 ] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
	static float lTCoords[ 8 ] = { 0, 0, 1, 0, 1, 1, 0, 1 };

	lVerts[ 3 ] = lVerts[ 6 ] = lVerts[ 7 ] = lVerts[ 10 ] = aSize;

	glTexCoordPointer( 2, GL_FLOAT, 0, lTCoords );
	glVertexPointer( 3, GL_FLOAT, 0, lVerts );

	glDrawArrays( GL_QUADS, 0, 4 );	

	/*
	glBegin( GL_QUADS );
	glTexCoord2f( 0.0, 0.0 );
	glVertex3f( 0.0, 0.0, 0.0 );

	glTexCoord2f( 1.0, 0.0 );
	glVertex3f( part.mCurrSize, 0.0, 0.0 );

	glTexCoord2f( 1.0, 1.0 );
	glVertex3f( part.mCurrSize, part.mCurrSize, 0.0 );

	glTexCoord2f( 0.0, 1.0 );
	glVertex3f( 0.0, part.mCurrSize, 0.0 );
	*/
}
