#ifndef __PARTICLE__
#define __PARTICLE__

#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include <vector>
#include "vect3.h"
#include "globals.h"
#include "textures.h"
#include "console.h"

struct particle_t
{
  particle_t() : mValid( false ) {}
  virtual ~particle_t() {}

  vect3_t mPos;
  vect3_t mVel, mVelStart;
  float mColor;
  int mEnergy, mEnergyStart;
  float mCurrSize;
  bool mValid;
};

enum { MAX_PARTICLES = 150 };

class particle_sys_t
{
public:
  particle_sys_t();
  virtual ~particle_sys_t();

  particle_sys_t( const particle_sys_t &aParticleSystem );
  const particle_sys_t &operator=( const particle_sys_t &aParticleSystem );
  
  virtual void init( int aStartNum, 
                     float aEnergy, float aEnergyDelta,
                     int aTexid, 
					           float aSizeStart, float aSizeInc, float aSizeMax,
					           float aGravity, float aFade, float aColor, bool aLoop,
					           const vect3_t &aVel, const vect3_t &aVelDelta,
                     bool aUpdate, GLenum aSrc, GLenum aDst );

  virtual void set_pos( const vect3_t &aPos ) { mPos = aPos; }
  virtual void set_vel( const vect3_t &aVel ) { mVel = aVel; }

  virtual void set_texture( int aTexture ) { mTexid = aTexture; }

  virtual bool draw( const vect3_t &viewerPos );

  virtual void reset( bool aRandXZ = true );
  virtual void disable() { mValid = false; }

  bool is_enabled() const { return mValid; }

  void disable_loops() { mLoopEnabled = false; }

  static void drawParticleQuad( float aSize );

protected:
  void update( particle_t &part );

  particle_t *mParticles;
  vect3_t mPos;                  // Source position
  vect3_t mVel, mVelDelta;
  bool mValid;
  int mStartNum, mCurrNum;
  int mTexid;
  bool mLoop;                    // Do particles loop?
  bool mLoopEnabled;             // Are loops currently enabled
  float mGravity, mGravityTick;
  float mEnergy, mEnergyDelta;
  float mStartSize, mSizeInc, mSizeMax;
  float mFade;
  float mColor;
  GLenum mSrc, mDst;

  bool mUpdate;
};

#endif
