#include "program.h"

///////////////////////////////////////////////////////////////////////////
// Loads both ARB vertex and fragment programs
///////////////////////////////////////////////////////////////////////////
int
program_manager_t::insert( const char *aFileName, GLenum aProgType )
{
  if ( aProgType != GL_VERTEX_PROGRAM_ARB && aProgType != GL_FRAGMENT_PROGRAM_ARB )
  {
    gLog->add_msg( "ERROR: Invalid program type for %s", aFileName );
    return 0;
  }

  string lPath;
  if ( aProgType == GL_VERTEX_PROGRAM_ARB )
    lPath = "data/vertex_progs/";
  else
    lPath = "data/frag_progs/";

  lPath += aFileName;

	FILE *lpFile = fopen( lPath.c_str(), "r" );
	if ( !lpFile )
	{
		gLog->add_msg( "ERROR: Could not open program '%s'", aFileName );
		return 0;
	}

	unsigned char *lData = 0;
	unsigned long lSize = 0;
   
	lSize = COM_FileSize( lpFile );   

	if ( !lSize )
  {
    fclose( lpFile );
		return 0;
  }
   
	if ( !( lData = new unsigned char [ lSize + 1 ] ) )
  {
    fclose( lpFile );
    return 0;
  }

	memset( lData, '\0', lSize + 1 );
   
	fread( lData, 1, lSize, lpFile );

	unsigned int lID;

	glGenProgramsARB( 1, &lID );
	glBindProgramARB( aProgType, lID );

	glProgramStringARB( aProgType, GL_PROGRAM_FORMAT_ASCII_ARB,  
					            strlen( ( char * ) lData ), ( unsigned char * ) lData );
	
	if( glGetError() != GL_NO_ERROR )
  {
		gLog->add_msg( "ERROR: Failed to load program %s. %s", aFileName,
                   ( char * ) glGetString( GL_PROGRAM_ERROR_STRING_ARB ) );
    glDeleteProgramsARB( 1, &lID );
    lID = 0;
  }

  delete [] lData;
  fclose( lpFile );  

	return lID;
}
