#ifndef __PROJECTILE_H__
#define __PROJECTILE_H__

#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include <assert.h>

#include "globals.h"
#include "vect3.h"
#include "model.h"
#include "console.h"

#include "castle_piece.h"

enum { PROJ_FIRE_ANGLE = 45 };

///////////////////////////////////////////////////////////////
class projectile_t
{
public:
  projectile_t() {}
  ~projectile_t() {}

	void set_pos( const vect3_t &aPos ) { mPos = aPos; mCannonPos = aPos; mHeight = mPos.y + 10; 
    mCannonPos.y += 10;}
	const vect3_t &get_pos() const { return mPos; }

	void set_target_piece( castle_piece_t *apPiece ) { mpPiece = apPiece; }

  void set_dir( const vect3_t &aDir ) { mDir = aDir; }
	const vect3_t &get_dir() const { return mDir; }

	void set_vel( float aVel ) { mVel = aVel; mVelSquared = aVel * aVel; }
	float get_vel() const { return mVel; }

	void set_model( model_node_t *apModel ) { mpModel = apModel; }

  bool update( vect3_t &aHitPos );
  void draw();

  void set_valid( bool aValid ) { mValid = aValid; mFlightTime = 0; }
  bool get_valid() const { return mValid; }

  void calc_physics();

private:
  bool mValid;

  vect3_t mPos;
  vect3_t mDir;
  vect3_t mCannonPos;
  float mGravity;
//  float mTotalFlightTime;
  float mTotalFlightTimeInv;
  float mDistance;
  float mVel;
  float mVelSquared;
  float mHeight;
  float mHeightIntermediate;

  vect3_t mTargetPos;
  float mFlightTime;

  castle_piece_t *mpPiece; // The castle piece this guy is targeting

	model_node_t *mpModel;
};

#endif