#include "projectile2.h"


///////////////////////////////////////////////////////////////////
// Calculate the projectile physics
// This function should be called after mVel and mDir are set
// It does not need to be run with every update
// Only after the projectile is created
///////////////////////////////////////////////////////////////////
void
projectile2_t::calc_physics()
{
  static float lVelCos, lVelCosSquared;
  static vect3_t lDir;
  
  static float COS_ANG, INV_COS_ANG, SIN_ANG;
//  static float SIN_2ANG;

  // Need to speed up the velocity in game_t::player_shoot() when an angle other
  // than 45 degrees is used.  If 300 is default 400 is good for 60 degrees.
  switch ( PROJ2_FIRE_ANGLE )
  {
  case 30 :
    COS_ANG = 0.86602540f;
    INV_COS_ANG = 1.15470054f;
    SIN_ANG = 0.5f;
//    SIN_2ANG = 0.86602540f;
    break;
  case 60 :
    COS_ANG = 0.5f;
    INV_COS_ANG = 2;
    SIN_ANG = 0.86602540f;
//    SIN_2ANG = 0.86602540f;
    break;
  default :   // 45 degrees
    COS_ANG = 0.70710678f;
    INV_COS_ANG = 1.41421356f;
    SIN_ANG = 0.70710678f;
//    SIN_2ANG = 1;
    break;
  }
  
  lDir = mDir;
  lDir.y = 0;     // we only want the magnitude in the x-z plane

  // The equations on the next line assume source and destination are at the same elevation
  // R = magnitude(mDir) = v^2 / g * SIN_2ANG = v * COS_ANG * t
  mDistance = lDir.mag();

  // Multiplying by 100 accounts for mFlightTime being 100ths of a second
//  mTotalFlightTime = mDistance * INV_COS_ANG / mVel * 100;
  mTotalFlightTimeInv = COS_ANG * mVel / mDistance;

  if ( mDir.y >= 0 )
    lDir.y = mDir.y;
  else
    lDir.y = -mDir.y;

  // in this world velocity is constant and gravity changes
//  mGravity = mVelSquared / mDistance * SIN_2ANG;
  mGravity = 2 * ( mVel * SIN_ANG * mTotalFlightTimeInv + 
    lDir.y * mTotalFlightTimeInv * mTotalFlightTimeInv );

  mTotalFlightTimeInv *= 0.01f;   // convert to 100ths of a second

  // (g / 2 / (v * COS_ANG)^2)
  lVelCos = mVel * COS_ANG;
  lVelCosSquared = lVelCos * lVelCos;
  mHeightIntermediate = mGravity * 0.5 / lVelCosSquared;
}


///////////////////////////////////////////////////////////////////
// Returns true when the projectile has reached its target and aPos
// is set to the projectile's position at that time
///////////////////////////////////////////////////////////////////
bool 
projectile2_t::update( vect3_t &aPos )
{
  if ( !mValid ) 
    return false;

  float magPosDiff;
  vect3_t PosDiff;

  static float TAN_ANG;

  switch ( PROJ2_FIRE_ANGLE )
  {
  case 30 :
    TAN_ANG = 0.57735027f;
    break;
  case 60 :
    TAN_ANG = 1.73205081f;
    break;
  default :   // 45 degrees
    TAN_ANG = 1;
    break;
  }

  mFlightTime += gFactTick;


  // acceleration = 0 so x - x0 = v * t
  // time is normalized
  mPos = mDir * mFlightTime * mTotalFlightTimeInv;  // x = v * t
  mPos.x += mCannonPos.x;   // add back x0
  mPos.z += mCannonPos.z;   // add back z0

  // y - y0 = TAN_ANG(x - x0) - (g / 2 / (v * COS_ANG)^2) * (x - x0)^2
  PosDiff = mPos - mCannonPos;
  PosDiff.y = 0;    // we only want the magnitude in the x-z plane
  magPosDiff = PosDiff.mag();
  mHeight = ((TAN_ANG * magPosDiff) - ( mHeightIntermediate 
    * magPosDiff * magPosDiff )) + mCannonPos.y;

  if ( mHeight <= 1 )
  {
    mValid = false;

    // This is the only time aPos is valid
    aPos.x = mPos.x;
    aPos.y = 0;
    aPos.z = mPos.z;
    return true;
  }

  return false;
}

///////////////////////////////////////////////////////////////////
void 
projectile2_t::draw()
{
  if ( !mValid ) 
    return;

	glPushMatrix(); 
	glTranslatef( mPos.x, mHeight, mPos.z );
	mpModel->draw();
	glPopMatrix();
}