#include "qsound.h"

//////////////////////////////////////////////////////////////////////
int 
sound_manager_t::init()  
{
  QMIX_DISTANCES qd;

	// Clear all loaded samples
  memset( mSamples, 0, sizeof( mSamples ) );
  mSampleCount = 0;

	// Init QSound
  mQMixer = QM(Init)();

  if ( !mQMixer ) 
	{
		gCon->add_msg( "Could not init QMixer!" );
    return 0;
  }

  gCon->add_msg( "Initializing sound system" );
	
  // Activate the mixer
  QM(Activate)( mQMixer, TRUE );

  QM(SetSpeakerPlacement)( mQMixer, QMIX_SPEAKERS_STEREO );

	// Open all channels
  QM(OpenChannel)( mQMixer, 0, QMIX_OPENALL );

  QM(SetVolume)( mQMixer, 0, QMIX_ALL, 32767 );

	// Setup distance mapping
  qd.cbSize = sizeof( QMIX_DISTANCES );
  qd.minDistance = 20.0f;
  qd.maxDistance = 400.0f;
  qd.scale = 0.8f;
  QM(SetDistanceMapping)( mQMixer, 0, QMIX_ALL, &qd );
  QM(SetRoomSize)( mQMixer, 400, 0 );
  return 1;
}

//////////////////////////////////////////////////////////////////////
void 
sound_manager_t::shut_down( void ) 
{
	int i;

  if ( mQMixer ) 
	{
		for ( i = 0; i < MAX_SAMPLES; i++ ) 
		{
			QM(FreeWave)( mQMixer, mSamples[ i ].mSample );
      mSamples[ i ].mSample = 0;
    }
		QM(CloseSession)( mQMixer );
		mQMixer = 0;
  }
}

//////////////////////////////////////////////////////////////////////
int  
sound_manager_t::load_sample( char *apFileName ) 
{
  int i;
  char lpPath[ 200 ];

  if ( !mQMixer ) 
		return -1;

  strcpy( lpPath, "data/sounds/" );
  strcat( lpPath, apFileName );

	// Find a free sample
  for ( i = 0; i < MAX_SAMPLES; i++ ) 
  {
    if ( mSamples[ i ].mSample )
	{
	  if ( strcmp( mSamples[ i ].mName, apFileName ) == 0 )
	  	return i;
	}
	else
	  break;
  }
    
  if ( i >= MAX_SAMPLES ) 
  {
	gCon->add_msg( "Not enough slots to load %s", lpPath );
	return -1;
  }

	// Move SampleCount past the empty slot
  mSampleCount++;

	// Put sample into i
  mSamples[ i ].mSample = QM(OpenWave)( 0, lpPath, 0, WMIX_FILE );
  mSamples[ i ].mChannel = 0;

  if ( !mSamples[ i ].mSample )
  {
	gCon->add_msg( "Could not load %s", apFileName );
	return -1;
  }

  return i;
}


//////////////////////////////////////////////////////////////////////
void 
sound_manager_t::free_sample( int aSampleID ) 
{
	if ( mQMixer && mSamples[ aSampleID ].mSample ) 
	{
		QM(FreeWave)( mQMixer, mSamples[ aSampleID ].mSample );
    mSamples[ aSampleID ].mSample = 0;
    mSamples[ aSampleID ].mChannel = 0;
  }
}

//////////////////////////////////////////////////////////////////////
void 
sound_manager_t::stop_sample( int aSampleID ) 
{
	QM(SetVolume)( mQMixer, mSamples[ aSampleID ].mChannel, 
                 QMIX_USEONCE, 0 );
	QM(StopChannel)( mQMixer, mSamples[ aSampleID ].mChannel,
   		             QMIX_STOP_FINISHLASTLOOP );
}

//////////////////////////////////////////////////////////////////////
void 
sound_manager_t::stop_all()
{
	QM(StopChannel)( mQMixer, 0, QMIX_ALL );
}

//////////////////////////////////////////////////////////////////////
void 
sound_manager_t::set_sample_volume( int aSampleID, int aVolume ) 
{
	QM(SetVolume)( mQMixer, mSamples[ aSampleID ].mChannel, 
                 QMIX_USEONCE, aVolume );
}

//////////////////////////////////////////////////////////////////////
void 
sound_manager_t::set_volume( int aVolume )
{	
  QM(SetVolume)( mQMixer, 0, QMIX_ALL, aVolume );
}

//////////////////////////////////////////////////////////////////////
void 
sound_manager_t::set_sample_pos( int aSampleID, const float *apPos )
{
  QM(SetSourcePosition)( mQMixer, mSamples[ aSampleID ].mChannel, 
                         QMIX_IMMEDIATE, ( QSVECTOR * ) apPos );
}

//////////////////////////////////////////////////////////////////////
int  
sound_manager_t::play_sample3d( int aSampleID, const float *apPos, int aLoop )
{
  mSamples[ aSampleID ].mChannel = QM(FindChannel)( mQMixer, QMIX_FINDCHANNEL_LRU, 0 );

	QM(SetVolume)( mQMixer, mSamples[ aSampleID ].mChannel, QMIX_USEONCE, 
 		             con_sndfx_vol );

  QM(SetSourcePosition)( mQMixer, mSamples[ aSampleID ].mChannel, 
                         QMIX_WAIT, ( QSVECTOR * ) apPos );
  QM(PlayEx)( mQMixer, mSamples[ aSampleID ].mChannel, 
  		        QMIX_CLEARQUEUE | QMIX_IMMEDIATE, 
              mSamples[ aSampleID ].mSample, aLoop, 0 );

  return mSamples[ aSampleID ].mChannel;
}

//////////////////////////////////////////////////////////////////////
int  
sound_manager_t::play_sample( int aSampleID, int aLoop )
{
  mSamples[ aSampleID ].mChannel = QM(FindChannel)( mQMixer, QMIX_FINDCHANNEL_LRU, 0 );

	QM(SetVolume)( mQMixer, mSamples[ aSampleID ].mChannel, QMIX_USEONCE, 
 		             con_sndfx_vol );

  QM(SetSourcePosition)( mQMixer, mSamples[ aSampleID ].mChannel, 
                         QMIX_WAIT, ( QSVECTOR * ) mListenPos.mVect );
  QM(PlayEx)( mQMixer, mSamples[ aSampleID ].mChannel, 
  		        QMIX_CLEARQUEUE | QMIX_IMMEDIATE, 
              mSamples[ aSampleID ].mSample, aLoop, 0 );

  return mSamples[ aSampleID ].mChannel;
}

static QSVECTOR upa = { 0, 1, 0 };

//////////////////////////////////////////////////////////////////////
void 
sound_manager_t::set_listener( const float *apPos, const float *apDir ) 
{
  mListenPos.mVect[ 0 ] = apPos[ 0 ];
  mListenPos.mVect[ 1 ] = apPos[ 1 ];
  mListenPos.mVect[ 2 ] = apPos[ 2 ];

	QM(SetListenerPosition)( mQMixer, ( QSVECTOR * ) apPos, 0 );
  QM(SetListenerOrientation)( mQMixer, ( QSVECTOR * ) apDir, &upa, 0 );
}


