#include "score_list.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

score_list_t::score_list_t()
{
  bool lReadStatus;
  
  mNameEntered = false;
  
  mHighScores = new list<score_data_t>;
  
  // Read data from the hard list file
  mHighScores = &mHighScoresHard;
  lReadStatus = read_bin_file( "scores_h.dat" );

  if ( !lReadStatus )
  {
    gLog->add_msg( "No hard high score list was found" );

    score_data_t lCurData;
    
    strcpy( lCurData.mName, " " );
    lCurData.mScore = 0;
    // Fill mHighScores list
    for ( int i = 0; i < 10; ++i )
      mHighScores->push_back( lCurData );
  }

  // Read data from the easy list file
  mHighScores = &mHighScoresEasy;
  lReadStatus = read_bin_file( "scores_e.dat" );

  if ( !lReadStatus )
  {
    gLog->add_msg( "No easy high score list was found" );

    score_data_t lCurData;
    
    strcpy( lCurData.mName, " " );
    lCurData.mScore = 0;
    // Fill mHighScores list
    for ( int i = 0; i < 10; ++i )
      mHighScores->push_back( lCurData );
  }

  if ( con_difficulty == 1 )
    mHighScores = &mHighScoresHard;
  else
    mHighScores = &mHighScoresEasy;
}

score_list_t::~score_list_t()
{
  delete mHighScores;
}

//////////////////////////////////////////////////////////////////////
// Reads the list data from a binary file and builds the list data structure
//////////////////////////////////////////////////////////////////////
bool
score_list_t::read_bin_file( string asFileName )
{
  score_data_t lCurData[ 10 ];
  
  // Load the high score file
  FILE *lpFile = fopen( asFileName.c_str(), "rb" );   // Read-only file

  if ( !lpFile )
  {
    gLog->add_msg( "ERROR: Could not open binary high score list" );
    return false;
  }

  //+ Create a header struct and use it to add some basic encryption
//  fread( &lHeader, sizeof( lHeader ), 1, lpFile );
  
  // Read the 10 high score names and numbers
  fread( &lCurData, sizeof( score_data_t ), 10, lpFile );

  // Fill mHighScores list
  for ( int i = 0; i < 10; ++i )
    mHighScores->push_back( lCurData[ i ] );

  fclose( lpFile );

  gLog->add_msg( "Loaded binary high score list" );

  return true;
}


//////////////////////////////////////////////////////////////////////
// Reads the list data from a text file and builds the list data structure
//////////////////////////////////////////////////////////////////////
bool
score_list_t::read_text_file( string asFileName )
{
  score_data_t lCurData;
  
  // Load the high score file
  FILE *lpFile = fopen( asFileName.c_str(), "r" );   // Read-only file

  if ( !lpFile )
  {
    gLog->add_msg( "ERROR: Could not open text high score list" );
    return false;
  }

  // Fill mHighScores list
  char lLine[ 100 ];
  while ( fgets( lLine, 100, lpFile ) )
  {
    char *tok = strtok( lLine, " \n\t" );
    char *tok2;

    sscanf( tok, "%s", lCurData.mName );
    tok2 = strtok( NULL, " \n\t" );
    if ( !tok2 )
    {
      sscanf( tok, "%d", &lCurData.mScore );
      lCurData.mName[ 0 ] = '\0';
    }
    else
      sscanf( tok2, "%d", &lCurData.mScore );

    //+ I probably need some error handling here
    mHighScores->push_back( lCurData );
  }

  fclose( lpFile );
  
  gLog->add_msg( "Loaded text high score list" );

  return true;
}


//////////////////////////////////////////////////////////////////////
// Writes the list data to a binary file
//////////////////////////////////////////////////////////////////////
bool
score_list_t::write_bin_file( string asFileName )
{
  int lListSize, i;
  score_data_t lCurData;
  
  FILE *lpFile = fopen( asFileName.c_str(), "wb" );   // Erases previous file

  if ( !lpFile )
  {
    gLog->add_msg( "ERROR: Could not open binary high score list for writing" );
    return false;
  }

  lListSize = mHighScores->size();

  list<score_data_t>::iterator it = mHighScores->begin();

  for ( i = 0; i < lListSize; ++i )
  {
    lCurData = *it;

    fwrite( &lCurData, sizeof( score_data_t ), 1, lpFile );

    ++it;
  }

  fclose( lpFile );
  
  gLog->add_msg( "Wrote binary high score list" );

  return true;
}


//////////////////////////////////////////////////////////////////////
// Writes the list data to a text file
//////////////////////////////////////////////////////////////////////
bool
score_list_t::write_text_file( string asFileName )
{
  int lListSize, i;
  score_data_t lCurData;
  
  FILE *lpFile = fopen( asFileName.c_str(), "w" );   // Erases previous file

  if ( !lpFile )
  {
    gLog->add_msg( "ERROR: Could not open text high score list for writing" );
    return false;
  }

  lListSize = mHighScores->size();

  list<score_data_t>::iterator it = mHighScores->begin();

  for ( i = 0; i < lListSize; ++i )
  {
    lCurData = *it;

    fprintf( lpFile, "%s %d\n", lCurData.mName, lCurData.mScore );

    ++it;
  }

  fclose( lpFile );
  
  gLog->add_msg( "Wrote text high score list" );

  return true;
}


//////////////////////////////////////////////////////////////////////
// Asks the player's name
//////////////////////////////////////////////////////////////////////
void 
score_list_t::ask_name( char aName[], font_t *apFont )
{
  char lName[16];
  
	apFont->setDefaultDim();
	apFont->print( 110, 340, 1, 1, 1, "Congratulations.  You have a new high score." );	
	apFont->setDim( 14, 14, 8 );
	apFont->print( 130, 140, 1, 1, 1, "Please type your name and press enter to continue" );	

  // Start receiving keyboard input
  gpNameTextBox->enable();

  // Draw the inputed text
  gpNameTextBox->draw_text( apFont, 280, 110 );
  gpNameTextBox->get_text( lName );

  strcpy( aName, lName );
}


//////////////////////////////////////////////////////////////////////
// Returns true if the new score is good enough to make the list
//////////////////////////////////////////////////////////////////////
bool 
score_list_t::check_score( int aScore )
{
  score_data_t lLowScore;

  lLowScore = mHighScores->back();

  // A tie for the last spot goes to the newest score
  if ( aScore >= lLowScore.mScore )
    return true;

  return false;
}


//////////////////////////////////////////////////////////////////////
// Updates the list with the new high score
// Assumes the score was already checked by check_score()
//////////////////////////////////////////////////////////////////////
void
score_list_t::update_list( score_data_t aScoreData )
{
  int lListSize, i;
  score_data_t lCurData;
  bool lWriteStatus;

  lListSize = mHighScores->size();

  list<score_data_t>::iterator it = mHighScores->begin();

  // Loop through the list, starting at the top, and check each score
  for ( i = 0; i < lListSize; ++i )
  {
    lCurData = *it;
    if ( aScoreData.mScore >= lCurData.mScore )
    {
      mHighScores->insert( it, aScoreData );   // insert new score
      break;                      // exit the for loop
    }
    ++it;
  }  

  mHighScores->pop_back();         // remove the 11th score

  // Write the new list to the proper file
  if ( con_difficulty == 1 )
    lWriteStatus = write_bin_file( "scores_h.dat" );
  else
    lWriteStatus = write_bin_file( "scores_e.dat" );

  if ( !lWriteStatus )
    gLog->add_msg( "ERROR: unable to write the high score list" );
}


//////////////////////////////////////////////////////////////////////
// Display the high score list on the screen
//////////////////////////////////////////////////////////////////////
void
score_list_t::display( font_t *apFont )
{
  int lListSize, i;
  score_data_t lCurData;
  char lsScore[10];

  apFont->setDim( 22.5, 30, 16 );
	apFont->print( 220, 380, 1, 0, 0, "HIGH SCORES" );
	
	apFont->setDefaultDim();
	apFont->print( 190, 340, 1, 1, 0, "Name" );	
	apFont->print( 390, 340, 1, 1, 0, "Score" );	

  lListSize = mHighScores->size();

  list<score_data_t>::iterator it = mHighScores->begin();

  // Loop through the list, starting with the top score
  for ( i = 0; i < lListSize; ++i )
  {
    lCurData = *it;

    // Print the name and score
    sprintf( lsScore, "%d", lCurData.mScore );
    apFont->print( 190, (300-20*i), 1, 1, 1, lCurData.mName );
    apFont->print( 390, (300-20*i), 1, 1, 1, lsScore );

    ++it;
  }  
}


//////////////////////////////////////////////////////////////////////
// Necessary to make sure the proper list is used when a game is 
// restarted.
//////////////////////////////////////////////////////////////////////
void
score_list_t::restart()
{
  if ( con_difficulty == 1 )
    mHighScores = &mHighScoresHard;
  else
    mHighScores = &mHighScoresEasy;
}


