#ifndef __SCORE_LIST_H__
#define __SCORE_LIST_H__

#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include <list>
#include <string>

using namespace std;

#include "globals.h"
#include "font.h"
#include "console.h"
#include "textbox.h"


//////////////////////////////////////////////////////////////////////
// Data that is stored in each high score list
//////////////////////////////////////////////////////////////////////
struct score_data_t
{
  int mScore;
  char mName[16];
};


//////////////////////////////////////////////////////////////////////
// The actual high score list
//////////////////////////////////////////////////////////////////////
class score_list_t  
{
public:
  score_list_t();
	~score_list_t();

  // Returns true if the new score is good enough to make the list
  bool check_score( int aScore );

  // Updates the list with the new high score
  void update_list( score_data_t aScoreData );

  void ask_name( char aName[], font_t *apFont );
  void display( font_t *apFont );

  bool get_name_entered() { return mNameEntered; }

  void set_name_entered() { mNameEntered = true; }
  void clear_name_entered() { mNameEntered = false; }

  void restart();

private:

  bool read_bin_file( string asFileName );
  bool read_text_file( string asFileName );
  bool write_bin_file( string asFileName );
  bool write_text_file( string asFileName );

  bool mNameEntered;

  list<score_data_t> *mHighScores;
  list<score_data_t> mHighScoresEasy;
  list<score_data_t> mHighScoresHard;
};

extern score_list_t *gpScoreList;

#endif
