#include "shader.h"

/////////////////////////////////////////////////////////////////////
shader_manager_t::shader_manager_t( texture_manager_t *apTextures ) : mTextures( apTextures )
{
	// Insert a default shader
	shader_t *shader = new shader_t;
	shader->mName = "default_shader";
	shaders.push_back( shader );
}

/////////////////////////////////////////////////////////////////////
shader_manager_t::~shader_manager_t()
{
  gDebug->add_msg( "~shader_manager_t() : start" );

	for ( int i = 0; i < shaders.size(); i++ )
		delete shaders[ i ];

    shaders.clear(); 

  gDebug->add_msg( "~shader_manager_t() : end" );
}

/////////////////////////////////////////////////////////////////////
shader_t *
shader_manager_t::getShader( const string &name )
{
	for ( int i = 0; i < shaders.size(); i++ )
	{
		if ( shaders[ i ]->mName == name )
		{
			gLog->add_msg( "Applying shader %s", name.c_str() );
			return shaders[ i ];
		}
	}
	
	gLog->add_msg( "-- Cannot apply shader '%s'", name.c_str() );

	// Shader not found -> return default
	return shaders[ 0 ];
}

/////////////////////////////////////////////////////////////////////
shader_t *
shader_manager_t::insert( const shader_t &shader )
{
	shader_t *shr = new shader_t;
	*shr = shader;
  
	//con->InsertMsg( "-- Loading shader '%s'", shader.mName.c_str() );

	shaders.push_back( shr );
	return shr;
};

/////////////////////////////////////////////////////////////////////
void
shader_manager_t::clear() 
{
	for ( int i = 0; i < shaders.size(); i++ )
		delete shaders[ i ];

    shaders.clear(); 

	// Insert a default shader
	shader_t *shader = new shader_t;
	shader->mName = "default_shader";
	shaders.push_back( shader );
}

/////////////////////////////////////////////////////////////////////
bool 
shader_manager_t::readFromFile( const char *aFileName )
{
  char lPathName[ 200 ];
  strcpy( lPathName, "data/shaders/" );
  strcat( lPathName, aFileName );

	FILE *fp = fopen( lPathName, "rb" );
	if ( !fp )
	{
		gLog->add_msg( "-- Cannot open shader file '%s'", aFileName );
		return false;
	}

	int lineCount = 0;
	char line[ 500 ];
  int lNumShaders = 0;
    
	while ( fgets( line, 500, fp ) != NULL )
	{
		lineCount++;
	  char *tok = strtok( line, " \n\t\r" );
	  
    if ( !tok || tok[ 0 ] == '-' || tok[ 0 ] == '#' )
			continue;
	  else if ( strcmp( tok, "shader_begin" ) == 0 )
		{
      lNumShaders++;

			tok = strtok( NULL, " \n\t\r" );

			parseShader( fp, tok, lineCount );
		}
	}

  gLog->add_msg( "-- Parsed %d shaders from file '%s'", lNumShaders, 
                 aFileName );
	return true;
}

/////////////////////////////////////////////////////////////////////
bool 
shader_manager_t::parseShader( FILE *apFile, const char *aShaderName, 
  	                           int &aLineCount )
{	
	string texture = "default.bmp";

	shader_t shader;
	ShaderPass pass;
	GenericData data;

	shader.mFlags = 0;
	shader.mName = aShaderName;

	bool clamptex = false;
    int numTextures = 0;

	char line[ 200 ];
    while ( fgets( line, 500, apFile ) != NULL )
    {
		aLineCount++;
		char *tok = strtok( line, " \n\t\r" );
		if ( !tok || tok[ 0 ] == '-' || tok[ 0 ] == '#' )
			continue;

		if ( strcmp( tok, "shader_end" ) == 0 )
			break;
		else if ( strcmp( tok, "depth_off" ) == 0 )
		{
			shader.mFlags |= SF_DEPTH_OFF;
		}
		else if ( strcmp( tok, "faceviewer" ) == 0 )
		{
			shader.mFlags |= SF_VIEWERFACING;
		}
		else if ( strcmp( tok, "transform" ) == 0 )
		{
			shader.mFlags |= SF_TRANSFORM;

			// Get function type
			tok = strtok( NULL, " \n\t\r" );
      if ( strcmp( tok, "sin" ) == 0 || strcmp( tok, "sine" ) == 0 )
        shader.mXferFunction.mFlags = SF_SINE;

      // Function params
      tok = strtok( NULL, " \n\t\r" );
			sscanf( tok, "%f", &shader.mXferFunction.mArgs[ 0 ] );	// Freq
      tok = strtok( NULL, " \n\t\r" );
			sscanf( tok, "%f", &shader.mXferFunction.mArgs[ 1 ] );	// Amplitude
		}
		else if ( strcmp( tok, "no_cull" ) == 0 )
		{
			shader.mFlags |= SF_NO_CULL;
		}
		else if ( strcmp( tok, "invisible" ) == 0 )
		{
			shader.mFlags |= SF_INVISIBLE;
		}
		else if ( strcmp( tok, "pass_begin" ) == 0 )
		{
			pass.reset();

			while ( fgets( line, 500, apFile ) != NULL )
			{
				aLineCount++;
				char *tok = strtok( line, " \n\t\r" );
				if ( !tok || tok[ 0 ] == '-' || tok[ 0 ] == '#' )
					continue;
			
				if ( strcmp( tok, "pass_end" ) == 0 )
				{		
					if ( numTextures == 0 )
					{
						unsigned int flags = ( clamptex ) ? texture_manager_t::CLAMP : 0;
		
						pass.mTextures.push_back( mTextures->insert( texture, flags ) );
					}
					else
					{
						char *ext, baseName[ 100 ], texName[ 100 ];
						ext = COM_FileExtension( texture.c_str() );
						COM_StripExtension( texture.c_str(), baseName );
						for ( int i = 0; i < numTextures; i++ )
						{	  
							sprintf( texName, "%s%d.%s", baseName, i, ext );
							pass.mTextures.push_back( mTextures->insert( texName ) );
						}
					}

					shader.passes.push_back( pass );
					break;
				}
				else if ( strcmp( tok, "mat" ) == 0 )
				{
					tok = strtok( NULL, " \n\t\r" );
					sscanf( tok, "%f", &pass.mRGB );
				}
				else if ( strcmp( tok, "rgbfunc" ) == 0 )
				{
					pass.mFlags |= SF_RGBGEN;

					tok = strtok( NULL, " \n\t\r" );
					if ( strcmp( tok, "square" ) == 0 )
					{
						pass.mFlags |= SF_SQUARE;
					}
					else if ( strcmp( tok, "triangle" ) == 0 )
					{
						pass.mFlags |= SF_TRIANGLE;
					}
					else if ( strcmp( tok, "sawtooth" ) == 0 )
					{
						pass.mFlags |= SF_SAWTOOTH;
					}
					else if ( strcmp( tok, "invsawtooth" ) == 0 )
					{
						pass.mFlags |= SF_INVSAWTOOTH;
					}
					else if ( strcmp( tok, "sin" ) == 0 )
					{
						pass.mFlags |= SF_SINE;
					}

					// Get function params
					tok = strtok( NULL, " \n\t\r" );
					sscanf( tok, "%f", &pass.rgbData.d0.fVal );	// RGB min

					tok = strtok( NULL, " \n\t\r" );
					sscanf( tok, "%f", &pass.rgbData.d1.fVal );	// RGB amplitude

					tok = strtok( NULL, " \n\t\r" );
					sscanf( tok, "%f", &pass.rgbData.d2.fVal );	// RGB period

					tok = strtok( NULL, " \n\t\r" );
					sscanf( tok, "%f", &pass.rgbData.d3.fVal );	// RGB increment
				}
				else if ( strcmp( tok, "blend_func" ) == 0 )
				{
					parseBlendFunc( pass );
				}
				else if ( strcmp( tok, "alpha_func" ) == 0 )
				{
					pass.mFlags |= SF_ALPHA;
				}
				else if ( strcmp( tok, "tex" ) == 0 )
				{
					texture = strtok( NULL, " \n\t\r" );
				}
				else if ( strcmp( tok, "clamptex" ) == 0 )
				{
					clamptex = true;
					texture = strtok( NULL, " \n\t\r" );
				}
				else if ( strcmp( tok, "cycletex" ) == 0 )
				{
					tok = strtok( NULL, " \n\t\r" );
					sscanf( tok, "%d", &numTextures ); // Number of textures
					pass.cycleData.d0.iVal = numTextures;
		
					tok = strtok( NULL, " \n\t\r" );
					sscanf( tok, "%d", &pass.cycleData.d1.iVal ); // Cycle interval
	
					pass.mFlags |= SF_TEXCYCLE;
				}
				else if ( strcmp( tok, "tcfunc" ) == 0 )
				{
					ShaderPassData passData;

					pass.mFlags |= SF_TCMOD;
					tok = strtok( NULL, " \n\t\r" );
					if ( strcmp( tok, "perturb" ) == 0 )
					{
					    tok = strtok( NULL, " \n\t\r" );
						sscanf( tok, "%f", &passData.d0.fVal );
						
					    tok = strtok( NULL, " \n\t\r" );
						sscanf( tok, "%f", &passData.d1.fVal );

					    tok = strtok( NULL, " \n\t\r" );
						sscanf( tok, "%f", &passData.d2.fVal );
	
					    tok = strtok( NULL, " \n\t\r" );
						sscanf( tok, "%f", &passData.d3.fVal );

					    pass.tcFlags.push_back( SF_PERTURB );
					}
					else if ( strcmp( tok, "scroll" ) == 0 )
					{
						tok = strtok( NULL, " \n\t\r" );
						sscanf( tok, "%f", &passData.d0.fVal ); // X increment

					    tok = strtok( NULL, " \n\t\r" );
						sscanf( tok, "%f", &passData.d1.fVal ); // Y increment
		
					    pass.tcFlags.push_back( SF_SCROLL );
					}
					else if ( strcmp( tok, "rotate" ) == 0 )
					{
					    tok = strtok( NULL, " \n\t\r" );
						sscanf( tok, "%f", &passData.d0.fVal ); // Angle increment
	
					    pass.tcFlags.push_back( SF_ROTATE );
					}
					else if ( strcmp( tok, "scale" ) == 0 )
					{
					    tok = strtok( NULL, " \n\t\r" );
						sscanf( tok, "%f", &passData.d0.fVal ); // Scale increment
					
					    tok = strtok( NULL, " \n\t\r" );
						sscanf( tok, "%f", &passData.d1.fVal ); // Scale min
					
					    tok = strtok( NULL, " \n\t\r" );
						sscanf( tok, "%f", &passData.d2.fVal ); // Scale max
	
					    pass.tcFlags.push_back( SF_SCALE );
					}

					pass.tcData.push_back( passData );
				}
				else if ( strcmp( tok, "gentc" ) == 0 )
				{
					pass.mFlags |= SF_TCGEN;
					shader.mFlags |= SF_TCGEN;
				}
			}
		}
		else
		{
			gLog->add_msg( "-- Error parsing shader file on line %d", aLineCount );
		}
	}

	insert( shader );
	
	return true;
}

//////////////////////////////////////////////////////////////////////
void
shader_manager_t::parseBlendFunc( ShaderPass &pass )
{
	pass.mFlags |= SF_BLEND;
	char *tok = strtok( NULL, " \n\t\r" );

	if ( strcmp( tok, "modulate" ) == 0 )
	{
		pass.mSrc = GL_DST_COLOR;
		pass.mDst = GL_ONE;
	}
	else if ( strcmp( tok, "add" ) == 0 )
	{
		pass.mSrc = GL_ONE;
		pass.mDst = GL_ONE;
	}
	else if ( strcmp( tok, "blend" ) == 0 )
	{
		pass.mSrc = GL_SRC_ALPHA;
		pass.mDst = GL_ONE_MINUS_SRC_ALPHA;
	}
	else if ( strcmp( tok, "replace" ) == 0 )
	{
		pass.mSrc = GL_ONE;
		pass.mDst = GL_ZERO;
	}
	else
	{
		pass.mSrc = getBlendFunc( tok );

		tok = strtok( NULL, " \n\t\r" );
		pass.mDst = getBlendFunc( tok );
	}	
}

//////////////////////////////////////////////////////////////////////
GLenum 
shader_manager_t::getBlendFunc( const char *tok )
{
  if ( strcmp( tok, "GL_ZERO" ) == 0 )
    return GL_ZERO;
  else if ( strcmp( tok, "GL_ONE" ) == 0 )
    return GL_ONE;
  else if ( strcmp( tok, "GL_SRC_COLOR" ) == 0 )
    return GL_SRC_COLOR;
  else if ( strcmp( tok, "GL_ONE_MINUS_SRC_COLOR" ) == 0 )
    return GL_ONE_MINUS_SRC_COLOR;
  else if ( strcmp( tok, "GL_DST_COLOR" ) == 0 )
    return GL_DST_COLOR;
  else if ( strcmp( tok, "GL_ONE_MINUS_DST_COLOR" ) == 0 )
    return GL_ONE_MINUS_DST_COLOR;
  else if ( strcmp( tok, "GL_SRC_ALPHA" ) == 0 )
    return GL_SRC_ALPHA;
  else if ( strcmp( tok, "GL_ONE_MINUS_SRC_ALPHA" ) == 0 )
    return GL_ONE_MINUS_SRC_ALPHA;
  else if ( strcmp( tok, "GL_DST_ALPHA" ) == 0 )
    return GL_DST_ALPHA;
  else if ( strcmp( tok, "GL_ONE_MINUS_DST_ALPHA" ) == 0 )
    return GL_ONE_MINUS_DST_ALPHA;

  return GL_ZERO;
}
