#ifndef _SHADER_
#define _SHADER_

#include <vector>
#include <string>
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "textures.h"
#include "log.h"

using namespace std;

enum { SF_TCMOD =			   0x00000001, 
	     SF_TCGEN =			   0x00000002, 
	     SF_BLEND =        0x00000004,
	     SF_DEPTH =        0x00000008,
	     SF_TEXCYCLE =     0x00000010,
	     SF_RGBGEN =       0x00000020,
	     SF_VIEWERFACING = 0x00000040,
	     SF_NO_CULL =      0x00000080,
	     SF_DEPTH_OFF =    0x00000100,
       SF_INVISIBLE =    0x00000200,
	     SF_ALPHA =        0x00000400,
       SF_TRANSFORM =    0x00000800       // Transform the vertices
};

// TC Functions
enum { SF_ROTATE =  0x00010000,
	     SF_SCROLL =  0x00020000,
	     SF_PERTURB = 0x00040000,
	     SF_SCALE  =  0x00080000 };

// RGB Functions
enum { SF_SQUARE =      0x01000000,
	     SF_TRIANGLE =    0x02000000,
       SF_SAWTOOTH =    0x04000000,
	     SF_INVSAWTOOTH = 0x08000000,
	     SF_SINE =        0x10000000 };

struct Mat
{
  Mat() : r( 1.0 ), g( 1.0 ), b( 1.0 ), a( 1.0 ) {}
  Mat( float pr, float pg, float pb, float pa ) :
    r( pr ), g( pg ), b( pb ), a( pa ) {}

  void set( float val ) { r = val; g = val; b = val; a = val; }

  float r, g, b, a;
};

class Textures;

//////////////////////////////////////////////////////////////////////
//                             BLENDER                              //
//                                                                  //
// Specifies the src and dest factors to be passed to glBlendFunc   //
//////////////////////////////////////////////////////////////////////
struct GLBlender
{
  GLBlender() : enable( false ) {}
  bool enable;
  GLenum src, dst;
};

union GenericData
{
	GenericData() : iVal( 0 ) {}

	float fVal;
	int iVal;
};

struct ShaderPassData
{
	ShaderPassData() {}
	GenericData d0, d1, d2, d3;
};

struct ShaderPass
{
	ShaderPass() : mFlags( 0 ) {}
	void reset() 
	{ 
		mFlags = 0; 
		mTextures.clear(); 
		mRGB = 0.7f; 
		
		mSrc = GL_ONE;
		mDst = GL_ZERO;
	}

	unsigned int mFlags;
	vector<unsigned int> mTextures;
	GLenum mSrc, mDst;
	float mRGB;
	ShaderPassData rgbData;
	ShaderPassData cycleData;
	vector<ShaderPassData> tcData;
	vector<unsigned int> tcFlags;
};

struct PolyPassData
{
	unsigned int mCurrTexture;
	ShaderPassData rgbData;
	ShaderPassData cycleData;
	vector<ShaderPassData> tcData;
	vector<unsigned int> tcFlags;
};

struct xfer_function_t
{
  unsigned int mFlags;
  float mArgs[ 2 ];
};

struct shader_t
{
	shader_t() : mFlags( 0 ) {}

	int getNumPasses() const { return passes.size(); }

	string mName;
	unsigned int mFlags;
	vector<ShaderPass> passes;

  xfer_function_t mXferFunction;
};

//////////////////////////////////////////////////
// class shader_manager_t
//////////////////////////////////////////////////
class shader_manager_t
{
public:
	shader_manager_t( texture_manager_t *aTextures = NULL );
	~shader_manager_t();

	void set_texture_manager( texture_manager_t *aTextures ) { mTextures = aTextures; }

	void clear(); 
	shader_t *insert( const shader_t &shader );
	shader_t *getDefault() { return shaders[ 0 ]; }
	shader_t *getShader( const string &name );

	bool readFromFile( const char *aFileName );

private:
	bool parseShader( FILE *apFile, const char *aShaderName, 
	                  int &aLineCount );
	void parseBlendFunc( ShaderPass &pass );
	GLenum getBlendFunc( const char *tok );

	vector<shader_t*> shaders;
	texture_manager_t *mTextures;
};

#endif
