#ifndef __SHIP__
#define __SHIP__

#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include <assert.h>

#include "globals.h"
#include "qsound.h"
#include "vect3.h"
#include "shadow_model.h"
#include "console.h"
#include "particle.h"

///////////////////////////////////////////////////////////////
class ship_t
{
public:
  ship_t() 
    { mpModel = 0; mpSmokeParticles = 0; mpWakeParticles = 0, mState = DEAD; }
	~ship_t() {}

  void set_id( int aID ) { mID = aID; }
  int get_id() const { return mID; }

  void reset();
	void draw();
	void draw_shadow( const vect3_t &aLightPos );
	void draw_shadow_two_pass( const vect3_t &aLightPos );
	void draw_debug();
	bool update( const vector<segment_t> &aSegments,
	             const vector<vect3_t> &aShipPoly,
               vector<ship_t> &aShips,
               const int aBattleRound );

	void set_pos( const vect3_t &aPos ) { mOrigPos = aPos; }
	const vect3_t &get_pos() const { return mPos; }

  void set_dir( const vect3_t &aDir );
	const vect3_t &get_dir() const { return mDir; }

	void set_vel( float aVel ) { mOrigVel = aVel; }
	float get_vel() const { return mVel; }

	float get_angle() const { return mAngle; }
  float get_angleinc() const { return mAngleInc; }

	void set_model( shadow_model_node_t *apModel ) { mpModel = apModel; }
  void set_particles( particle_sys_t *apParticles ) { mpSmokeParticles = apParticles; }
  void set_wake_particles( particle_sys_t *apParticles ) { mpWakeParticles = apParticles; }
  void set_shiptype( int shipType ) { mShipType = shipType; }
  int get_shiptype() const { return mShipType; }

  bool hit();
  int get_health() const { return mHealth; }

  bool is_alive() const { return !( mState == DEAD || mState == DIEING ); }

  void clear_ship_collide() { mCollidingWithShip = false; }
  void set_ship_collide() { mCollidingWithShip = true; }

private:

	enum { STRAIGHT = 0, TURNING, AVOIDING, DEAD, DIEING };

  bool check_collision( const vector<segment_t> &aSegments,
	                      const vector<vect3_t> &aShipPoly,
                        vector<ship_t> &aShips, 
                        float aMoveTime = 100.0f );

  void calc_dir();

  int mID;             // This ship's ID
	vect3_t mPos, mOrigPos;
	vect3_t mDir, mOrigDir;
	float mVel, mOrigVel;
	float mAngle, mOrigAngle;
	float mTargetAngle;
	float mAngleInc;
  float mAvoidAngle;    // The angle between colliding ships
  float mShootCount;

  bool mCollidingWithShip; // Is this ship colliding with another one?

  vect3_t mPredictPos;

	int mState;

	shadow_model_node_t *mpModel;

	float mTurnCount;

  float mTurnTime;

  int mHealth;
  int mShipType;

  particle_sys_t *mpSmokeParticles;
  particle_sys_t *mpWakeParticles;
};

#endif
