#include "textbox.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

textbox_t::textbox_t()
{
  reset();
}

textbox_t::~textbox_t()
{

}


//////////////////////////////////////////////////////////////////////
// This function handles keystrokes
//////////////////////////////////////////////////////////////////////
bool 
textbox_t::add_char( unsigned short aKey )
{
	if ( !mEnabled )
		return false;

	//+ is this check needed here?  add() should get this key
  if ( aKey == VK_RETURN )
	{
		mEnabled = false;
		return true;
	}

	if ( isprint( aKey ) )
	{
		mTextInput[ mCursorPos++ ] = aKey;
		
		if ( mCursorPos > mMaxPos )
			mMaxPos = mCursorPos;
		
		mTextInput[ mMaxPos ] = '\0';

    if ( mCursorPos > (TEXT_WIDTH - 2) )
			mCursorPos = TEXT_WIDTH - 2;    // Keep from exceeding the array size
	}

  return false;
}


//////////////////////////////////////////////////////////////////////
// This function handles keystrokes
//////////////////////////////////////////////////////////////////////
bool 
textbox_t::add( unsigned short aKey )
{
	if ( !mEnabled )
		return false;

	switch ( aKey )
	{
	case VK_BACK : 
		if ( mCursorPos > 0 )
		{
			mTextInput[ --mCursorPos ] = ' ';
			for ( int i = mCursorPos + 1; i < mMaxPos; i++ )
				mTextInput[ i ] = mTextInput[ i + 1 ];
			mMaxPos--;
		}
		break;
	case VK_DELETE : 
		if ( mMaxPos > mCursorPos )
		{
			for ( int i = mCursorPos; i < mMaxPos; i++ )
				mTextInput[ i ] = mTextInput[ i + 1 ];
			mMaxPos--;
		}
		break;
	case VK_LEFT :
		if ( mCursorPos > 0 )
		{
			--mCursorPos;
		}
		break;
	case VK_RIGHT :
		if ( mCursorPos < mMaxPos )
		{
			++mCursorPos;
		}
		break;
	case VK_RETURN :
    mEnabled = false;
		return true;
	}

  return false;
}


//////////////////////////////////////////////////////////////////////
// Reset the text box
//////////////////////////////////////////////////////////////////////
void 
textbox_t::reset()
{
  mCursorPos = 0;
  mMaxPos = 0;
  mEnabled = false;
  mCursorCount = 0.0f;

  mTextInput[ 0 ] = '\0';
}


//////////////////////////////////////////////////////////////////////
// Draw the text
//////////////////////////////////////////////////////////////////////
void 
textbox_t::draw_text( font_t *apFont, int aX, int aY )
{	
	apFont->setDefaultDim();
  apFont->print( aX, aY, mTextInput );

  mCursorCount += gFactTick;

	if ( mCursorCount < 50.0f )
		apFont->print( apFont->getSpace() * mCursorPos + aX, aY, "_" );
	else if ( mCursorCount >= 100.0f )
		mCursorCount = 0.0f;
}


//////////////////////////////////////////////////////////////////////
// Return the text as an argument
//////////////////////////////////////////////////////////////////////
void 
textbox_t::get_text( char aText[] )
{	
  strcpy( aText, mTextInput );
}


