#ifndef _TEXTURES_
#define _TEXTURES_

#include <stdlib.h>
#include <stdio.h>
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <string>
#include <list>
#include <vector>

#include "glext.h"
#include "ijl.h"
#include "misc.h"
#include "log.h"

using namespace std;

struct CTexture
{
  string name;  
  unsigned int number;
  GLint sizeX, sizeY;
  GLint components;
  unsigned char *data;
};

struct CTGAHeader
{
  unsigned char id_length;
  unsigned char color_map_type;
  unsigned char image_type;
  unsigned char color_map_spec[ 5 ];

  unsigned short xorigin, yorigin;
  unsigned short width;
  unsigned short height;
  unsigned char pixel_depth;
  unsigned char image_descriptor;
};

//////////////////////////////////////////////////
// CLASS TEXTURES
//////////////////////////////////////////////////
class texture_manager_t
{
public:
  texture_manager_t() {}
  ~texture_manager_t();

  void clear();  
  int insert( const string &tName, unsigned flags = 0 );
  void remove( const string &tName );

  bool loadBMP( const char *fileName, CTexture *tex );
  bool loadBMP( FILE *file, CTexture *tex );
  bool loadTGA( const char *fileName, CTexture *tex );
  bool loadJPG( const char *fileName, CTexture *tex );
  void saveBMP( const char *fileName, int numComponents, int w, int h,
		const unsigned char *data );

  void set_filtering( int aFiltering ) { mFiltering = aFiltering; }

  enum { NEAREST = 1, 
	     CLAMP = 2 };

  enum { BILINEAR = 0, TRILINEAR };
private:

  list<CTexture> tList;
  int mFiltering;       // Used to specify bilinear or trilinear filtering
};

#endif
