#ifndef __VECT_T__
#define __VECT_T__

#include <math.h>
#include <string>
#include <assert.h>

#include <windows.h>

#include <vector>

#include <GL/gl.h>
#include <GL/glu.h>

#include "fmath.h"

using namespace std;

#define _2PI_RAD 6.2830

inline float deg2rad( float ang );
inline float rad2deg( float ang );

//////////////////////////////////////////////////////////////////////
//                              Vector                              //
//////////////////////////////////////////////////////////////////////
struct vect3_t
{
	vect3_t( float val = 0.0f ) 
	{
		mVect[ 0 ] = val;  
		mVect[ 1 ] = val;  
		mVect[ 2 ] = val;  
	}

	vect3_t( float val0, float val1, float val2 ) 
	{
		mVect[ 0 ] = val0;  
		mVect[ 1 ] = val1;  
		mVect[ 2 ] = val2;  
	}

	vect3_t( int val ) 
	{
		mVect[ 0 ] = ( float ) val;  
		mVect[ 1 ] = ( float ) val;  
		mVect[ 2 ] = ( float ) val;  
	}

	vect3_t( float *apVal ) 
	{
		assert( apVal );
		mVect[ 0 ] = apVal[ 0 ];  
		mVect[ 1 ] = apVal[ 1 ];   
		mVect[ 2 ] = apVal[ 2 ];  
	}

	vect3_t( const vect3_t &aVect )
	{
		mVect[ 0 ] = aVect.mVect[ 0 ];
		mVect[ 1 ] = aVect.mVect[ 1 ];
		mVect[ 2 ] = aVect.mVect[ 2 ];
	}

	const vect3_t &operator=( const vect3_t &aVect )
	{
		if ( this != &aVect )
		{
			mVect[ 0 ] = aVect.mVect[ 0 ];
			mVect[ 1 ] = aVect.mVect[ 1 ];
			mVect[ 2 ] = aVect.mVect[ 2 ];
		}
		return *this;
	}

	float operator[]( int i ) const 
	{ 
		assert( i >= 0 && i < 3 );
		return mVect[ i ];
	}

	float &operator[]( int i ) 
	{ 
		assert( i >= 0 && i < 3 );
		return mVect[ i ];
	}

	const vect3_t &operator=( float aVal )
	{
		mVect[ 0 ] = aVal;
		mVect[ 1 ] = aVal;
		mVect[ 2 ] = aVal;
		return *this;
	}

	const vect3_t &operator=( float *apVal )
	{
		mVect[ 0 ] = apVal[ 0 ];
		mVect[ 1 ] = apVal[ 1 ];
		mVect[ 2 ] = apVal[ 2 ];
		return *this;
	}

	const vect3_t &operator-=( const vect3_t &aVect )
	{
		mVect[ 0 ] -= aVect.mVect[ 0 ];
		mVect[ 1 ] -= aVect.mVect[ 1 ];
		mVect[ 2 ] -= aVect.mVect[ 2 ];
		return *this;
	}

	vect3_t operator-( const vect3_t &aVect ) const
	{
		vect3_t lVect = *this;
		return lVect -= aVect;
	}

	vect3_t operator-( float *apVal ) const
	{
		vect3_t lVect = *this;
		lVect.mVect[ 0 ] -= apVal[ 0 ];
		lVect.mVect[ 1 ] -= apVal[ 1 ];
		lVect.mVect[ 2 ] -= apVal[ 2 ];
		return lVect;
	}

	const vect3_t &operator+=( const vect3_t &aVect )
	{
		mVect[ 0 ] += aVect.mVect[ 0 ];
		mVect[ 1 ] += aVect.mVect[ 1 ];
		mVect[ 2 ] += aVect.mVect[ 2 ];
		return *this;
	}

	vect3_t operator+( const vect3_t &aVect ) const
	{
		vect3_t lVect = *this;
		return lVect += aVect;
	}

	const vect3_t &operator*=( const vect3_t &aVect )
	{
		mVect[ 0 ] *= aVect.mVect[ 0 ];
		mVect[ 1 ] *= aVect.mVect[ 1 ];
		mVect[ 2 ] *= aVect.mVect[ 2 ];
		return *this;
	}

	vect3_t operator*( const vect3_t &aVect ) const
	{
		vect3_t lVect = *this;
		return lVect *= aVect;
	}

	const vect3_t &operator*=( float aVal )
	{
		mVect[ 0 ] *= aVal;
		mVect[ 1 ] *= aVal;
		mVect[ 2 ] *= aVal;
		return *this;
	}

	vect3_t operator*( float aVal ) const
	{
		vect3_t lVect = *this;
		return lVect *= aVal;
	}

	const vect3_t &operator/=( float aVal )
	{
		x /= aVal;
		y /= aVal;
		z /= aVal;
		return *this;
	}

	vect3_t operator/( float aVal ) const
	{
		vect3_t lVect = *this;
		return lVect /= aVal;
	}
	
	void negate()
	{
		x *= -1.0f;
		y *= -1.0f;
		z *= -1.0f;
	}

	float dot( const vect3_t &r ) const
	{
	    return ( x * r.x + 
			     y * r.y + 
				 z * r.z );
	}

	//float mag() const { return fsqrt( dot( *this ) ); }
	float mag() const { return sqrt( dot( *this ) ); }
	float magSquared() const { return dot( *this ); }
	vect3_t cross( const vect3_t &r ) const
	{
	    vect3_t t;
	    t.x = y * r.z - z * r.y;
	    t.y = z * r.x - x * r.z;
	    t.z = x * r.y - y * r.x;
	    return t;
	}

	//////////////////////////////////////////////////////////////////////
	// Angle between two vectors
	//////////////////////////////////////////////////////////////////////
	float angle( const vect3_t &r ) const
	{
		float dp = dot( r );
		float prod = mag() * r.mag();
		return acos( dp / prod );
	}

	//////////////////////////////////////////////////////////////////////
	// Cosine of the angle between two vectors
	//////////////////////////////////////////////////////////////////////
	float cosAngle( const vect3_t &r ) const
	{
		float dp = dot( r );
		float prod = mag() * r.mag();
		return dp / prod;
	}

	void normalize()
	{
		static float d, lMag;
		lMag = mag();
		FP_INV2( d, lMag ); // d = 1 / lMag
		x *= d;
		y *= d;
		z *= d;
	}

	float distance( const vect3_t &v ) const
	{
		float diffx2 = v.x - x;
		diffx2 *= diffx2;
		float diffy2 = v.y - y;
		diffy2 *= diffy2;
		float diffz2 = v.z - z;
		diffz2 *= diffz2;
		return sqrt( diffx2 + diffy2 + diffz2 );
	}

	float distanceSquared( const vect3_t &v ) const
	{
		float diffx2 = v.x - x;
		diffx2 *= diffx2;
		float diffy2 = v.y - y;
		diffy2 *= diffy2;
		float diffz2 = v.z - z;
		diffz2 *= diffz2;
		return diffx2 + diffy2 + diffz2;
	}

	bool equal( const vect3_t &v ) const
	{
		return x == v.x && y == v.y && z == v.z;
		
		static const float e = 0.0000001f;

		static float dx, dy, dz;

		dx = fastabs( x - v.x );
		dy = fastabs( y - v.y );
		dz = fastabs( z - v.z );

		return ( dx < e && dy < e && dz < e );
	}

	// Rotate about Y axis
	void rotateY( float aAngle )
	{
		x = x * cos( aAngle ) - z * sin( aAngle );
		z = z * cos( aAngle ) + x * sin( aAngle );
	}

	// Rotate about Z axis
	void rotateZ( float aAngle )
	{
		x = x * cos( aAngle ) - y * sin( aAngle );
		y = y * cos( aAngle ) + x * sin( aAngle );
	}

	// Rotate about X axis
	void rotateX( float aAngle )
	{
		y = y * cos( aAngle ) - z * sin( aAngle );
		z = z * cos( aAngle ) + y * sin( aAngle );
	}

	union
	{
		struct
		{
			float x, y, z;
		};
		float mVect[ 3 ];
	};
};

//////////////////////////////////////////////////////////////////////
inline void add( const vect3_t &v0, const vect3_t &v1, vect3_t &res );
inline void sub( const vect3_t &v0, const vect3_t &v1, vect3_t &res );
inline void mult( const vect3_t &v0, const vect3_t &v1, vect3_t &res );
inline void div( const vect3_t &v0, const vect3_t &v1, vect3_t &res );
inline void mult( const vect3_t &v0, float c, vect3_t &res );
inline void div( const vect3_t &v0, float c, vect3_t &res );
inline void cross( const vect3_t &l, const vect3_t &r, vect3_t &res );
inline void reflect( const vect3_t &aRay, const vect3_t &aNorm, 
			               vect3_t &aReflected );
inline void muladd( const vect3_t &l, float c, const vect3_t &r, vect3_t &res );
inline void muladd( const vect3_t &v0, const vect3_t &v1, const vect3_t &v2, 
                    vect3_t &res );

// Distance from point to line segment
float dist_from_line_seg( const vect3_t &aP, 
				                  const vect3_t &aStart, const vect3_t &aEnd );
void mult_look_at( const vect3_t &aEye,
		               const vect3_t &aLookat );
void multLookAt( float eyex, float eyey, float eyez,
            	   float atx, float aty, float atz,
	               float upx, float upy, float upz );

//////////////////////////////////////////////////////////////////////
//                              PLANE                               //
//////////////////////////////////////////////////////////////////////
struct plane_t
{
	float distanceToPoint( const vect3_t &p ) const
	{
		return ( mNorm.x * p.x + mNorm.y * p.y + 
			     mNorm.z * p.z + d );
	}

	bool pointOnPlane( const vect3_t &p ) const
	{ 
		//return distanceToPoint( p ) == 0.0;
		return distanceToPoint( p ) <= 0.01;
	}

	void rayIntersection( const vect3_t &pos, const vect3_t &ray,
		                  vect3_t &intersect ) const
	{
		static float_t dis;
		dis.f = mNorm[ 0 ] * ray[ 0 ] + mNorm[ 1 ] * ray[ 1 ] + mNorm[ 2 ] * ray[ 2 ];
		
		if ( dis.i == 0 ) 
		{
			intersect = pos;
			return;	//error line parallel to plane
		}
		
		static float distDiv, temp;
		//distDiv = distanceToPoint( pos ) / dis.f;
		FP_INV2( temp, dis.f );
		distDiv = distanceToPoint( pos ) * temp;

		static vect3_t prod;
		mult( ray, distDiv, prod );
		sub( pos, prod, intersect );

		//return pos - ray * ( distanceToPoint( pos ) / dis );
	}

	vect3_t mNorm;
	float d;
};

//////////////////////////////////////////////////////////////////////
//                              SPHERE                              //
//////////////////////////////////////////////////////////////////////
struct sphere_t
{
	sphere_t() : mCenter( 0 ) { mRad.f = 0.0f; }

  void set_rad( float aRad ) { mRad.f = aRad; }

  void square_rad() { mRad2.f = mRad.f * mRad.f; }

	vect3_t mCenter;
	float_t mRad, mRad2;
};

//////////////////////////////////////////////////////////////////////
//                     LINE SEGMENT                                 //
//////////////////////////////////////////////////////////////////////
class segment_t
{
public:

	void set_point( int aNum, const vect3_t &aPoint )
	{
		assert( aNum < 0 || aNum < 2 );
		mPoints[ aNum ] = aPoint;

		if ( aNum == 1 )
    {
      mP1MinusP0 = mPoints[ 1 ] - mPoints[ 0 ];

      // Calculate magnitude
			mag();
    }
	}

	float dist2_from_line_seg( const vect3_t &aP ) const;

  float dist_from_line_seg( const vect3_t &aP ) const 
    { return sqrt( dist2_from_line_seg( aP ) ); }

	const vect3_t &get_point( int aNum ) const
	{
		assert( aNum < 0 || aNum < 2 );
		return mPoints[ aNum ];
	}

  const vect3_t &get_vector() const { return mP1MinusP0; }

private:

	void mag() 
  { 
    mMag = mPoints[ 0 ].distance( mPoints[ 1 ] ); 
    mRecipMag2 = 1.0f / ( mMag * mMag );
  }

	vect3_t mPoints[ 2 ]; // 0 = start, 1 = end
	float mMag;           // Magnitude of lin segment
  float mRecipMag2;     // Reciprocal of the mag squared
  vect3_t mP1MinusP0;   // Unormalized vector from P0 to P1;
};

bool inpoly( const vector<vect3_t> &poly,
             unsigned int xt,
             unsigned int yt );

//////////////////////////////////////////////////////////////////////
//                              MATRIX                              //
//////////////////////////////////////////////////////////////////////
struct mat4_t
{
  mat4_t() {}
  ~mat4_t() {}

  void set_id()
  {
    a00 = a11 = a22 = a33 = 1;
    a10 = a20 = a30 = 0;
    a01 = a21 = a31 = 0;
    a02 = a12 = a32 = 0;
    a03 = a13 = a23 = 0;
  }
  void set_rot( float theta, const vect3_t &v );
  void set_trans( const vect3_t &aPos )
  {
    a03 = aPos.x;
    a13 = aPos.y;
    a23 = aPos.z;
  }

  union 
  {
    struct 
    {
      float a00, a10, a20, a30;
      float a01, a11, a21, a31;
      float a02, a12, a22, a32;
      float a03, a13, a23, a33;
    };
    float mData[ 16 ];
  };
};

// B = transpose( A )
inline mat4_t &transpose( mat4_t &B, const mat4_t &A );
// B = inverse( A )
inline mat4_t &invert( mat4_t &B, const mat4_t &A );
inline mat4_t &invert_fast( mat4_t &B, const mat4_t &A );

inline vect3_t &mult( vect3_t &u, const mat4_t &M, const vect3_t &v );

inline void mult( float *M, float *v, float *res );

inline void mult( mat4_t &C, const mat4_t &A, const mat4_t &B );

inline float det2x2( float a1, float a2, float b1, float b2 );
inline float det3x3( float a1, float a2, float a3, 
                     float b1, float b2, float b3, 
                     float c1, float c2, float c3 );

//////////////////////////////////////////////////////////////////////
vect3_t &
mult( vect3_t &u, const mat4_t &M, const vect3_t &v )
{
  u.x = M.a00 * v.x + M.a01 * v.y + M.a02 * v.z + M.a03;
  u.y = M.a10 * v.x + M.a11 * v.y + M.a12 * v.z + M.a13;
  u.z = M.a20 * v.x + M.a21 * v.y + M.a22 * v.z + M.a23;
  return u;
}

//////////////////////////////////////////////////////////////////////
mat4_t &
transpose( mat4_t &B, const mat4_t &A )
{
  B.a00 = A.a00; B.a01 = A.a10; B.a02 = A.a20; B.a03 = A.a30;
  B.a10 = A.a01; B.a11 = A.a11; B.a12 = A.a21; B.a13 = A.a31;
  B.a20 = A.a02; B.a21 = A.a12; B.a22 = A.a22; B.a23 = A.a32;
  B.a30 = A.a03; B.a31 = A.a13; B.a32 = A.a23; B.a33 = A.a33;
  return B;
}

void 
mult( float *M, float *v, float *res )
{
	res[ 0 ] = M[ 0 ] * v[ 0 ] + M[ 4 ] * v[ 1 ] + M[ 8 ] * v[ 2 ] + M[ 12 ] * v[ 3 ];
	res[ 1 ] = M[ 1 ] * v[ 0 ] + M[ 5 ] * v[ 1 ] + M[ 9 ] * v[ 2 ] + M[ 13 ] * v[ 3 ];
	res[ 2 ] = M[ 2 ] * v[ 0 ] + M[ 6 ] * v[ 1 ] + M[10 ] * v[ 2 ] + M[ 14 ] * v[ 3 ];
	res[ 3 ] = M[ 3 ] * v[ 0 ] + M[ 7 ] * v[ 1 ] + M[11 ] * v[ 2 ] + M[ 15 ] * v[ 3 ];
}

// C = A * B
void 
mult( mat4_t &C, const mat4_t &A, const mat4_t &B )
{
  C.a00 = B.a00 * A.a00 + B.a01 * A.a10 + B.a02 * A.a20 + B.a03 * A.a30;
  C.a10 = B.a10 * A.a00 + B.a11 * A.a10 + B.a12 * A.a20 + B.a13 * A.a30;
  C.a20 = B.a20 * A.a00 + B.a21 * A.a10 + B.a22 * A.a20 + B.a23 * A.a30;
  C.a30 = B.a30 * A.a00 + B.a31 * A.a10 + B.a32 * A.a20 + B.a33 * A.a30;
  C.a01 = B.a00 * A.a01 + B.a01 * A.a11 + B.a02 * A.a21 + B.a03 * A.a31;
  C.a11 = B.a10 * A.a01 + B.a11 * A.a11 + B.a12 * A.a21 + B.a13 * A.a31;
  C.a21 = B.a20 * A.a01 + B.a21 * A.a11 + B.a22 * A.a21 + B.a23 * A.a31;
  C.a31 = B.a30 * A.a01 + B.a31 * A.a11 + B.a32 * A.a21 + B.a33 * A.a31;
  C.a02 = B.a00 * A.a02 + B.a01 * A.a12 + B.a02 * A.a22 + B.a03 * A.a32;
  C.a12 = B.a10 * A.a02 + B.a11 * A.a12 + B.a12 * A.a22 + B.a13 * A.a32;
  C.a22 = B.a20 * A.a02 + B.a21 * A.a12 + B.a22 * A.a22 + B.a23 * A.a32;
  C.a32 = B.a30 * A.a02 + B.a31 * A.a12 + B.a32 * A.a22 + B.a33 * A.a32;
  C.a03 = B.a00 * A.a03 + B.a01 * A.a13 + B.a02 * A.a23 + B.a03 * A.a33;
  C.a13 = B.a10 * A.a03 + B.a11 * A.a13 + B.a12 * A.a23 + B.a13 * A.a33;
  C.a23 = B.a20 * A.a03 + B.a21 * A.a13 + B.a22 * A.a23 + B.a23 * A.a33;
  C.a33 = B.a30 * A.a03 + B.a31 * A.a13 + B.a32 * A.a23 + B.a33 * A.a33;
}

/*
    calculate the determinent of a 2x2 matrix in the from

    | a1 a2 |
    | b1 b2 |

*/
float 
det2x2( float a1, float a2, float b1, float b2 )
{
  return a1 * b2 - b1 * a2;
}

/*
    calculate the determinent of a 3x3 matrix in the from

    | a1 a2 a3 |
    | b1 b2 b3 |
    | c1 c2 c3 |

*/
float 
det3x3( float a1, float a2, float a3, 
        float b1, float b2, float b3, 
        float c1, float c2, float c3 )
{
  return a1 * det2x2( b2, b3, c2, c3 ) - 
         b1 * det2x2( a2, a3, c2, c3 ) + 
         c1 * det2x2( a2, a3, b2, b3 );
}

mat4_t & 
invert( mat4_t &B, const mat4_t &A )
{
  float det, oodet;

  B.a00 =  det3x3(A.a11, A.a21, A.a31, A.a12, A.a22, A.a32, A.a13, A.a23, A.a33);
  B.a10 = -det3x3(A.a10, A.a20, A.a30, A.a12, A.a22, A.a32, A.a13, A.a23, A.a33);
  B.a20 =  det3x3(A.a10, A.a20, A.a30, A.a11, A.a21, A.a31, A.a13, A.a23, A.a33);
  B.a30 = -det3x3(A.a10, A.a20, A.a30, A.a11, A.a21, A.a31, A.a12, A.a22, A.a32);

  B.a01 = -det3x3(A.a01, A.a21, A.a31, A.a02, A.a22, A.a32, A.a03, A.a23, A.a33);
  B.a11 =  det3x3(A.a00, A.a20, A.a30, A.a02, A.a22, A.a32, A.a03, A.a23, A.a33);
  B.a21 = -det3x3(A.a00, A.a20, A.a30, A.a01, A.a21, A.a31, A.a03, A.a23, A.a33);
  B.a31 =  det3x3(A.a00, A.a20, A.a30, A.a01, A.a21, A.a31, A.a02, A.a22, A.a32);

  B.a02 =  det3x3(A.a01, A.a11, A.a31, A.a02, A.a12, A.a32, A.a03, A.a13, A.a33);
  B.a12 = -det3x3(A.a00, A.a10, A.a30, A.a02, A.a12, A.a32, A.a03, A.a13, A.a33);
  B.a22 =  det3x3(A.a00, A.a10, A.a30, A.a01, A.a11, A.a31, A.a03, A.a13, A.a33);
  B.a32 = -det3x3(A.a00, A.a10, A.a30, A.a01, A.a11, A.a31, A.a02, A.a12, A.a32);

  B.a03 = -det3x3(A.a01, A.a11, A.a21, A.a02, A.a12, A.a22, A.a03, A.a13, A.a23);
  B.a13 =  det3x3(A.a00, A.a10, A.a20, A.a02, A.a12, A.a22, A.a03, A.a13, A.a23);
  B.a23 = -det3x3(A.a00, A.a10, A.a20, A.a01, A.a11, A.a21, A.a03, A.a13, A.a23);
  B.a33 =  det3x3(A.a00, A.a10, A.a20, A.a01, A.a11, A.a21, A.a02, A.a12, A.a22);

  det = (A.a00 * B.a00) + (A.a01 * B.a10) + (A.a02 * B.a20) + (A.a03 * B.a30);

  oodet = 1.0f / det;

  B.a00 *= oodet;
  B.a10 *= oodet;
  B.a20 *= oodet;
  B.a30 *= oodet;

  B.a01 *= oodet;
  B.a11 *= oodet;
  B.a21 *= oodet;
  B.a31 *= oodet;

  B.a02 *= oodet;
  B.a12 *= oodet;
  B.a22 *= oodet;
  B.a32 *= oodet;

  B.a03 *= oodet;
  B.a13 *= oodet;
  B.a23 *= oodet;
  B.a33 *= oodet;

  return B;
}

////////////////////////////////////////////////////////////////////////////
#define MAT(m,r,c) (m)[(c)*4+(r)]

mat4_t & 
invert_fast( mat4_t &B, const mat4_t &A )
{
  B.a03 = -A.a03;
  B.a13 = -A.a13;
  B.a23 = -A.a23;

  B.a00 = A.a00; B.a01 = A.a10; B.a02 = A.a20;
  B.a10 = A.a01; B.a11 = A.a11; B.a12 = A.a21;
  B.a20 = A.a02; B.a21 = A.a12; B.a22 = A.a22;

  return B;
}

///////////////////////////////////////////////////////////////////
void 
add( const vect3_t &v0, const vect3_t &v1, vect3_t &res )
{
	res.x = v0.x + v1.x;
	res.y = v0.y + v1.y;
	res.z = v0.z + v1.z;
}

//////////////////////////////////////////////////////////////////////
void 
sub( const vect3_t &v0, const vect3_t &v1, vect3_t &res )
{
	res.x = v0.x - v1.x;
	res.y = v0.y - v1.y;
	res.z = v0.z - v1.z;
}

//////////////////////////////////////////////////////////////////////
void 
mult( const vect3_t &v0, const vect3_t &v1, vect3_t &res )
{
	res.x = v0.x * v1.x;
	res.y = v0.y * v1.y;
	res.z = v0.z * v1.z;
}

//////////////////////////////////////////////////////////////////////
void 
div( const vect3_t &v0, const vect3_t &v1, vect3_t &res )
{
	res.x = v0.x / v1.x;
	res.y = v0.y / v1.y;
	res.z = v0.z / v1.z;
}

//////////////////////////////////////////////////////////////////////
void 
mult( const vect3_t &v0, float c, vect3_t &res )
{
	res.x = v0.x * c;
	res.y = v0.y * c;
	res.z = v0.z * c;
}

//////////////////////////////////////////////////////////////////////
void 
div( const vect3_t &v0, float c, vect3_t &res )
{ 
  float recip = 1.0f / c;
	res.x = v0.x * recip;
	res.y = v0.y * recip;
	res.z = v0.z * recip;
}

void
cross( const vect3_t &l, const vect3_t &r, vect3_t &res )
{
    res.x = l.y * r.z - l.z * r.y;
    res.y = l.z * r.x - l.x * r.z;
    res.z = l.x * r.y - l.y * r.x;
}

//////////////////////////////////////////////////////////////////////
void 
muladd( const vect3_t &l, float c, const vect3_t &r, vect3_t &res )
{
	res.x = l.x * c + r.x;
	res.y = l.y * c + r.y;
	res.z = l.z * c + r.z;
}

//////////////////////////////////////////////////////////////////////
void 
muladd( const vect3_t &v0, const vect3_t &v1, const vect3_t &v2, vect3_t &res )
{
	res.x = v0.x * v1.x + v2.x;
	res.y = v0.y * v1.y + v2.y;
	res.z = v0.z * v1.z + v2.z;
}

/////////////////////////////////////////////////////////////////////
void reflect( const vect3_t &aRay, const vect3_t &aNorm, 
			  vect3_t &aReflected )
{
   vect3_t vn = aNorm * ( aNorm.dot( aRay ) );
   vect3_t vt = aRay - vn;
   aReflected = vt - vn;

   //dlog->InsertMsg( "i %f %f %f, r %f %f %f", aRay.x, aRay.y, aRay.z, 
   //                 aReflected.x, aReflected.y, aReflected.z );
}

/////////////////////////////////////////////////////////////////////
#define pi_over_180 0.0174532925f
#define _180_over_pi 57.2957795130823208767981548141052f

float
deg2rad( float ang )
{
  return ang * pi_over_180;
}

float
rad2deg( float ang )
{
  return ang * _180_over_pi;
}

#endif